#!/usr/bin/perl 
# P860 version 0.91 -- date: Sept 23 2008
# Philips P860 emulator with 32k 16 bit words
#
my $version = 0.91;
# limitations:
#=============
# 1) the memory protection option is not implemented, so there is no "user mode"
#    (LKM does not switch to system mode, and the SMD operation acts as a no
#    operation. The cpu is always in system mode; no checking on illegal use
#    of registers and instructions in user mode)
# 2) only high speed arithmetic and the multiplex channel options are 
#    implemented (no dma)
#
# modifications:
#===============
# version 0.5
#  - simulated card reader added + mpx channel logic update
# version 0.6
#  - simulated tty adapted for MSWin32 (added module IO:Handle + autoflush for STDOUT)
# version 0.7
#  - added loading of relocatable load modules
#  - moving head disk support via mpx channel (2 drives md0, md1 on a single control unit)
#  - disk boot from unit md0 (+ IPL from disk)
# version 0.8
#  - error removed from routine format10. The condition(r1) was not in the requested
#    position for the test by evalcond. Therefor absolute conditional branches failed
#    sometimes.
#  - for the TTY (console) either the ASCII 64 character set (ASR33,..) or the 128
#    character set (glass tty) can be selected 
# version 0.9
#  - The relocatable load module format is adapted to the origional one:
#    word 3 is the program length in words, but counted in bytes (load <filename.rel> <address>)
#  - The disk bootstrap is adapted to the sector ident of the origional diskimage (sector ident= cyl number)
#  - The diskdriver is adapted to a sector ident = cylinder number
#
# version 0.91
#  - @MD0 $MD0[OTR]=\&otr_mdo; => $MD0[OTR]=\&otr_md0;
#  - absolute loader: 	$wc = nextword();            # word count
#			            $ba = nextword() + 2;        # begin address of load area
#                   =>	$wc = nextword();                  # word count
#            			$ba = (nextword() + 2) & 0xFFFE;   # begin address of load area
#  - relocatable loader
#						$M[$ba+($k*2)] = $M[$ba+($k*2)] + $ba; # relocate
#                   =>  $M[$ba+($k*2)] = ($M[$ba+($k*2)] + $ba) & 0xFFFF ; # relocate
#  - boot ptr command
#                       if($w[2] = "h")
#				    =>  if($w[2] eq "h")
#				        if($w[2] = "r")
#				    =>  if($w[2] eq "r")
#  - boot md0 command
#		                if($w[1] eq "md0")
#		            =>  elsif($w[1] eq "md0")
#  - 1 complement instruction (line 4047)
#	                    if($r1 != 0)
#                   =>	if($r1 == 0)
#  - cio_clk routine (line 5138)
#	                    if($f = 0)
#	                =>  if($f == 0)
#  - interrupt: save the whole psw on the a15 stack (was save PL/CR)
#               a15 stack overflow in case of a15 < /100
#  - call function: save the whole psw on the stack (was save PL/CR)
#                   a15 stack overflow in case of a15 < /100
#  - MSR(MLR)       adapted for a15 being stored on the stack when msr uses a15
#                   a15 stack overflow in case of a15 < /100
#  - rtc 10Hz => 50Hz
#  - rtc service routine: status not to $R[$r3] when rtc goes to inactive
#  - load command
#    mode 'a'           print "Load $w[2]\n";
#                   =>  print "Load $w[1]\n";
#    mode 'r'           print "Load $w[3]\n";
#                   =>  print "Load $w[1]\n";
#    added the possibility to load an .LM
#    added the possibility to load an .rel or .LM with address mon (=> load at 0)
#  - tst routines of all devices:
#    in case device not busy make $R[$r3] =0;
#  - AB instruction: position the condition code, so added:	$r3 = $r3 >> 1; # position the condition code
#  - ML instructions: CR must be set by value A1 (	$R[2] = setCRvalue($R[2]);)
#  - SLN/DLN changed: bit 0 and bit 1 comparison was wrong
#  - CC* r1,m / CC* r1,m,r2 adapted: t1 value not maske with 0xFF
#  - RIT instruction: allow reseting more than one internal interrupt in a single RIT instruction
#  - 2 * ierror() changed into ilcerror() (line 3754 and 4217)
#  - DRN updated
#  - DOUBLE SUBTRACT: UNDERFLOW FORM:  # SIGN => DPSIGN
#  - Divide code: give the remainder the sign of the dividend
#  - $pcsaved made global + mpxerror() introduced
#  - CWK changed to make it possible to compare the long constant with the right value of register P (instruction address + 4)
#  - change LDKL, ADKL SUKL as well?? (other long constant instructions do not allow A0 in r1)
#  - trace command extended with option to write the trace output in a file
#  - lineprinter logic extented: either select a skip before or after print
#    current selection is: before print
#    changed are: the printer driver and the LPT device block page parameters; SKIP_NO added
#  - tty service input routine: input of /0A => print CR
#  - time measurement changed because it went wrong with Continue
#    split now in ctime and time: time for the accumulated ticks and ctime for the current time
#    (current-time = time + (current interval - interval)
#                      ^accu   +        ^ executed part on curent interval
#  - bug removed from TM and TNM instructions
#
#
# Author: Theo Engel
# Info:   Info@theoengel.nl
#
# Usage:  p860-xxx.pl [<do-script-name>]  (where xxx is some version number)
#
# The emulator is built as a prototype, because at least some hardware details of the original 
# P860 are still lacking. The emulator is partly based on concepts used in SIMH (e.g. the event 
# mechanism). Also the control commands mirror some of the SIMH scp commands.
# (for SIMH, look at simh.trailing-edge.com)
#
# Currently implemented control commands:
# - run [<address>]										(includes reset-all = reset cpu+tty+rtc+ptr+ptp+lpt+crd+md0+md1)
# - go  [<address>]                                     (like run, but no reset)
# - continue|c
# - step|s [<nbr of instructions to execute>]           (no argument = execute 1 instruction)
# - do <filename>                                       (filename is name of the do-script)
# - breakpoint|bp set|s   <breakpoint address>
# - breakpoint|bp clear|c <breakpoint address>|all|a
# - breakpoint|bp print|p                               (show active breakpoints)
# - examine|ex <register>                               (show register content)
# - examine|ex m <address>                              (interactively shows content of <address>, <address>+2, .. until s is replied)
# - deposit|de <register> [<16 bit hex value>]          (in case no content specified, then it is requested from the keyboard)
# - deposit|de m <address>                              (interactively changes content of <address>, <address>+2, .. until s is replied)
# - deposit|de m <address> <16 bit hex value>
# - reset [cpu|tty|rtc|ptr|ptp|lpt|crd|md0|md1|mem|power] (no argument = reset-all; reset power = reset-all + reset-mem + P=0)
# - trace [on|off] [<filename>]                         (no argument = on; traceoutput => file in case file is specified)
# - load <filename.abs>                                 (load absolute (8*8 format) executable with an .abs extension)
# - load <filename.rel>|<filename.LM> <address>|mon     (load relocatable with a .rel or .LM extension at <address>) (mon loads a monitor at address 0)
# - boot ptr [h|r]                                      (h=halt,r= run after boot; default = run, starting at 1st address)
# - boot md0 [p] [h|r]                                  (ipl's from md0; p requests the in the ipl embedded papertape bootstrap, h=halt,r=run after boot, default=run)
# - attach|att ptr|ptp|lpt|crd|md0|md1 <filename>       (attach a simulated peripheral to a file)
# - detach|det ptr|ptp|lpt|crd|md0|md1
# - state [cpu|tty|rtc|ptr|ptp|lpt|crd|md0|md1]			(no argument = state cpu+tty+rtc+ptr+ptp+lpt+crd+md0+md1)
# - queue|q                                             (show event queue)
# - echo <string>                                       (character string is displayed on console)
# - dump|du <ba> <ea> [<fn.abs>]                        (dump in hex memory area on screen or on papertape in abs 8+8 format)
# - console 64|128                                      (select either 64 or 128 character set)
# - exit
#
# |    means logical OR
# [..] means OPTIONAL argument
# <address> is hex value of 0..FFFF
# <register> is P,A0..A15
# <string> is double quoted character string and may include codes like \n
# The key combination CntrE stops the emulation
# The key combination CntrP simulates a Panel interrupt
# do-scripts do not support arguments
#
# The following devices are currently are supported:
# 1) TTY - keyboard/printer
# 2) High speed papertape reader (PTR) - Digitronics 25540EP, 333/600 cps
# 3) High speed papertape punch (PTP)  - Facit 4070, 75 cps
# 4) Real time clock (RTC)             - Variable frequency (set to 10Hz)
# 5) Line printer (LPT)                - Data products, 132col, 245/670 lpm
# 6) Card reader (CRD)                 - Documation M200, 285 cpm
# 7) Moving head disk (MD0, MD1)       - 2 * Philips X1210 
#
# Questions/Remarks
# 1) which part of the PSW is stored/restored on/from the stack?
#    assumptions:
#    - CF           : whole PSW stored on the stack
#    - Interrupt    : whole PSW stored on the stack
#    - Depending on the register being used, part of the PSW is restored with RTN:
#      RTN P855/P860 restores saved CR and PL; RTN A15 also enables interrupt (P852 also)
#
# 2) reset state? ENB=1? PL=00? This implementation has ENB=0 an PL=/63 at this moment
#
# 3) adapt LDKL, ADKL SUKL so they load proper A0?
#    CWK instruction changed to make it possible to compare the long constant
#    with the right value of register P (instruction address + 4)
#    still adapt LDKL, ADKL SUKL ?? (other long constant may not have A0 as r1)
#
# 4) The double shift instructions DLA,DRA,DLN,DRN bypass bit0 of A2
#    On the P855/P860?? and certainly on P852/P856/P857 bit0 of A2 is
#    not changed during shift actions. On later models P854 .. bit0
#    of A2 is always reset to 0 (also in case of siht count = 0). In this
#    implementation bit0 of A2 is always reset to 0 (at this moment)
#
# 5) For some instructions on the real machine the l/s bit is not significant 
#    and may contain a 0 or a 1. The assembler puts in those cases always a 0
#    in the l/s bit In this implementation (at this moment) a 0 is always assumed 
#    in the l/s bit in those cases. A 1 will result in the detection of an illegal
#    instruction.
#
#
use strict;
use bytes;
use Term::ReadKey;
use Time::HiRes qw(gettimeofday tv_interval);
use IO::Handle;

# ascii conversion table
my %ch;
$ch{0x20}=' ';
$ch{0x21}='!';
$ch{0x22}="\"";
$ch{0x23}="\#";
$ch{0x24}='$';
$ch{0x25}='%';
$ch{0x26}='&';
$ch{0x27}="\'";
$ch{0x28}='(';
$ch{0x29}=')';
$ch{0x2a}='*';
$ch{0x2b}='+';
$ch{0x2c}=',';
$ch{0x2d}='-';
$ch{0x2e}='.';
$ch{0x2f}="\/";
$ch{0x30}='0';
$ch{0x31}='1';
$ch{0x32}='2';
$ch{0x33}='3';
$ch{0x34}='4';
$ch{0x35}='5';
$ch{0x36}='6';
$ch{0x37}='7';
$ch{0x38}='8';
$ch{0x39}='9';
$ch{0x3a}="\:";
$ch{0x3b}="\;";
$ch{0x3c}='<';
$ch{0x3d}='=';
$ch{0x3e}='>';
$ch{0x3f}='?';
$ch{0x40}='@';
$ch{0x41}='A';
$ch{0x42}='B';
$ch{0x43}='C';
$ch{0x44}='D';
$ch{0x45}='E';
$ch{0x46}='F';
$ch{0x47}='G';
$ch{0x48}='H';
$ch{0x49}='I';
$ch{0x4a}='J';
$ch{0x4b}='K';
$ch{0x4c}='L';
$ch{0x4d}='M';
$ch{0x4e}='N';
$ch{0x4f}='O';
$ch{0x50}='P';
$ch{0x51}='Q';
$ch{0x52}='R';
$ch{0x53}='S';
$ch{0x54}='T';
$ch{0x55}='U';
$ch{0x56}='V';
$ch{0x57}='W';
$ch{0x58}='X';
$ch{0x59}='Y';
$ch{0x5a}='Z';
$ch{0x5b}='[';
$ch{0x5c}="\\";
$ch{0x5d}=']';
$ch{0x5e}='^';
$ch{0x5f}='_';

 # digit conversion
my %cvdig;             
$cvdig{'0'}=0x0;
$cvdig{'1'}=0x1;
$cvdig{'2'}=0x2;
$cvdig{'3'}=0x3;
$cvdig{'4'}=0x4;
$cvdig{'5'}=0x5;
$cvdig{'6'}=0x6;
$cvdig{'7'}=0x7;
$cvdig{'8'}=0x8;
$cvdig{'9'}=0x9;
$cvdig{'A'}=0xa;
$cvdig{'B'}=0xb;
$cvdig{'C'}=0xc;
$cvdig{'D'}=0xd;
$cvdig{'E'}=0xe;
$cvdig{'F'}=0xf;

# register conversion table
my %R16;
$R16{"P"}   = 0;	# 0000
$R16{"A0"}  = 0;	# 0000
$R16{"A1"}  = 0x2;	# 0010
$R16{'A2'}  = 0x4;	# 0100
$R16{'A3'}  = 0x6;	# 0110
$R16{'A4'}  = 0x8;	# 1000
$R16{'A5'}  = 0xA;	# 1010
$R16{'A6'}  = 0xC;	# 1100
$R16{'A7'}  = 0xE;	# 1110
$R16{'A8'}  = 0x1;	# 0001
$R16{'A9'}  = 0x3;	# 0011
$R16{'A10'} = 0x5;	# 0101
$R16{'A11'} = 0x7;	# 0111
$R16{'A12'} = 0x9;	# 1001
$R16{'A13'} = 0xB;	# 1011
$R16{'A14'} = 0xD;	# 1101
$R16{'A15'} = 0xF;	# 1111

# register index conversion table
my @A;
$A[0]  = 0x0;
$A[1]  = 0x2;
$A[2]  = 0x4;
$A[3]  = 0x6;
$A[4]  = 0x8;
$A[5]  = 0xA;
$A[6]  = 0xC;
$A[7]  = 0xE;
$A[8]  = 0x1;
$A[9]  = 0x3;
$A[10] = 0x5;
$A[11] = 0x7;
$A[12] = 0x9;
$A[13] = 0xB;
$A[14] = 0xD;
$A[15] = 0xF;
my @Reg;
$Reg[0]  = 'A0';
$Reg[1]  = 'A8';
$Reg[2]  = 'A1';
$Reg[3]  = 'A9';
$Reg[4]  = 'A2';
$Reg[5]  = 'A10';
$Reg[6]  = 'A3';
$Reg[7]  = 'A11';
$Reg[8]  = 'A4';
$Reg[9]  = 'A12';
$Reg[10] = 'A5';
$Reg[11] = 'A13';
$Reg[12] = 'A6';
$Reg[13] = 'A14';
$Reg[14] = 'A7';
$Reg[15] = 'A15';

# register symbols
use constant P   => 0x0;
use constant A0  => 0x0;
use constant A1  => 0x2;
use constant A2  => 0x4;
use constant A3  => 0x6;
use constant A4  => 0x8;
use constant A5  => 0xA;
use constant A6  => 0xC;
use constant A7  => 0xE;
use constant A8  => 0x1;
use constant A9  => 0x3;
use constant A10 => 0x5;
use constant A11 => 0x7;
use constant A12 => 0x9;
use constant A13 => 0xB;
use constant A14 => 0xD;
use constant A15 => 0xF;

####################### emulator controller data ############################
my $DEBUG = 1;
my $StopOnIllInstr = 1;     # stop on illegal instruction
my $StopOnIllDev   = 1;     # stop on illegal device
my $StopOnMpxError = 1;     # stop on an illegal mpx condition

my $tracefile      = "";    # tracefile name in case specified
my $EmTrace = 0;			# trace on if 1, trace off 0
my $EmBreak = 0;            # 1 in case a breakpoint exist, else 0
my %BP;                     # breakpoint hash
my $program_started = 0;    # 1 if new program is started by RUN/STEP
# papertape format
my $null 	= 0;
my $rub     = 0377;         # start of tape
my $xoff    = 0224;         # end of tape

# time/event handling
my $interval;               # decreasing: number of instructions still to be executed till next event
my $currentinterval;		# fix: number of instructions to be executed till next event
my $event_queue;
my $queue_empty_time = 5000;
my $time;                   # accumulated number of executed instruction
my $ctime;                  # current time: always calculated form $time+executed part of current interval

# elapsed runtime measurement
my @t0;		                # start point elapsed time measurement during run command
my $secs;                   # run elapsed time in secs
my $cps;                    # cycles per second during run command
# real time clock synchronization
my @clk0;                   # start point of measurement period
my $laps;                   # actual measured period in second
my $pulses;                 # number of events in measured period

############################ emulator data ##################################
# cpu/memory resources and state
#===============================
my @M=();               # memory array
my $IR;                 # instruction register
my $PSW;                # status word
	my $PL;             # priority level 			psw bit 0:5
	my $CR;             # condition register		psw bit 6:7
	my $overflow=0;     # (aid to set CR, must be 0)
	my $RUN;            # run bit               	psw bit 8
	my $ENB;            # interrupt enable bit		psw bit 9
#	my $CP;             # control panel interrupt	psw bit 10
#	my $PWR;            # power fail interrupt      psw bit 11
#	my $RTC;            # real time clock interrupt psw bit 12
#	my $LKM;            # link monitor interrupt    psw bit 13
#						# + stack overflow and illegal instruction
#	my $MP;             # memory protect interrupt  psw bit 14
	my $UM;             # user mode indicator       psw bit 15
my @R;                  # registers A0..A15
my $pcsaved;
#
# Interrupt system
#=================
# maskeable interrupt bits of the common interrupt line are located
# in the lower 16 bits of the interrupt register.
# the 5 nonmaskeable interrupt bits (which are also part of the PSW) are
# located in the upper half of the interrupt register
#
my $intreg;             # interrupt register
my $intmsk;             # interrupt mask for maskeable interrupts 
#
# a maskeable interrupt is disabled when its flag = 1 (and enabled when its flag = 0)
# bit assignment of the interrupt register:
# maskeable interrupt flipflops of the common interrupt line
use constant ICR   => 0x0002;          # Card reader
use constant IPTR1 => 0x0004;          # Papertape reader 1
use constant ICT   => 0x0010;          # Cassette tape
use constant IPTP1 => 0x0020;          # Papertape punch 1
use constant IASR1 => 0x0080;          # Teletype 1
use constant IMT   => 0x0200;          # Mag tape
use constant IFHD  => 0x0400;          # Fixed head disk
use constant IMD0  => 0x0800;          # Moving head disk cu-A
use constant IMD2  => 0x1000;          # Moving head disk cu-B
use constant ILP   => 0x2000;          # Line printer
# nonmaskeable interrupt flipflops (as part of PSW)
use constant MP    => 0x010000;        # Memory protect (6)
use constant LKM   => 0x020000;        # Link monitor, stackoverflow, illegal instruction (1)
use constant RTC   => 0x040000;        # Real time clock (2)
use constant PWR   => 0x080000;        # Power failure, auto restart (0)
use constant CP    => 0x100000;        # Panel interrupt (7)
#
# Interrupt locations
#
use constant AdPWR  => 0x40;           # power failure interrupt location (highest priority = level 0)
use constant AdLKM  => 0x42;           # link monitor (priority level 1)
use constant AdRTC  => 0x44;           # real time clock (priority level 2)
use constant ilres1 => 0x46;           # reserved (memory incr data break option (priority level 3))
use constant ilres2 => 0x48;           # reserved (priority level 4)
use constant AdCOM  => 0x4A;           # common line (priority level 5)
use constant ilres3 => 0x4C;           # reserved (memory protection option (priority level 6))
use constant AdCP   => 0x4E;           # control panel (priority level 7)

# working registers for CPU, CU's, etc simulation
my $opc;                     # operation code bit1:4
my $r1;                      # r1 or cnd shifted 1 left	
my $r2;                      # r2
my $r3;                      # r3             (in format 0 instructions)
my $k;                       # short constant (in format 0 instructions)
my $sc;                      # shiftcount     (in format 0 instructions)
my $sfun;                    # shift function (in format 0 instructions)
my $md;                      # adressing mode (format 1)
my $ls;                      # load/store bit (format 1)
my $t1;                      # temporary 1 (value)
my $t2;                      # temporary 2 (value)
my $t3;                      # temporary 3 (value)
my $a;                       # temporary for effective address computation

# masks
use constant mskf10 => 0x8000;      	      # format 1 instruction with ls bit 0

use constant SIGN    => 0x8000;               # sign bit short
use constant DPSIGN  => 0x40000000;		      # 31 bit double precision sign (in 32bit word)
use constant DMASK   => 0xFFFF;        	      # data mask short
use constant DPDMASK => 0x7FFFFFFF;           # 31 bit double precision data mask (in 32bit word)
use constant MMASK   => DMASK & (~ SIGN); 	  # magnitude mask short
use constant DPMMASK => DPDMASK & (~ DPSIGN); # 31 bit magnitude mask (in 32bit word)

# devices/control units
#
# The control unit STATE consists of two parts
# 1) ls 16 bits  = device status word, different for each device, but
#    bit 15 is 1 always means: device not operable
#
# 2) 4 bits before the ls 16 bits
#    
#    bit 2:    0 = output request
#              1 = input request
#    bit 3:4  00 = inactive state
#             01 = execute state
#             11 = exchange state
#             10 = wait state
#    bit 1     1 = stop io requested
#    the device is busy if not in inactive state
#
# control unit states (the 4 bits as part of STATE)
use constant INAC => 0;                  # inactive
use constant EXE  => 1;                  # execute
use constant IEXE => (0b100 + 0b001);    # input  + execute
use constant OEXE => (0b000 + 0b001);    # output + execute
use constant EXC  => 3;                  # exchange
use constant IEXC => (0b100 + 0b011);    # input  + exchange
use constant OEXC => (0b000 + 0b011);    # output + exchange
use constant WAIT => 2;                  # wait state
use constant WREQ => 0b1000;             # stop io (or wait state) requested

# device control block declaration
my @TTY;
my @PTR;
my @PTP;
my @CLK;
my @LPT;
my @CRD;
my @MD0;
my @MD1;

# device control block layout
use constant NEXT		=> 0;
use constant NAME   	=> 1;
use constant TIMEOUT	=> 2;
use constant SERVICE   	=> 3;
use constant STATE   	=> 4;
use constant BUFFER		=> 5;
use constant TIME		=> 6;
use constant CIO   		=> 7;
use constant INR		=> 8;
use constant OTR   		=> 9;
use constant SST		=> 10;
use constant TST   		=> 11;
use constant DSTAT      => 12;
use constant RESET      => 13;
use constant FILE       => 14;
use constant FC         => 15;

# tty device control block (teletype keyboard/printer - programmed channel)
# tty specific entries in device control block
use constant CSET => 14;     # 64 or 128 Ascii character set

$TTY[NEXT]  	= 0;					# next pointer field
$TTY[NAME]  	= "tty";                # device name
$TTY[TIMEOUT]  	= 0;                    # event timeout value
$TTY[SERVICE]  	= \&svc_tty;            # event service routine
$TTY[STATE]  	= 0;                    # control unit state
$TTY[BUFFER]  	= 0;                    # TTY buffer
$TTY[TIME]      = 5000;                 # default device event time
$TTY[CIO]  		= \&cio_tty;            # cio processor
$TTY[INR]  		= \&inr_tty;            # inr processor
$TTY[OTR]  		= \&otr_tty;            # otr processor
$TTY[SST]  		= \&sst_tty;            # sst processor
$TTY[TST] 		= \&tst_tty;            # sst processor
$TTY[DSTAT]     = \&state_tty;          # display state
$TTY[RESET]     = \&reset_tty;          # reset device
$TTY[CSET]    = 64;                   # default is 64 character set 

# ptr device control block (paper tape reader - programmed channel )
$PTR[NEXT]      = 0;                    # next pointer field
$PTR[NAME]      = "ptr";                # device name
$PTR[TIMEOUT]  	= 0;                    # event timeout value
$PTR[SERVICE]  	= \&svc_ptr;            # event service routine
$PTR[STATE]  	= 0;                    # control unit state
$PTR[BUFFER]  	= 0;                    # PTR buffer
$PTR[TIME]      = 250;                  # default device event time
$PTR[CIO]  		= \&cio_ptr;            # cio processor
$PTR[INR]  		= \&inr_ptr;            # inr processor
$PTR[OTR]  		= 0;                    # otr processor
$PTR[SST]  		= \&sst_ptr;            # sst processor
$PTR[TST] 		= \&tst_ptr;            # sst processor
$PTR[DSTAT]     = \&state_ptr;          # display state
$PTR[RESET]     = \&reset_ptr;          # reset device
$PTR[FILE]      = "";                   # attached file (name; 0 if not attached/inoperable)
$PTR[FC];                               # file handle of attached file

# ptp device control block (paper tape punch - programmed channel)
$PTP[NEXT]      = 0;                    # next pointer field
$PTP[NAME]      = "ptp";                # device name
$PTP[TIMEOUT]  	= 0;                    # event timeout value
$PTP[SERVICE]  	= \&svc_ptp;            # event service routine
$PTP[STATE]  	= 0;                    # control unit state
$PTP[BUFFER]  	= 0;                    # PTP buffer
$PTP[TIME]      = 750;                  # default device event time
$PTP[CIO]  		= \&cio_ptp;            # cio processor
$PTP[INR]  		= 0;                    # inr processor
$PTP[OTR]  		= \&otr_ptp;            # otr processor
$PTP[SST]  		= \&sst_ptp;            # sst processor
$PTP[TST] 		= \&tst_ptp;            # sst processor
$PTP[DSTAT]     = \&state_ptp;          # display state
$PTP[RESET]     = \&reset_ptp;          # reset device
$PTP[FILE]      = "";                   # attached file (name; 0 if not attached/inoperable)
$PTP[FC];                               # file handle of attached file

# crd device control block and buffer (card reader - multiplex channel )
# card reader specific entries in device control block
use constant RTIME      => 16;          # card moving time
use constant RCLEN      => 17;          # read number of bytes

my $crdbuf;                             # card reader buffer

$CRD[NEXT]      = 0;                    # next pointer field
$CRD[NAME]      = "crd";                # device name
$CRD[TIMEOUT]  	= 0;                    # event timeout value
$CRD[SERVICE]  	= \&svc_crd;            # event service routine
$CRD[STATE]  	= 0;                    # control unit state
$CRD[BUFFER]  	= 0;                    # buffer
$CRD[TIME]      = 5;                    # mpx character transfer time
$CRD[CIO]  		= \&cio_crd;            # cio processor
$CRD[INR]  		= \&inr_crd;            # inr processor
$CRD[OTR]  		= 0;                    # otr processor
$CRD[SST]  		= \&sst_crd;            # sst processor
$CRD[TST] 		= \&tst_crd;            # sst processor
$CRD[DSTAT]     = \&state_crd;          # display state
$CRD[RESET]     = \&reset_crd;          # reset device
$CRD[FILE]      = "";                   # attached file (name; 0 if not attached/inoperable)
$CRD[FC];                               # file handle of attached file
$CRD[RTIME]     =10000;                 # time to move card to read station
$CRD[RCLEN]     = 0;                    # read length in bytes

# lpt device control block and buffer (line printer - multiplex channel)
# printer specific entries in printer device control block
use constant PTIME      => 16;          # buffer print time
use constant LMAX       => 17;          # 60 print lines per page
use constant NPAG       => 18;          # nbr of lines to skip between print pages
use constant LCUR       => 19;          # current line on page
# the following entry directs the page printing (skip page before or after printing the buffer).
use constant SKIP       => 20;          
# this entry must be set to one of the following two constants.
use constant SKIP_BP    => 1;           # skip page before printing the buffer
use constant SKIP_AP    => 2;           # skip page after  printing the buffer
use constant SKIP_NO    => 3;           # no printer control at all, store chars forwarded to lpt in the printfile

my $lptbuf;                             # lineprinter buffer

$LPT[NEXT]      = 0;                    # next pointer field
$LPT[NAME]      = "lpt";                # device name
$LPT[TIMEOUT]  	= 0;                    # event timeout value
$LPT[SERVICE]  	= \&svc_lpt;            # event service routine
$LPT[STATE]  	= 0;                    # control unit state
$LPT[BUFFER]  	= "";                   # buffer
$LPT[TIME]      = 5;                    # mpx character transfer time
$LPT[CIO]  		= \&cio_lpt;            # cio processor
$LPT[INR]  		= 0;                    # inr processor
$LPT[OTR]  		= \&otr_lpt;            # otr processor
$LPT[SST]  		= \&sst_lpt;            # sst processor
$LPT[TST] 		= \&tst_lpt;            # sst processor
$LPT[DSTAT]     = \&state_lpt;          # display state
$LPT[RESET]     = \&reset_lpt;          # reset device
$LPT[FILE]      = "";                   # attached file (name; 0 if not attached/inoperable)
$LPT[FC];                               # file handle of attached file
$LPT[PTIME]     = 5000;                 # buffer print time
# page parameters: 66 lines per page, of which 64 being printed + 1 empty line at top and end
# (in case of no printer control chars and printer control is set by SKIP_AP or SKIP_BP)
$LPT[LMAX]      = 64;                   # 64 print lines per page
$LPT[NPAG]      = 2;                    # nbr of lines to skip between print pages
$LPT[LCUR]      = 0;                    # current line on page
$LPT[SKIP]      = SKIP_BP;              # skip page before printing the buffer
#$LPT[SKIP]      = SKIP_AP;              # skip page after  printing the buffer

# X1210 moving head disk device control blocks (1 per drive) connected to multiplex channel
#
# two drives per control unit; the control and drive 0 data are in the md0 control
# block. the drive specific data of the 2nd drive are in md1. only one drive can
# use the control unit and multiplex channel at a time; even seeks can not be
# overlapped.
#
# moving head disk specific constants
use constant RWTIME => 15;              # latency to move to sector in cylinder
use constant TRTIME => 16;              # time to move 1 cylinder
use constant XLEN   => 17;              # transferred length in bytes
use constant DN     => 18;              # active drive == drive with a pending operation
use constant CYL    => 19;              # current cylinder
use constant SEC    => 20;              # sector to read or write

# resulting state of a disk operation
use constant DREADY   => 0x4000; # unit ready
use constant DONCYL   => 0x0400; # on requested cylinder
use constant DSEEKERR => 0x0200; # seek error
use constant DPROGERR => 0x0010; # program error
use constant DLENGERR => 0x0008; # incorrect length
use constant DDATAERR => 0x0004; # data error
use constant DTHROERR => 0x0002; # throughput error
use constant DNOTOP   => 0x0001; # not operable

# disk parameters (X1210)
use constant CYLS     => 203;    # ncyl 0..202 (203 tracks/cylinders)
use constant NH       => 2;      # number of heads
use constant SPT      => 16;     # 16 sectors per track
use constant LSEC     => 205;    # sector length in 16 bit words

my $md0buf;                             # internal disk buffer

# drive0 + control unit (device number = 0)
$MD0[NEXT]      = 0;                    # next pointer field
$MD0[NAME]      = "md0";                # device name
$MD0[TIMEOUT]  	= 0;                    # event timeout value
$MD0[SERVICE]  	= \&svc_md0;            # event service routine
$MD0[STATE]  	= 0;                    # control unit state
$MD0[BUFFER]  	= "";                   # buffer
$MD0[TIME]      = 4;                    # mpx transfer time (30 musec)
$MD0[CIO]  		= \&cio_md0;            # cio processor/driver selector
$MD0[INR]  		= \&inr_md0;            # inr processor
$MD0[OTR]  		= \&otr_md0;            # otr processor
$MD0[SST]  		= \&sst_md0;            # sst processor
$MD0[TST] 		= \&tst_md0;            # tst processor
$MD0[DSTAT]     = \&state_md0;          # display state
$MD0[RESET]     = \&reset_md0;          # reset device
$MD0[FILE]      = "";                   # attached file (name; "" if not attached/inoperable)
                                        # if attached, then also opened. filecode MD0
$MD0[RWTIME]    = 1000;                 # latency to move to sector in cylinder (10 msec)
$MD0[TRTIME]    = 100;                  # time to move 1 cylinder (1 msec)
$MD0[XLEN]      = 0;                    # transferred length in bytes
$MD0[DN]        = 0;                    # active drive == drive with a pending operation
$MD0[CYL]       = 0;                    # current cylinder
$MD0[SEC]       = 0;                    # physical sector to read or write

# drive1 (device number = 1)
$MD1[NAME]      = "md1";                # device name
$MD1[CIO]  		= \&cio_md1;            # cio processor/drive selector
$MD1[SST]       = \&sst_md0;
$MD1[TST]       = \&tst_md0;
$MD1[FILE]      = "";                   # attached file (name; "" if not attached/inoperable)
                                        # if attached, then also opened. filecode MD1
$MD1[CYL]       = 0;                    # current cylinder

# clk device control block (10Hz clock)
# clock specific entries in clock device control block
use constant FREQ  		=> 14;
use constant SYSEC      => 15;
use constant SYNC       => 16;
use constant CTIME      => 17;

$CLK[NEXT]      = 0;                    # next pointer field
$CLK[NAME]      = "rtc";                # device name
$CLK[TIMEOUT]  	= 0;                    # event timeout value
$CLK[SERVICE]  	= \&svc_clk;            # event service routine
$CLK[STATE]  	= 0;                    # control unit state
$CLK[BUFFER]  	= 0;                    # buffer
$CLK[TIME]      = 3000;                 # default device event time
$CLK[CIO]  		= \&cio_clk;            # cio processor
$CLK[INR]  		= \&inr_clk;            # inr processor
$CLK[OTR]  		= \&otr_clk;            # otr processor
$CLK[SST]  		= \&sst_clk;            # sst processor
$CLK[TST] 		= \&tst_clk;            # sst processor
$CLK[DSTAT]     = \&state_clk;          # display state
$CLK[RESET]     = \&reset_clk;          # reset device
$CLK[FREQ]      = 50;                   # 50 HZ clock
$CLK[SYSEC]     = 2;                    # sync measurement period (secs) 
$CLK[SYNC]      = $CLK[SYSEC] * $CLK[FREQ]; # number sync pulses for measurement period
$CLK[CTIME]     = 0;                    # calculated synced device event time

# defined devices
my %DEV;
$DEV{0x0D} = \@LPT;
$DEV{0x10} = \@TTY;
$DEV{0x20} = \@PTR;
$DEV{0x30} = \@PTP;
$DEV{0x3F} = \@CLK;
$DEV{0x05} = \@CRD;
$DEV{0x02} = \@MD0; # the cu-A has device address 2, device number is 0
$DEV{0x12} = \@MD1; # the cu-A has device address 2, device number is 1

# speudo device for stepping function
my @STEP;             
$STEP[NEXT] 	= 0;					# next pointer field
$STEP[NAME] 	= "step";               # device name
$STEP[TIMEOUT] 	= 0;                    # event timeout value
$STEP[SERVICE] 	= \&svc_step;           # event service routine
# stepper service routine (scheduled after stepper event)
sub svc_step {
	$RUN = 0;          # stop cpu
}

# Multiplex channel
# Each multiplex channel uses 2 words at fixed addresses in memory to control the channel.
# 15 cu's max can be connected to multiplex channels. The multiplex channel used be a 
# particular device is determed by it device address. The multiplex channels are processed
# in priority order: channel 1 has the highest priority.

# Multiplex channel addresses and registers
#                          bit15
# brkreg 0111 1111 1111 1111
#         ^                ^
#         channel 15       channel 1 break request (highest priority)
#
my $brkreg;             # break register
my $mpx1;               # multiplex word 1 register
my $mpx2;               # multiplex word 2 register
my $tx3;				# temporary for multiplex channel processing
my $tx4;                # temporary for multiplex channel processing
my $xf;                 # function bit
# channel memory addresses
my @MPX;
$MPX[0x1] = 0x84; # channel 1 uses word 84 and 86
$MPX[0x2] = 0x88; # moving head disk cu-A with address 0x02 uses this channel
$MPX[0x3] = 0x8C; # moving head disk cu-B with address 0x03 uses this channel
$MPX[0x4] = 0x90;

$MPX[0x5] = 0x94; # card reader with address 0x05 uses this channel
$MPX[0x6] = 0x98;
$MPX[0x7] = 0x9C;
$MPX[0x7] = 0xA0;

$MPX[0x9] = 0xA4;
$MPX[0xA] = 0xA8;
$MPX[0xB] = 0xAC;
$MPX[0xC] = 0xB0;

$MPX[0xD] = 0xB4; # line printer with address 0x0D uses this channel
$MPX[0xE] = 0xB8;
$MPX[0xF] = 0xBC; # channel 15

# Multiplex channel break flipflops in breakregister
use constant BMD0  => 0x0002;   # MH disk cu-A  addr 0x2 => channel  2 => bit 14
use constant BLP   => 0x1000;   # Line printer: addr 0xD => channel 13 => bit  3
use constant BCR   => 0x0010;   # Card reader:  addr 0x5 => channel  5 => bit 11 

######################### Emulator controller ##############################
print "P855/860 Emulator, version $version\n";
my $cmd;     # command string
my $file;
my $do = "";
my $n;
my $m;
my $s;
my @w; 		 # itemized command string
@w = @ARGV;  # possible initial do string

# do proc called at startup?
if($w[0] ne "")
{
	# yes
	$w[1] = $w[0];
	$w[0] = "do";
}

while(1)
{
	# read a command
	if($do)
	{
		$cmd = <DO>;
		if($cmd eq "")
		{
			close DO || die "can't close $do: $!";			
			$do = ""; # reset after eof
		}
		else
		{
			print $cmd;
			chop $cmd;
			@w = split(/ /, $cmd);			
		}
	}
	
	# in case no command from a do string, then from keyboard
	if($w[0] eq "")
	{	
		print "> ";
		$cmd = <STDIN>;
		chop $cmd;
		@w = split(/ /, $cmd);
	}

	if($w[0] eq "run")
	{
		if($program_started)
		{
			print "Proceed with \"continue\", \"step\" or \"reset\"\n";
		}
		else
		{
			if($w[1] ne "")
			{
				$n = cvhex(UP($w[1]));
				$n = $n & 0xFFFE;
				$R[0] = $n;                    # set PC
			}
			$interval = 0;
			$currentinterval = 0;
			$time = 0;
			$event_queue = 0;
			reset_all();
			# tty active to listen to CntrP (panel interrupt) or
			#                         CntrE (stop simulation)
			activate_event(\@TTY,$TTY[TIME]); 
			@t0 = gettimeofday ();             # fix t0 to measure elapsed time
			$RUN = 1;
			ReadMode 4;                        # prepare simulated console mode
			STDOUT->autoflush(1);              # for MSWin32
			while($RUN)
			{
				instr();
			}
			ReadMode 0;                        # reset console
			STDOUT->autoflush(0);              # for MSWin32
			$secs = tv_interval (\@t0);        # elapsed run time in secs
			$ctime = $time + ($currentinterval - $interval);   # accumulated runned cycles
			$cps = int($ctime / $secs);        # cycles per second
			$s = prhex(16,$R[0]);
			$program_started = 1;              # program is still in a running state
			print "\nRun: emulation stopped at address $s\n";
			print "$ctime instructions in $secs seconds: $cps instructions per second\n";
		}
		mem_check(1);
	}
	elsif($w[0] eq "go")
	{
		if($program_started)
		{
			print "Proceed with \"continue\", \"step\" or \"reset\"\n";
		}
		else
		{
			if($w[1] ne "")
			{
				$n = cvhex(UP($w[1]));
				$n = $n & 0xFFFE;				
				$R[0] = $n;                    # set PC
			}
			$interval = 0;
			$currentinterval = 0;
			$time = 0;
			$event_queue = 0;
			# tty active to listen to CntrP (panel interrupt) or
			#                         CntrE (stop simulation)
			activate_event(\@TTY,$TTY[TIME]); 
			@t0 = gettimeofday ();             # fix t0 to measure elapsed time
			$RUN = 1;
			ReadMode 4;                        # prepare simulated console mode
			STDOUT->autoflush(1);              # for MSWin32
			while($RUN)
			{
				instr();
			}
			ReadMode 0;                        # reset console
			STDOUT->autoflush(0);              # for MSWin32
			$secs = tv_interval (\@t0);        # elapsed run time in secs
			$ctime = $time + ($currentinterval - $interval);   # accumulated runned cycles
			$cps = int($ctime / $secs);         # cycles per second
			$s = prhex(16,$R[0]);
			$program_started = 1;              # program is still in a running state
			print "\nRun/Go: emulation stopped at address $s\n";
			print "$ctime instructions in $secs seconds: $cps instructions per second\n";
		}
		mem_check(1);
	}	
	elsif($cmd eq "continue" || $cmd eq "c")
	{
		if(! $program_started)
		{
			print "Continue only after a run, go or step\n";
		}
		else
		{
			$RUN = 1;
			ReadMode 4;                        # prepare simulated console mode
			STDOUT->autoflush(1);              # for MSWin32
			while($RUN)
			{
				instr();
			}
			ReadMode 0;                        # reset console
			STDOUT->autoflush(0);              # for MSWin32
			$ctime = $time + ($currentinterval - $interval);   # accumulated runned cycles
			$s = prhex(16,$R[0]);
			print "\nContinue: emulation stopped at address $s, $ctime instructions processed\n";
		}
		mem_check(1);
	}
	elsif($w[0] eq "step" || $w[0] eq "s")
	{
		print "step $w[1]\n";
		$n = 1;                              # default nbr of steps
		if($w[1] ne "")
		{
			if( ($n = cvdec(UP($w[1]))) <= 1)
			{
				$n=1;
			};
		}
		if(! $program_started)
		{
			$interval = 0;
			$currentinterval = 0;
			$time = 0;
			$event_queue = 0;
			reset_all();
			# tty active to listen to CntrP (panel interrupt) or
			#                         CntrE (stop simulation)
			activate_event(\@TTY, $TTY[TIME]); 
			$program_started = 1;              # program is in a running state
		}
		activate_event(\@STEP, $n);          # set stepper event
		$RUN = 1;
		ReadMode 4;                          # prepare simulated console mode	
		STDOUT->autoflush(1);                # for MSWin32			
		while($RUN)
		{
			instr();
		}
		ReadMode 0;                          # reset console
		STDOUT->autoflush(0);                # for MSWin32
		$ctime = $time + ($currentinterval - $interval);   # accumulated runned cycles
		$s = prhex(16,$R[0]);
		print "Step: $ctime instruction(s) (remaining interval steps $interval), now at address $s\n";
		mem_check(1);
	}
	elsif($w[0] eq "do")
	{
		$do = $w[1];
		if(! open (DO, $do))
		{
			print "Unable to open file: $do\n";
			$do = "";
		}
	}
	elsif($w[0] eq "breakpoint" || $w[0] eq "bp")
	{
		if($w[1] eq "set" || $w[1] eq "s")
		{
			if(($n = cvhex(UP($w[2]))) >= 0)
			{
				$n = $n & 0xFFFE;
				$BP{$n} = $n;
			}
			else
			{
				print "Wrong address, no breakpoint set\n";
			}
		}
		elsif($w[1] eq "clear" || $w[1] eq "c")
		{
			if($w[2] eq "all" || $w[2] eq "a")
			{
				%BP = ();
			}
			else
			{
				$n = cvhex(UP($w[2]));
				$n = $n & 0xFFFE;
				delete $BP{$n};
			}
		}
		elsif($w[1] eq "print" || $w[1] eq "p")
		{
			print "Defined Breakpoints\n";
			foreach $n (sort keys( %BP))
			{
				print prhex(16,$n) . "\n";
			}
		}
		else
		{
			print "????\n";
		}
		$n = scalar(keys(%BP));
		if($n)
		{
			$EmBreak = 1;
		}
		else
		{
			$EmBreak = 0;
		}
		print "$n Breakpoints defined (EmBreak=$EmBreak)\n";
	}
	elsif($w[0] eq "examine" || $w[0] eq "ex")
	{
		if(UP($w[1]) =~ /^M/)
		{
			$s = UP($w[2]);
			if(($m=cvhex($s)) >= 0)
			{
				$m=$m & 0xFFFE;
				while(1)
				{
					print prhex(16, $m) . "  " . prhex(16,$M[$m]) . " ";
					$s = <STDIN>;
					$m = $m + 2;
					if(($s =~ /s/) || ($m > 0xFFFF))
					{
						last;
					}
				}
			}
			else
			{
				print "Wrong or no hex address\n";
			}
		}
		elsif(($n=$R16{UP($w[1])}) ne undef)
		{
			print "$w[1] = " . prhex(16,$R[$n]) . "\n";
		}
		else
		{
			print "Wrong memory or register reference $w[1]\n";
		}
	}
	elsif($w[0] eq "dump" || $w[0] eq "du")
	{
		my $str;
		my $c;
		$s = UP($w[1]);		
		if(($m=cvhex($s)) >= 0)
		{
			$m=$m & 0xFFE0;
			$s = UP($w[2]);	
			if(($n=cvhex($s)) >= 0)
			{
				if($w[3] eq "")
				{
					# dump on screen				
					while($m < $n)
					{
						$str="";
						$s=0;
						print prhex(16,$m) . " ";
						while($s<8)
						{
							print prhex(16,$M[$m]);
							$c=$ch{($M[$m] >> 8) & 0xFF};
							if($c ne undef)
							{
								$str = $str . $c;
							}
							else
							{
								$str = $str . ".";
							}
							$c=$ch{$M[$m] & 0xFF};
							if($c ne undef)
							{
								$str = $str . $c;
							}
							else
							{
								$str = $str . ".";
							}						
							$m=$m+2;
							$s=$s+1;
						}
						print " ";
						$s=0;
						while($s<8)
						{
							print prhex(16,$M[$m]);
							$c=$ch{($M[$m] >> 8) & 0xFF};
							if($c ne undef)
							{
								$str = $str . $c;
							}
							else
							{
								$str = $str . ".";
							}
							$c=$ch{$M[$m] & 0xFF};
							if($c ne undef)
							{
								$str = $str . $c;
							}
							else
							{
								$str = $str . ".";
							}						
							$m=$m+2;
							$s=$s+1;
						}					
						print "  $str\n";
					}
				}
				else
				{
					# dump on an absolute papertape
		            # print "$w[3]\n";					
					if($w[3] =~ /\.abs/)
					{
						open (BIN,">:raw", $w[3]) || die "cannot open outputfile: $w[3]: $!";
						print "\nPunch in absolute 8+8 papertape load format: $w[3] \n\n";
						my $byte;
						my $w;
						my $chksum=0;
                        my $low=$m;
						my $high=$n+2;
						my $i;
						
						punch($rub);                         # rubout
						$w = ($high >> 1) - ($low >> 1);     # word count (high = endaddr + 2)
						$byte = $w >> 8;
						punch($byte);
						$byte = $w & 0xFF;
						punch($byte);

						$w = $low-2;	                     # begin address - 2
						$byte = $w >> 8;
						punch($byte);
						$byte = $w & 0xFF;
						punch($byte);

						$i = $low;                           # begin address 

						while($i < $high)                    # content
						{
							$w = $M[$i];
							$chksum = $chksum ^ $w;
							$byte = $w >> 8;
							punch($byte);
							$byte = $w & 0xFF;
							punch($byte);
							$i = $i + 2;
						}

						$w=$chksum;                          # checksum
						$byte = $w >> 8;
						punch($byte);
						$byte = $w & 0xFF;
						punch($byte);
						punch($xoff);                        # xoff	
						print "Papertape ready\n";
						close BIN || die "can't close $w[3]: $!";
					}
					else
					{
						print "Wrong filename: no .abs extension\n";
					}				
				}
			}
		}					
	}
	elsif($w[0] eq "deposit" || $w[0] eq "de")
	{
		if(UP($w[1]) =~ /^M/)
		{
			$s = UP($w[2]);                   # address
			if(($s=cvhex($s)) >= 0)
			{
				$s=$s & 0xFFFE;
				$m = UP($w[3]);               # content
				if(($m=cvhex($m)) >= 0)
				{
					$M[$s] = $m;
				}
				else
				{
					# read content from keyboard
					while(1)
					{
						print prhex(16, $s) . "  " . prhex(16,$M[$s]) . " ";
						$m = <STDIN>;
						if($m =~ /s/)
						{
							last;                 # stop
						}
						elsif($m =~ /[\da-fA-F]/)
						{
							if(($m=cvhex(UP($m))) >= 0)
							{
								$M[$s] = $m;
							}
							else
							{
								print "Wrong content\n";
								$s = $s - 2;       # redo current address
							}
						}
						$s = $s + 2;
					}
				}
			}
			else
			{
				print "Wrong or no hex address\n";
			}
		}
		elsif(($n=$R16{UP($w[1])}) ne undef)
		{
			if($w[2] eq undef)
			{
				print "$w[1] = ";
				$s = <STDIN>;			# input 16 bit hex number
				chomp $s;
				$s = UP($s);
				if(($m=cvhex($s)) >= 0)
				{
					$R[$n] = $m;
				}
				else
				{
					print "Error input hex number\n";
				}
			}
			else
			{
				$s = UP($w[2]);                   # content
				if(($m=cvhex($s)) >= 0)
				{
					$R[$n] = $m;
				}
				else
				{
					print "Error input hex number\n";
				}
			}
		}
		else
		{
			print "Wrong register $w[1]\n";
		}
	}
	elsif($w[0] eq "reset")
	{
		if($w[1] eq "cpu")
		{
			print "CPU reset\n";
			reset_cpu();
		}
		elsif($w[1] eq "tty")
		{
			print "TTY reset\n";
			reset_tty();
		}
		elsif($w[1] eq "rtc")
		{
			print "Realtime clock reset\n";
			reset_clk();
		}
		elsif($w[1] eq "ptr")
		{
			print "Papertape reader reset\n";
			reset_ptr();
		}	
		elsif($w[1] eq "ptp")
		{
			print "Papertape punch reset\n";
			reset_ptp();
		}
		elsif($w[1] eq "lpt")
		{
			print "Line printer reset\n";
			reset_lpt();
		}	
		elsif($w[1] eq "crd")
		{
			print "Card reader reset\n";
			reset_crd();
		}
		elsif($w[1] eq "md0" || $w[1] eq "md1")
		{
			print "Moving head disk reset\n";
			reset_md0();
		}		
		elsif($w[1] eq "mem")
		{
			print "Memory reset\n";
			@M=();
		}
		elsif($w[1] eq "power")
		{
		    reset_all();
			@M=();
			$R[0]=0;
			print "Power reset\n"; 
		}
		else
		{
			print "Reset All, except A0/P and Memory\n"; 
			reset_all();
		}
	}
    elsif($w[0] eq "trace" || $w[0] eq "tr")
	{
		if($w[1] eq "on")
		{
			if($w[2] eq "")
			{
				$EmTrace = 1;
				print "set trace $w[1]\n";
			}
			else
			{
				if($tracefile ne "")
				{
					close TRACE || die "can't close $tracefile: $!";
					print "Previous tracefile $tracefile is closed\n";
				}
				$tracefile = $w[2];
				if(! open (TRACE, ">$tracefile"))
				{
					print "Cannot open the new tracefile: $tracefile\n";
				}
				else
				{
					$EmTrace = 1;
					print "set trace $w[1], trace dumped in $tracefile\n";
				}					
			}
		}
		elsif($w[1] eq "off")
		{
			$EmTrace = 0;
			print "set trace $w[1]\n";
		}
		else
		{
			$EmTrace = 1;
			print "set trace on\n";
		}		
	}
	elsif($w[0] eq "load")
	{
		my $leader = "false";
		my $c;
		my $w;
		my $err=0;
		my $ba=0;     # begin address of load area
		my $wc=0;     # word count
		my $l;
		my $mode = ""; 

		$file = $w[1];
		if ($file eq "" && ( $file !~ /\.abs/ ) && ( $file !~ /\.rel/ ) && ( $file !~ /\.LM/ ))
		{
			print "No filename, or wrong extension (must be .abs or .rel/.LM)\n";
			$err=1;
		}	
				
		if($w[1] =~ /\.abs/)
		{
			$mode = 'a';
		}
		
		if($w[1] =~ /\.rel/ || $w[1] =~ /\.LM/)
		{
			if($w[2] ne "" && $w[2] =~ /^[0-9a-f]/)
			{
				# load a relocatable load module
				$ba = cvhex(UP($w[2]));
				$ba = $ba & 0xFFFE;	
				$mode = 'r';
			}
			elsif($w[2] ne "" && $w[2] eq "mon")
			{
				# load a relocatable monitor load module at address 0
				$ba = -8;
				$mode = 'm';				
			}			
			else
			{
				print "Load address missing\n";
				$err = 1;
			}			
		}

		if($err == 0)
		{
			if(! open (IN, $file))
			{
				print "Unable to open requested load file: $file\n";
				$err = 1;
			}
		}
		
		if($err == 0 && $mode eq 'a')
		{
			binmode IN, ":raw";			
			print "Load $w[1]\n";
			# Read bootable 8+8 formatted paper tape
			# skip leader
			while( 1 )
			{
				$c = nextchar();
				# skip leader in case there
				if ($leader eq "false")
				{
					next if $c eq $null;
					$leader = "true";
					print "Leader skipped\n";
					last;
				}
			}
			$wc = nextword();                       # word count
			$ba = (nextword() + 2) & 0xFFFE;        # begin address of load area
			my $low = $ba;
			my $high = $ba + $wc*2;
			my $chksum = 0;
			print "Load " . prhex(16,$wc) . " words\; start loading at" . prhex(16,$ba) . "\n";
			$l = 0;
			while($l < $wc)
			{
				$w = nextword();
				$chksum = $chksum ^ $w;
				$M[$ba] = $w;
				$ba = $ba + 2;
				$l = $l + 1;
			}	
			my $chk = nextword();	
			if( ($chksum - $chk) != 0)
			{
				print "Checksum error\n";
			}
			$c =nextchar();
			if($c != $xoff)
			{
				print "No x-off marking end of tape\n";
			}
			close (IN) || die "can't close $file: $!";
			print "LOADING COMPLETE\; 1st free address is" . prhex(16,$high) . "\n";
			$R[0] = $low;		# default start address
			print "Default start address (P) is:" . prhex(16,$low) . "\n";
		}
		
		if($err == 0 && $mode eq 'r')
		{		
			# load relocatable load module
			binmode IN, ":raw";			
			print "Load $w[1]\n";
			nextword(); # header word 1
			nextword(); # header word 2
            my $sa = nextword();    # start address
			my $ns = nextword();    # number of sectors
			$wc    = nextword();    # length in words, counted in bytes
			$wc    = $wc >> 1;      # length in words
			nextword();             # sym table pointer for debugging: not used at this moment
			print "Load " . prhex(16,$wc) . " words\; start loading at" . prhex(16,$ba) . "\n";
            # $ba is the relocator
		    $M[$ba]   = $sa;
			$M[$ba+2] = $ns;
			$M[$ba+4] = $wc << 1;
			$M[$ba+6] = 0;          # sym table pointer for debugging: not used at this moment
				
			my $k = 0;              # total number of words relocated
			$l=4;					# total number of words loaded
			
            my $i=4;				# number of code/data words retrieved from record
			while($l<$wc)
			{
				if($i == 0)
				{
					# skip header
			        nextword(); # header word 1
			        nextword(); # header word 2
                }					
				# load record
				while(($i<188) && ($l<$wc))
				{
					$w=nextword();
					$M[$ba+($l*2)] = $w;
					$i=$i+1;
					$l=$l+1;
				}
				# skip to relocation table, if not there yet
				while($i<188)
				{
					nextword();
					$i=$i+1;
				}
				# relocate this record
				$i=0;
				while($i<12)    # scan the relocation table
				{
					my $rw = nextword();
					my $j = 0;
					while(($j<16) && ($k<$l))
					{
						if($rw & 0x8000)
						{
							$M[$ba+($k*2)] = ($M[$ba+($k*2)] + $ba) & 0xFFFF ; # relocate
						}
						$rw = $rw << 1;
						$j=$j+1;
						$k=$k+1;
					}
					$i=$i+1;
				}
				# skip to end of record
				nextword(); # trailer word 1
				nextword(); # trailer word 1
				nextword(); # trailer word 1
				$i=0;
			}
			close (IN) || die "can't close $file: $!";
			$i=$ba+($l*2);      # first free address 
			print "LOADING COMPLETE\; 1st free address is" . prhex(16,$i) . "\n";
			$R[0] = $M[$ba];		# start address
			print "Start address (P) is: " . prhex(16,$R[0]) . "\n";			
		}
		if($err == 0 && $mode eq 'm')
		{		
			# load relocatable monitor load module at address 0
			binmode IN, ":raw";			
			print "Load monitor $w[1]\n";
			nextword(); # header word 1
			nextword(); # header word 2
            my $sa = nextword()-8;  # start address (relocated)
			my $ns = nextword();    # number of sectors
			$wc    = nextword();    # length in words, counted in bytes
			$wc    = $wc >> 1;      # length in words
			nextword();             # sym table pointer for debugging: not used at this moment
			print "Load " . prhex(16,$wc) . " -4 words\; start loading at" . prhex(16,0) . "\n";
            # $ba is the relocator (-8 !)
			my $sec0=1; # sector 0 indicator
			my $low=0;
			my $high=0;	
			my $k = 4;              # total number of words relocated
			$l=4;					# total number of words loaded
            my $i=4;				# number of code/data words retrieved from record
			while($l<$wc)
			{
				if($i == 0)
				{
					# skip header
			        nextword(); # header word 1
			        nextword(); # header word 2
                }					
				# load record
				while(($i<188) && ($l<$wc))
				{
					$w=nextword();
					$M[$ba+($l*2)] = $w;
					$i=$i+1;
					$l=$l+1;
				}
				# skip to relocation table, if not there yet
				while($i<188)
				{
					nextword();
					$i=$i+1;
				}
				# relocate this record
				$i=0;
				while($i<12)    # scan the relocation table
				{
					my $rw = nextword();
					my $j = 0;
					if($sec0)
					{
						# skip 1st 4 words of sector 0
						$rw = $rw << 4;
						$j  = 4;
						$sec0=0;
					}					
					while(($j<16) && ($k<$l))
					{
						if($rw & 0x8000)
						{
							$M[$ba+($k*2)] = ($M[$ba+($k*2)] + $ba) & 0xFFFF ; # relocate
						}
						$rw = $rw << 1;
						$j=$j+1;
						$k=$k+1;
					}
					$i=$i+1;
				}
				# skip to end of record
				nextword(); # trailer word 1
				nextword(); # trailer word 1
				nextword(); # trailer word 1
				$i=0;
			}
			close (IN) || die "can't close $file: $!";
			$i=$ba+($l*2);      # first free address 
			print "LOADING COMPLETE\; 1st free address is" . prhex(16,$i) . "\n";
			$high=$ba+($l*2)-2;
			$i = "Used memory area is: low=" . prhex(16,$low) . " high=" . prhex(16,$high);
			print "$i\n";					
			$R[0] = $sa;		# start address
			print "Start address (P) is:" . prhex(16,$R[0]) . "\n";			
		}				
	}
	elsif($w[0] eq "boot")
	{
		if($w[1] eq "ptr")
		{
			if($PTR[FILE] eq "")
			{
				print "No file attached to paper tape reader\n";
			}
			else
			{
				reset_all();
				@M=();
				# put papertape 8+8 bootstrap in memory
				#                 * CODE   EQU   *+/2E
				#                   PR     EQU   /20
				#                   S      EQU   1
				$M[0]  = 0x20BF;  #        INH
				$M[2]  = 0x41E0;  #        CIO   A1,S,PR
				$M[4]  = 0x4A20;  # SEARCH INR   A2,0,PR
				$M[6]  = 0x5C04;  #        RB(4) *-2
				$M[8]  = 0x227F;  #        ANK   A2,/7F
				$M[10] = 0x5808;  #        RB(0) SEARCH              skip leader
				$M[12] = 0x0704;  #        LDK   A7,4                read 4 bytes:
				$M[14] = 0x0540;  #        LDK   A5,/40              word count, and
				#                   INPUT  EQU   *
				$M[16] = 0x4A20;  #        INR   A2,0,PR             begin address - 2
				$M[18] = 0x5C04;  #        RB(4) *-2
				$M[20] = 0xE235;  #        SCR   A2,A5
				$M[22] = 0x1501;  #        ADK   A5,1
				$M[24] = 0x1F01;  #        SUK   A7,1
				$M[26] = 0x590C;  #        RB(1) INPUT               > 0
				$M[28] = 0x5210;  #        RF(2) CODE                < 0
				$M[30] = 0xBAC0;  #        ML    5,/3A               = 0
				$M[32] = 0x003A;  #
				$M[34] = 0x1502;  #        ADK   A5,2
				$M[36] = 0x9508;  #        ADR   A5,A2
				$M[38] = 0x3C41;  #        SLL   A4,1
				$M[40] = 0x8614;  #        LDR   A6,A5               A6 default start address
				$M[42] = 0x0100;  #        LDK   A1,0
				$M[44] = 0x5F1E;  #        RB    INPUT
				#                   CODE   EQU   *
				$M[46] = 0xE104;  #        ECR   A1,A1
				$M[48] = 0xB108;  #        XRR   A1,A2
				$M[50] = 0x1C01;  #        SUK   A4,1
				$M[52] = 0x5C26;  #        RB(4) INPUT               ne 0
				$M[54] = 0x8F18;  #        ABR   A6                  jump to start address (or HLT)           
				$M[56] = 0x0000;  #        DATA  0 UNUSED
				$M[58] = 0x0000;  # (3A)   DATA  0 UNUSED            A1
				$M[60] = 0x0000;  #        DATA  0 BASE ADDRESS      A2
				$M[62] = 0x0000;  #        DATA  0 UNUSED            A3
				########################################
				$M[64] = 0x0000;  # (40)   DATA  0 WORD COUNT        A4
				$M[66] = 0x0000;  #        DATA  0 BEGIN ADDRESS -2  A5
				#
				if($w[2] eq "h")
				{
					$M[54] = 0x207F;               # HALT, after loading complete
				}
				if($w[2] eq "r")
				{
					$M[54] = 0x8F18;               # ABR A6 => RUN starting at begin 1st address
				}
				$R[0] = 0;                         # PC = 0, start at address 0
				$interval = 0;
				$currentinterval = 0;
				$time = 0;
				$event_queue = 0;			
				activate_event(\@TTY,$TTY[TIME]); 
				@t0 = gettimeofday ();             # fix t0 to measure elapsed time
				$RUN = 1;
				ReadMode 4;                        # prepare simulated console mode
			  STDOUT->autoflush(1);              # for MSWin32				
				while($RUN)
				{
					instr();
				}
				ReadMode 0;                        # reset console
			  STDOUT->autoflush(0);              # for MSWin32				
				$secs = tv_interval (\@t0);        # elapsed run time in secs
				$ctime = $time + ($currentinterval - $interval);   # accumulated runned cycles
				$cps = int($ctime / $secs);        # cycles per second
				$s = prhex(16,$R[0]);
				$program_started = 1;              # program is still in a running state
				print "\nRun/Boot: emulation stopped at address $s\n";
				print "$ctime instructions in $secs seconds: $cps instructions per second\n";				
			}
		}
		elsif($w[1] eq "md0")
		{
			if($MD0[FILE] eq "")
			{
				print "No disk image file attached to disk drive\n";
			}
			else
			{
				reset_all();
				@M=(); # reset memory
				# put disk bootstrap for the X1210 connected to the multiplex channel in memory
				# THE BOOTSTRAP loads logical sector 1 (= physical sector 3) as IPL, which
				# loads and starts the monitor
				#                   S      EQU   1
				$M[0]  = 0x20BF;  #00 INH    INH
				$M[2]  = 0xBBC0;  #02        ML   7,IOROUT
				$M[4]  = 0x0032;  #04 
				$M[6]  = 0x273F;  #06        ANK  A7,/3F        ADDRESS OF DISK UNIT
				$M[8]  = 0x911C;  #08        ADR  A1,A7         UPDATE CIO
				$M[10] = 0x921C;  #OA        ADR  A2,A7         UPDATE SST
				$M[12] = 0x270F;  #0C        ANK  A7,/F         CALCULATE MPX
				$M[14] = 0x3F42;  #0E        SSL  A7,2          CHANNEL ADDR USED BY DISK
				$M[16] = 0x1778;  #10        ADK  A7,/78        /78 = /80 - 4 WORDS
				$M[18] = 0xBB3D;  #12        MSR  6,A7          IO ROUTINE + MPX SET
				$M[20] = 0x0503;  #14        LDK  A5,3          FUNTION = SEEK 0
				$M[22] = 0x041A;  #16        LDK  A4,/1A        RETURN ADDR = BOOT1
				$M[24] = 0x8F1C;  #18        ABR  A7            CALL IO ROUTINE
				$M[26] = 0x050C;  #1A BOOT1  LDK  A5,/C         FUNCTION = READ SECTOR 3 (LOGICAL 1)
				$M[28] = 0x0420;  #1C        LDK  A4,/20        RETURN ADDR = BOOT2
				$M[30] = 0x8F1C;  #1E        ABR  A7            CALL IO ROUTINE
				$M[32] = 0x8340;  #20 BOOT2  LD   A3,CW2        CALCULATE START ADDRESS
				$M[34] = 0x003C;  #22        
				$M[36] = 0x9B20;  #24        SUKL A3,/198
				$M[38] = 0x0198;  #26
				$M[40] = 0x812C;  #28        LDR* A1,A3         SECTOR IDENT IN PLACE? (0)
				# A1 gets the sector identifier:
				# the instruction below in the original bootstrap is an RB(4) INH.
				# Because in this implementation (for the time being) the sector
				# identifier is extended with the physical sector number in the track
				# the sector identifier of physical sector 3 is now NE 0 !!
				# the RB(4) is changed therefore into an RB(0)
				#$M[42] = 0x582C;  #2A        RB(0) INH          NO ***5C2C RB(4) INH*** (SEC_IDENT=TEST)
				$M[42] = 0x5C2C;  #2A        RB(4) INH          NO (SEC_IDENT=P800)
				$M[44] = 0x1304;  #2C        ADK  A3,4          A3 = START ADDRESS
				$M[46] = 0x8F0C;  #2E        ABR  A3            JUMP TO IPL
				$M[48] = 0x0000;  #30        DATA 0     
				$M[50] = 0x45C0;  #32 IOROUT CIO  A5,S,0   A1   IO ROUTINE TO BE TRANSFERRED
				$M[52] = 0x4DC0;  #34        SST  A5,0     A2   BEFORE THE MPX CONTROL WORDS
				$M[54] = 0x5C04;  #36        RB(4) -*2     A3    
				$M[56] = 0x8F10;  #38        ABR  A4       A4   RETURN TO CALLER
				$M[58] = 0x4E66;  #3A CW1    DATA /4E66    A5   MPX CW1 (INPUT 410 BYTES IN WORD MODE)
				$M[60] = 0x7E00;  #3C CW2    DATA /7E00    A6   MPX CW2 BUFFER END 16KW (3E00 FOR 8KW)
				$M[62] = 0x0002;  #3E        DATA 2        A7   DISK DEVICE ADDRESS (2 boots md0, /12 boots md1)
                # DEVICE ADDRESS IN /3E IS AN EXAMPLE: WITH /8002 THE PAPERTAPE BOOTSTRAP VIA DISK IPL
                # IS REQUESTED (IF SUPPORTED BY IPL)
				#
				if($w[2] eq 'p')
				{
					# papertape bootstrap via disk ipl is requested
					# ptr attached??
					if($PTR[FILE] eq "")
					{
						print "The paper tape bootstrap via disk IPL is requested\n";
						print "But no file is attached to paper tape reader\n";
					}
					else
					{
						$M[62] = 0x8002;                   # set papertape bootstrap request bit
                        # default is to continue after boot
						if($w[3] eq "h")
						{
							$M[56] = 0x207F;               # HALT, after loading complete
							print "Boot stops after loading IPL\n";
						}
						if($w[3] eq "r")
						{
							$M[56] = 0x8F10;               # ABR A4 => RUN starting at begin 1st address of IPL
							print "Boot starts IPL after loading IPL\n";			
						}
					    $R[0] = 0;                         # PC = 0, start at address 0
					    $interval = 0;
					    $currentinterval = 0;
					    $time = 0;
					    $event_queue = 0;
					    activate_event(\@TTY,$TTY[TIME]); 
					    @t0 = gettimeofday ();             # fix t0 to measure elapsed time
					    $RUN = 1;
					    ReadMode 4;                        # prepare simulated console mode
			        STDOUT->autoflush(1);              # for MSWin32
					    while($RUN)
					    {
							instr();
						}
						ReadMode 0;                        # reset console
			      STDOUT->autoflush(0);              # for MSWin32						
						$secs = tv_interval (\@t0);        # elapsed run time in secs
						$ctime = $time + ($currentinterval - $interval);   # accumulated runned cycles
						$cps = int($ctime / $secs);         # cycles per second
						$s = prhex(16,$R[0]);
						$program_started = 1;              # program is still in a running state
						print "\nRun/Boot: emulation stopped at address $s\n";
						print "$ctime instructions in $secs seconds: $cps instructions per second\n";
					}						
				}
				else
				{
					# default is to continue after boot					
					if($w[2] eq "h")
					{
						$M[56] = 0x207F;               # HALT, after loading complete
						print "Boot stops after loading IPL\n";
					}
					if($w[2] eq "r")
					{
						$M[56] = 0x8F10;               # ABR A4 => RUN starting at begin 1st address of IPL
						print "Boot starts IPL after loading IPL\n";			
					}						
					$R[0] = 0;                         # PC = 0, start at address 0
					$interval = 0;
					$currentinterval = 0;
					$time = 0;
					$event_queue = 0;
					activate_event(\@TTY,$TTY[TIME]); 
					@t0 = gettimeofday ();             # fix t0 to measure elapsed time
					$RUN = 1;
					ReadMode 4;                        # prepare simulated console mode
			    STDOUT->autoflush(1);              # for MSWin32
					while($RUN)
					{
						instr();
					}
					ReadMode 0;                        # reset console
			    STDOUT->autoflush(0);              # for MSWin32					
					$secs = tv_interval (\@t0);        # elapsed run time in secs
					$ctime = $time + ($currentinterval - $interval);   # accumulated runned cycles
					$cps = int($ctime / $secs);         # cycles per second
					$s = prhex(16,$R[0]);
					$program_started = 1;              # program is still in a running state
					print "\nRun/Boot: emulation stopped at address $s\n";
					print "$ctime instructions in $secs seconds: $cps instructions per second\n";
				}
			}
		}		
		else
		{
			print "Unsupported boot device $w[1]\n";
		}
	}
	elsif($w[0] eq "attach" || $w[0] eq "att")
	{
		if($w[1] eq "ptr")
		{
			$file = $w[2];
			if($file eq "")
			{
				print "No proper filename\n";
			}
			else
			{
				if($PTR[FILE] ne "")
				{
					close ($PTR[FC]) || die "can't close $PTR[FILE]: $!";
					print "File $PTR[FILE] detached from device PTR and closed\n";
					$PTR[FILE] = "";         # file detached and closed; device inoperable	
				}
				if(! open (PTR, $file))
				{
					print "Cannot open papertape input file: $file\n";
				}
				else
				{
					$PTR[FC] = \*PTR;
					binmode $PTR[FC], ":raw";
					print "File $file attached to device PTR\n";
					$PTR[FILE] = $file;           # file attached and open
				}
			}
		}
		elsif($w[1] eq "crd")
		{
			$file = $w[2];
			my($fi,$fy) = split(/\./, $file);
			if($file eq "" || $fy ne "crd")
			{
				print "No filename, or wrong extension (must be a .crd file)\n";
			}
			else
			{
				if($CRD[FILE] ne "")
				{
					close ($CRD[FC]) || die "can't close $CRD[FILE]: $!";
					print "File $CRD[FILE] detached from device CRD and closed\n";
					$CRD[FILE] = "";         # file detached and closed; device inoperable	
				}
				if(! open (CRD, $file))
				{
					print "Cannot open card reader input file: $file\n";
				}
				else
				{
					$CRD[FC] = \*CRD;
					binmode $CRD[FC], ":raw";
					print "File $file attached to device CRD\n";
					$CRD[FILE] = $file;           # file attached and open
				}
			}
		}		
		elsif($w[1] eq "ptp")
		{
			$file = $w[2];
			if($file eq "")
			{
				print "No proper filename\n";
			}
			else
			{
				if($PTP[FILE] ne "")
				{
					close ($PTP[FC]) || die "can't close ($PTP[FILE]: $!";
					print "File $PTP[FILE] detached from device PTP and closed\n";
					$PTP[FILE] = "";         # file detached and closed; device inoperable	
				}				
				if(! open (PTP, ">:raw", $file))
				{
					print "Cannot open papertape punch file: $file\n";
				}
				else
				{
					$PTP[FC] = \*PTP;
					print "File $file attached to device PTP\n";
					$PTP[FILE] = $file;		      # file attached and open
				}
			}				
		}
		elsif($w[1] eq "lpt")
		{
			$file = $w[2];
			if($file eq "")
			{
				print "No proper filename\n";
			}
			else
			{
				if($LPT[FILE] ne "")
				{
					close ($LPT[FC]) || die "can't close ($LPT[FILE]: $!";
					print "File $LPT[FILE] detached from device LPT and closed\n";
					$LPT[FILE] = "";         # file detached and closed; device inoperable	
				}				
				if(! open (LPT, ">$file"))
				{
					print "Cannot open lineprinter file: $file\n";
				}
				else
				{
					$LPT[FC] = \*LPT;
					print "File $file attached to device LPT\n";
					$LPT[FILE] = $file;		      # file attached and open
					$LPT[LCUR]  = 0;              # set on line 0 current page 	
				}
			}					
		}
		elsif($w[1] eq "md0")
		{
			$file = $w[2];
			my($fi,$fy) = split(/\./, $file);
			if($file eq "" || $fy ne "img")
			{
				print "No filename, or wrong extension (must be a .img file)\n";
			}
			else
			{
				if($MD0[FILE] ne "")
				{
					close (MD0) || die "can't close $MD0[FILE]: $!";
					print "File $MD0[FILE] detached from device MD0 and closed\n";
					$MD0[FILE] = "";         # file detached and closed; device inoperable	
				}
				if(! open (MD0, "+<:raw", $file)) # open for update
				{
					print "Cannot open disk file: $file\n";
				}
				else
				{
					print "File $file attached to device MD0\n";
					$MD0[FILE] = $file;           # file attached and open
					$MD0[CYL]  = 0;
				}
			}
		}			
		elsif($w[1] eq "md1")
		{
			$file = $w[2];
			my($fi,$fy) = split(/\./, $file);
			if($file eq "" || $fy ne "img")
			{
				print "No filename, or wrong extension (must be a .img file)\n";
			}
			else
			{
				if($MD1[FILE] ne "")
				{
					close (MD1) || die "can't close $MD1[FILE]: $!";
					print "File $MD1[FILE] detached from device MD1 and closed\n";
					$MD1[FILE] = "";         # file detached and closed; device inoperable	
				}
				if(! open (MD1,"+<:raw", $file)) # open for update
				{
					print "Cannot open disk file: $file\n";
				}
				else
				{
					print "File $file attached to device MD1\n";
					$MD1[FILE] = $file;           # file attached and open
					$MD1[CYL]  = 0;
				}
			}
		}					
		else
		{
			print "Unsupported device $w[1] requested for file attachement\n";
		}
	}
	elsif($w[0] eq "detach" || $w[0] eq "det")
	{
		if($w[1] eq "ptr")
		{
			$file = $PTR[FILE];
			if($file eq "")
			{
				print "Device PTR not attached\n";
			}
			else
			{
				close ($PTR[FC]) || die "can't close $file: $!";
				print "File $file detached from device PTR and closed\n";
				$PTR[FILE] = "";         # file detached and closed; device inoperable	
			}
		}
		if($w[1] eq "cdr")
		{
			$file = $CRD[FILE];
			if($file eq "")
			{
				print "Device CRD not attached\n";
			}
			else
			{
				close ($CRD[FC]) || die "can't close $file: $!";
				print "File $file detached from device CRD and closed\n";
				$CRD[FILE] = "";         # file detached and closed; device inoperable	
			}
		}		
		elsif($w[1] eq "ptp")
		{
			$file = $PTP[FILE];
			if($file eq "")
			{
				print "Device PTP not attached\n";
			}
			else
			{
				close ($PTP[FC]) || die "can't close $file: $!";
				print "File $file detached from device PTP and closed\n";			
				$PTP[FILE] = "";			# file detached and closed; device inoperable		
			}				
		}
		elsif($w[1] eq "lpt")
		{
			$file = $LPT[FILE];
			if($file eq "")
			{
				print "Device LPT not attached\n";
			}
			else
			{
				close ($LPT[FC]) || die "can't close $file: $!";
				print "File $file detached from device LPT and closed\n";			
				$LPT[FILE] = "";			# file detached and closed; device inoperable		
			}				
		}
		elsif($w[1] eq "md0")
		{
			$file = $MD0[FILE];
			if($file eq "")
			{
				print "Device MD0 not attached\n";
			}
			else
			{
				close (MD0) || die "can't close $file: $!";
				print "File $file detached from device MD0 and closed\n";			
				$MD0[FILE] = "";			# file detached and closed; device inoperable		
			}				
		}
		elsif($w[1] eq "md1")
		{
			$file = $MD1[FILE];
			if($file eq "")
			{
				print "Device MD1 not attached\n";
			}
			else
			{
				close (MD1) || die "can't close $file: $!";
				print "File $file detached from device MD1 and closed\n";			
				$MD1[FILE] = "";			# file detached and closed; device inoperable		
			}				
		}		
		else
		{
			print "Unsupported device $w[1]\n";
		}
	}	
	elsif($w[0] eq "state")
	{
		if($w[1] eq "cpu")
		{
			print "CPU state\n";
			trace($R[0]);
		}
		elsif($w[1] eq "tty")
		{
			print "TTY state: ";
			state_tty();
		}
		elsif($w[1] eq "rtc")
		{
			print "RTC state: ";
			state_clk();
		}	
		elsif($w[1] eq "ptr")
		{
			print "PTR state: ";
			state_ptr();
		}
		elsif($w[1] eq "ptp")
		{
			print "PTP state: ";
			state_ptp();
		}
		elsif($w[1] eq "lpt")
		{
			print "LPT state: ";
			state_lpt();
		}	
		elsif($w[1] eq "crd")
		{
			print "CRD state: ";
			state_crd();
		}
		elsif($w[1] eq "md0" || $w[1] eq "md1")
		{
			print "MD* state: ";
			state_md0();
		}		
		else
		{
			print "CPU state\n";
			trace($R[0]);
			print "TTY state: ";
			state_tty();
			print "RTC state: ";
			state_clk();
			print "PTR state: ";
			state_ptr();
			print "PTP state: ";
			state_ptp();
   			print "LPT state: ";
			state_lpt();
   			print "CRD state: ";
			state_crd();
			print "MD* state: ";
			state_md0();            			
		}
	}
	elsif($cmd eq "queue" || $cmd eq "q")
	{
		show_event_queue();
	}
	elsif($w[0] eq "echo")
	{	
		print eval($w[1]);
	}
	elsif($w[0] eq "console")
	{
		if($w[1] eq "64")
		{
			$TTY[CSET] = 64;
		}
		elsif($w[1] eq "128")	
		{
			$TTY[CSET] = 128;	
		}	
		else
		{
			print "Wrong console character set selected, set to 64\n";
			$TTY[CSET] = 64;
		}
	}
	elsif($cmd eq "exit")
	{
		# close still open/attached files
		if($PTR[FILE] ne "")
		{
			close ($PTR[FC]) || die "can't close $PTR[FILE]: $!";
			print "File $PTR[FILE] detached from device PTR and closed\n";	
		}
		if($PTP[FILE] ne "")
		{
			close ($PTP[FC]) || die "can't close $PTP[FILE]: $!";
			print "File $PTP[FILE] detached from device PTP and closed\n"
		}
		if($LPT[FILE] ne "")
		{
			close ($LPT[FC]) || die "can't close $LPT[FILE]: $!";
			print "File $LPT[FILE] detached from device LPT and closed\n"
		}
		if($CRD[FILE] ne "")
		{
			close ($CRD[FC]) || die "can't close $CRD[FILE]: $!";
			print "File $CRD[FILE] detached from device CRD and closed\n"
		}
		if($MD0[FILE] ne "")
		{
			close (MD0) || die "can't close $MD0[FILE]: $!";
			print "File $MD0[FILE] detached from device MD0 and closed\n"
		}
		if($MD1[FILE] ne "")
		{
			close (MD1) || die "can't close $MD1[FILE]: $!";
			print "File $MD1[FILE] detached from device MD1 and closed\n"
		}		
		if($do ne "")
		{
			close DO || die "can't close $do: $!";
		}
		if($tracefile ne "")
		{
			close TRACE || die "can't close $tracefile: $!";
			print "Tracefile $tracefile closed\n"			
		}		
		print "EXIT!\n";
		exit;
	}
	else
	{
		print "Unknown command\n";
	}
	@w = (); # reset
}

############################## Emulator #################################
# execute a single instruction at address $R[0]
#
sub instr {
    $interval = $interval - 1;
	if($interval <= 0)
	{
		process_events();
	}
	if($brkreg)
	{
		# Break
		$t1 = $brkreg;
		$t2 = 1;
		while(! ($t1 & 1))
		{
			$t1 = $t1 >> 1;
			$t2 = $t2 + 1;
		}
		# $t2 = mpx channel number requesting a break
		# the control unit must be in EXCHANGE state
		$t1 = 1 << ($t2-1);
		$brkreg = $brkreg & (~ $t1);    # reset break request
		$t1 = $DEV{$t2};                # cu requesting the break
		$t3 = $$t1[STATE] >> 16;        # cu state
		# print "MPX DEV " . prhex(8,$t2) . "\n";
		$a  = $MPX[$t2];                # memory address of mpx channel words		
		$mpx1 = $M[$a];	             	# w1: function and position
		$mpx2 = $M[$a+2];               # w2: buffer end (last byte/word to be read/written)
		$tx3 = ((~ $mpx1) + 1) & 0xFFF; # current position in buffer (last posion being read/written) 
		# mpx increments memory address BEFORE memory access
		$tx4 = $mpx2 - $tx3 + 1;        # new byte position in memory to address
		# print "MPX w1 w2 x3 x4" . prhex(16,$mpx1) . prhex(16,$mpx2) . prhex(16,$tx3) . prhex(16,$tx4) . "\n";
		$xf  = $mpx1 & 0x2000;          # multiplex function bit
	
		if($mpx1 & 0x4000)	# check I-bit
		{	
			# I-bit is 1 => INPUT
			if($t3 == IEXC)
			{
				# indeed the CU is in inputmode
				if($tx3 == 0)
				{
					# end of range; transfer ready
					# switch control unit to the wait request state
					$$t1[STATE] = $$t1[STATE] & 0xFFFF;           # reset cu state bits
					$$t1[STATE] = $$t1[STATE] | (WREQ << 16);    									
				}
				else
				{
					$t3 = $$t1[INR];            # call input service routine				
					if($mpx1 & 0x8000) #check C-bit
					{
						&$t3($xf,1);	# ask for a byte
						# C-bit is 1 => character transfer
						if($tx4 & 1)
						{
							# righthand byte
							$M[$tx4 & 0xFFFE] = ($M[$tx4 & 0xFFFE] & 0xFF00) | ($$t1[BUFFER] & 0xFF);
						}
						else
						{
							# lefthand byte
							$M[$tx4 & 0xFFFE] = ($M[$tx4 & 0xFFFE] & 0xFF) | (($$t1[BUFFER] & 0xFF) << 8);
						}
						$mpx1 = $mpx1 + 1;		# update position
					}
					else
					{
						&$t3($xf,0);	# ask for a word
						# C-bit is 0 => word transfer
						$tx4=$tx4+1;            # new word position in memory to address
						$M[$tx4 & 0xFFFE] = $$t1[BUFFER] & 0xFFFF;
						$mpx1 = $mpx1 + 2;		# update position
					}
					$$t1[BUFFER] = 0;           # reset buffer
					$M[$a] = $mpx1;		        # update MPX word 1 (position)
				}
			}
			else
			{
				# control unit state error	
				print "Error Input Multiplex channel, DEV:" . prhex(8,$t2) . " STATE:" . prhex(20,$t3) . "\n"; 
				mpxerror();
			}
		}
		else
		{
			# I-bit is 0 => OUTPUT
			if($t3 == OEXC)
			{
				# indeed the CU is in outputmode
				if($tx3 == 0)
				{
					# end of range; transfer ready
					# switch control unit to the execute state
					$$t1[STATE] = $$t1[STATE] & 0xFFFF;           # reset cu state bits
					$$t1[STATE] = $$t1[STATE] | (OEXE << 16);     # state = output + execute										
				}				
				else
				{
					$t3 = $$t1[OTR];            # call output service routine					
					if($mpx1 & 0x8000) #check C-bit
					{
						# C-bit is 1 => character transfer
						if($tx4 & 1)
						{
							$$t1[BUFFER] = $M[$tx4 & 0xFFFE] & 0xFF;        # righthand byte
						}
						else
						{
							$$t1[BUFFER] = ($M[$tx4 & 0xFFFE] >> 8) & 0xFF; # lefthand byte
						}
						$mpx1 = $mpx1 + 1;		# update position
						&$t3($xf,1);            # transfer a byte						
					}
					else
					{
						# C-bit is 0 => word transfer
						$tx4=$tx4+1;            # word position in memory						
						$$t1[BUFFER] = $M[$tx4 & 0xFFFE];	
						$mpx1 = $mpx1 + 2;		# update position
						&$t3($xf,0);            # transfer a word
					}
					$M[$a] = $mpx1;		        # update MPX word 1 (position)
				}
			}
			else
			{
				# control unit state error		
				print "Error Output Multiplex channel, DEV:" . prhex(8,$t2) . " STATE:" . prhex(20,$t3) . "\n"; 		
				mpxerror();	
			}
		}
	}
    if($ENB)
	{
		# interrupt?
		if($intreg) # check (for speed) unmasked, so it still can be a masked interrupt
		{
			$t1 = $intreg & 0xFFFF0000;                 # priority interrupts
			$t2 = ($intreg & (~ $intmsk)) & 0xFFFF;     # common line interrupts
			# an interrupt
			# check according priority chain
			# checks for levels 0,3,4 and 6 are not included !!!
			if(($t1 & LKM) && ($PL >= 1))
			{
				# LKM, stackoverflow or illegal instruction (prio/level 1)
			    $ENB = 0;                               # inhibit
			    # P, PSW on stack
			    $M[$R[A15]] = $R[A0];                   # save P
			    $R[A15] = $R[A15] - 2;                  # decrease link stack
			    if($R[A15] < 0x100)
			    {
					trap(LKM);		                    # A15 stack overflow
				}				
				$M[$R[A15]] = Fpsw();                   # save PSW
				$R[A15] = $R[A15] - 2;                  # decrease link stack
				if($R[A15] < 0x100)
				{
					trap(LKM);		                    # A15 stack overflow
				}							
				$PL = 1;
				$R[A0] = $M[AdLKM];
			}
			elsif(($t1 & RTC) && ($PL >= 2))
			{
				# real time clock (prio/level 2)
			    $ENB = 0;                               # inhibit
			    # P, PSW on stack
			    $M[$R[A15]] = $R[A0];                   # save P
			    $R[A15] = $R[A15] - 2;                  # decrease link stack
			    if($R[A15] < 0x100)
			    {
					trap(LKM);		                    # A15 stack overflow
				}				
				$M[$R[A15]] = Fpsw();                   # save PSW(PL+CR)
				$R[A15] = $R[A15] - 2;                  # decrease link stack
				if($R[A15] < 0x100)
				{
					trap(LKM);		                    # A15 stack overflow
				}												
				$PL = 2;
				$R[A0] = $M[AdRTC];
			}
			elsif($t2 && ($PL >= 5))
			{
				# unmasked common line interrupt (prio/level 5)
			    $ENB = 0;                               # inhibit
			    # P, PSW on stack
			    $M[$R[A15]] = $R[A0];                   # save P
			    $R[A15] = $R[A15] - 2;                  # decrease link stack
			    if($R[A15] < 0x100)
			    {
					trap(LKM);		                    # A15 stack overflow
				}				
				$M[$R[A15]] = Fpsw();                  # save PSW(PL+CR)
				$R[A15] = $R[A15] - 2;                  # decrease link stack
				if($R[A15] < 0x100)
				{
					trap(LKM);		                    # A15 stack overflow
				}												
				$PL = 5;
				$R[A0] = $M[AdCOM];
			}
			elsif(($t1 & CP) && ($PL >= 7))
			{
				# a panel interrupt (prio/level 7)
			    $ENB = 0;                               # inhibit
			    # P, PSW on stack
			    $M[$R[A15]] = $R[A0];                   # save P
			    $R[A15] = $R[A15] - 2;                  # decrease link stack
			    if($R[A15] < 0x100)
			    {
					trap(LKM);		                    # A15 stack overflow
				}				
				$M[$R[A15]] = Fpsw();                   # save PSW(PL+CR)
				$R[A15] = $R[A15] - 2;                  # decrease link stack
				if($R[A15] < 0x100)
				{
					trap(LKM);		                    # A15 stack overflow
				}												
				$PL = 7;
			    $R[A0] = $M[AdCP];
			}
		}
		# it was a masked interrupt; continue
	}
	$R[A0] = $R[A0] & 0xFFFE;            # even address and 16 bit
	$IR = $M[$R[A0]];                    # fetch instruction
    $pcsaved = $R[A0];                   # save pc for trace
	$R[A0] = ($R[A0] + 2);               # incr location counter
	if(! ($IR & mskf10))
	{
		# format 0
		# format 0
		# - short constant instructions
		# - shift instructions
		# - i/o instructions
		# - miscellanious instructions
		# execute a format 0 instruction		
		format0();
	}	
	elsif(!($IR & 1))
	{
		# format 1, ls bit 0
		format10();
	}
	else
	{
		# format 1, ls bit 1
		format11();
	}
	
	if($EmBreak)
	{
		# breakpoint?
		if($BP{$pcsaved} ne undef)
		{
			$RUN = 0;
			print "\nBreakpoint, last instruction executed and result:\n";
			if(! $EmTrace)
			{
				trace($pcsaved);
			}
		}
	}
	if($EmTrace)
	{
		# display the result of the executed instruction
		trace($pcsaved);
	}
}

sub format10 {
#
# emulate the format 1 instructions with the ls bit == 0:	

$opc = ($IR >> 11) & 0xF;                   # opcode
$r1  = ($IR >> 7)  & 0xF;                   # r1 or condition shifted 1 left	
$r2  = ($IR >> 1)  & 0xF;                   # r2
$md  = ($IR >> 5)  & 0x3;                   # addressing mode

if($opc == 0)
{
	# load
	if($md == 0)				            # register operand
	{
		# T1 LDR r1,r2
		$R[$r1] = $R[$r2];
	}
	elsif($md == 1)				            # indirect register operand
	{
		# T3A/B LDR* r1,r2, T2 LDKL r1,lk
		if($r2 == 0xF)                      # A15 ?
		{
			# A15, the stack pointer is addressed
			$R[0xF] = ($R[0xF] + 2) & 0xFFFE;
		}
		$R[$r1] = $M[($R[$r2] & 0xFFFE)];
		if($r2 == 0)
		{
			# T2 indirect via A0: LDKL r1,lk
			$R[0] = ($R[0] + 2);         	# increment A0 (PC)
		}
	}
	elsif($md == 2)				            # memory/indexed memory operand
	{
		# T4/T5 LD r1,m or LD r1,m,r2
		$a = $M[$R[0]];			            # memory operand address
		$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
		if($r2 != 0)			            # indexing ?
		{
			$a = ($a + $R[$r2]);            # yes
		}
		$a = $a & 0xFFFE;                   # effective address
		$R[$r1] = $M[$a];                   # load content of effective address
	}
	elsif($md == 3)				            # indirect memory/indexed memory operand
	{
		# T6/T7 LD* r1,m or LD* r1,m,r2
		$a = $M[$R[0]];			            # memory operand address
		$R[0] = ($R[0] + 2);         	    # increment A0 (PC)
		if($r2 != 0)			            # indexing ?
		{
			$a = ($a + $R[$r2]);   # yes
		}
		$a = $M[$a & 0xFFFE] & 0xFFFE;      # memory content = effective address
		$R[$r1] = $M[$a];                   # load content of effective address
	}
	setCR1($r1);
}
elsif($opc == 1)
{
	$r1=$r1>>1;								    # position condition (added in 0.8)
	# absolute branche
    if($md == 0)				                # register operand
	{
		# T1 ABR(cnd) r2
		# $r1 is the condition code in the operation
		if(evalcond($r1))
		{
			$R[0] = $R[$r2];
		}
	}
	elsif($md == 1)				               # indirect register operand
	{
		# T3 ABR(cnd)* r2
		# $r1 is the condition code in the operation
		if(evalcond($r1))
		{
			$R[0] = $M[$R[$r2] & 0xFFFE];      # jump
		}
		elsif($r2 == 0)                        # no jump, skip address in case of an ABL
		{
			# T2 indirect via A0: ABL(cnd) lk
			$R[0] = $R[0] + 2;           	   # increment A0 (PC)
		}
	}
	elsif($md == 2)				               # memory/indexed memory operand
	{	
		# T4/T5 ABI(cnd) m or ABI(cnd) m,r2
		# $r3 is the condition code in the operation
		$a = $M[$R[0]];			               # memory operand address
		$R[0] = ($R[0] + 2);         	       # increment A0 (PC)
		if($r2 != 0)			               # indexing ?
		{
			$a = ($a + $R[$r2]);               # yes
		}
		$a = $a & 0xFFFE;                      # effective address
		if(evalcond($r1))
		{
			$R[0] = $M[$a];                    # load P with content of effective address
		}
	}
	elsif($md == 3)				               # indirect memory/indexed memory operand
	{
		# T6/T7 ABI(cnd)* m or ABI(cnd)* m,r2
		$a = $M[$R[0]];			               # memory operand address
		$R[0] = ($R[0] + 2);         	       # increment A0 (PC)
		if($r2 != 0)			               # indexing ?
		{
			$a = ($a + $R[$r2]);               # yes
		}
		$a = $M[$a & 0xFFFE] & 0xFFFE;         # memory content = effective address
		if(evalcond($r1))
		{
			$R[0] = $M[$a];                    # load P with content of effective address
		}
	}		
}
elsif($opc == 2)
{
	# add, increment memory
	if($md == 0)				            # register operand
	{
		# T1 ADR r1,r2
		$t1 = $R[$r1]; $t2 = $R[$r2]; $t3 = $t1 + $t2;
	}
	elsif($md == 1)				            # indirect register operand
	{
		# T3 ADR* r1,r2, T2 ADKL r1,lk
		$t1 = $R[$r1]; $t2 = $M[($R[$r2] & 0xFFFE)]; $t3 = $t1 + $t2;
		if($r2 == 0)
		{
			# T2 indirect via A0: ADKL r1,lk
			$R[0] = ($R[0] + 2);        	# increment A0 (PC)
		}
	}
	elsif($md == 2)				            # memory/indexed memory operand
	{
		# T4/T5 AD r1,m or AD r1,m,r2
		$a = $M[$R[0]];			            # memory operand address
		$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
		if($r2 != 0)			            # indexing ?
		{
			$a = ($a + $R[$r2]);            # yes
		}
		$a = $a & 0xFFFE;                   # effective address
		$t1 = $R[$r1]; $t2 = $M[$a]; 
		$t3 = $t1 + $t2;                    # add content of effective address
	}
	elsif($md == 3)				            # indirect memory/indexed memory operand
	{
		# T6/T7 AD* r1,m or AD* r1,m,r2
		$a = $M[$R[0]];			            # memory operand address
		$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
		if($r2 != 0)			            # indexing ?
		{
			$a = ($a + $R[$r2]);            # yes
		}
		$a = $M[$a & 0xFFFE] & 0xFFFE;      # memory content = effective address
		$t1 = $R[$r1]; $t2 = $M[$a];        
		$t3 = $t1 + $t2;                    # add content of effective address
	}
	# overflow? (+)
	if((((~$t1) ^ $t2) & ($t1 ^ $t3)) & SIGN)
	{
		$overflow = 1;
	}
	$R[$r1] = setCRvalue($t3);
}
elsif($opc == 3)
{
	# subtract, two complement
	if($md == 0)				            # register operand
	{
		# T1 SUR r1,r2
		$t1 = $R[$r1]; $t2 = $R[$r2]; $t3 = $t1 - $t2;
	}
	elsif($md == 1)				            # indirect register operand
	{
		# T3 SUR* r1,r2, T2 SUKL r1,lk
		$t1 = $R[$r1]; $t2 = $M[($R[$r2] & 0xFFFE)]; $t3 = $t1 - $t2;
		if($r2 == 0)
		{
			# T2 indirect via A0: ADKL r1,lk
			$R[0] = ($R[0] + 2);        	# increment A0 (PC)
		}
	}
	elsif($md == 2)				            # memory/indexed memory operand
	{
		# T4/T5 SU r1,m or SU r1,m,r2
		$a = $M[$R[0]];			            # memory operand address
		$R[0] = ($R[0] + 2);	            # increment A0 (PC)
		if($r2 != 0)			            # indexing ?
		{
			$a = ($a + $R[$r2]);            # yes
		}
		$a = $a & 0xFFFE;                   # effective address
		$t1 = $R[$r1]; $t2 = $M[$a]; 
		$t3 = $t1 - $t2;                    # subtract content of effective address
	}
	elsif($md == 3)				            # indirect memory/indexed memory operand
	{
		# T6/T7 SU* r1,m or SU* r1,m,r2
		$a = $M[$R[0]];			            # memory operand address
		$R[0] = ($R[0] + 2);         	    # increment A0 (PC)
		if($r2 != 0)			            # indexing ?
		{
			$a = ($a + $R[$r2]);            # yes
		}
		$a = $M[$a & 0xFFFE] & 0xFFFE;      # memory content = effective address
		$t1 = $R[$r1]; $t2 = $M[$a];        
		$t3 = $t1 - $t2;                    # subtract content of effective address
	}
	# underflow? (-)
	if((($t1 ^ $t2) & ((~$t2) ^ $t3)) & SIGN)
	{
		$overflow = 1;
	}	
	$R[$r1] = setCRvalue($t3);
}
elsif($opc == 4)
{
	# AND
	if($r1 != 0)
	{
		if($md == 0)				            # register operand
		{
			# T1 ANR r1,r2
			$t3 = $R[$r1] & $R[$r2];
		}
		elsif($md == 1)				            # indirect register operand
		{
			# T3 ANR* r1,r2, T2 ANKL r1,lk
			$t3 = $R[$r1] & $M[($R[$r2] & 0xFFFE)];
			if($r2 == 0)
			{
				# T2 indirect via A0: ADKL r1,lk
				$R[0] = ($R[0] + 2);	        # increment A0 (PC)
			}
		}
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 AN r1,m or AN r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);         	    # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $a & 0xFFFE;                   # effective address
			$t3 = $R[$r1] & $M[$a];             # and content of effective address         
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 AN* r1,m or AN* r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $M[$a & 0xFFFE] & 0xFFFE;      # memory content = effective address
			$t3 = $R[$r1] & $M[$a];             # and content of effective address               
		}
		$R[$r1] = setCRvalue($t3);
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "AND10");
	}			
}
elsif($opc == 5)
{
	# OR
	if($r1 != 0)
	{
		if($md == 0)				            # register operand
		{
			# T1 ORR r1,r2
			$t3 = $R[$r1] | $R[$r2];
		}
		elsif($md == 1)				            # indirect register operand
		{
			# T3 ORR* r1,r2, T2 ORKL r1,lk
			$t3 = $R[$r1] | $M[($R[$r2] & 0xFFFE)];
			if($r2 == 0)
			{
				# T2 indirect via A0: ADKL r1,lk
				$R[0] = ($R[0] + 2);	# increment A0 (PC)
			}
		}
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 OR r1,m or OR r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $a & 0xFFFE;                   # effective address
			$t3 = $R[$r1] | $M[$a];             # or content of effective address         
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 OR* r1,m or OR* r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $M[$a & 0xFFFE] & 0xFFFE;      # memory content = effective address
			$t3 = $R[$r1] | $M[$a];             # or content of effective address               
		}
		$R[$r1] = setCRvalue($t3);
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "OR10");
	}	
}
elsif($opc == 6)
{
	# XOR/TNM
	if($r1 != 0)
	{
		if($md == 0)				            # register operand
		{
			# T1 XRR r1,r2
			$t3 = $R[$r1] ^ $R[$r2];
		}
		elsif($md == 1)				            # indirect register operand
		{
			# T3 XRR* r1,r2, T2 XRKL r1,lk
			$t3 = $R[$r1] ^ $M[($R[$r2] & 0xFFFE)];
			if($r2 == 0)
			{
				# T2 indirect via A0: ADKL r1,lk
				$R[0] = ($R[0] + 2);	        # increment A0 (PC)
			}
		}
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 XR r1,m or XR r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $a & 0xFFFE;                   # effective address
			$t3 = $R[$r1] ^ $M[$a];             # xr content of effective address         
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 XR* r1,m or XR* r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $M[$a & 0xFFFE] & 0xFFFE;      # memory content = effective address
			$t3 = $R[$r1] ^ $M[$a];             # xr content of effective address               
		}
		$R[$r1] = setCRvalue($t3);
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "XOR10");
	}			
		
}
elsif($opc == 7)
{
	# multiple load
    if($md == 1)				                # indirect register operand
	{
		# T3A/B MLR n,r2, T2 MLK n
		# $r1 is n
		$t1 = 1;
		if($r2 == 0xF)
		{
			# A15, the stack pointer is addressed
			$a = $R[0xF] = ($R[0xF] + $r1*2) & 0xFFFE;
			while($t1 <= $r1)
			{
				$R[$A[$t1]] = $M[$a];
				$a = $a - 2;
				$t1 = $t1 + 1;
			}
		}
		else
		{
			$a = $R[$r2] & 0xFFFE;              # effective address
			while($t1 <= $r1)
			{
				$R[$A[$t1]] = $M[$a];
				$a = $a + 2;
				$t1 = $t1 + 1;
			}
		}
		if($r2 == 0)
		{
			# T2 indirect via A0: MLK n ($r1 = n)
			$R[0] = ($R[0] + $r1*2);	       # increment A0 (PC)
		}
	}
	elsif($md == 2)				               # memory/indexed memory operand
	{
		# T4/T5 ML n,m or ML n,m,r2
		$a = $M[$R[0]];			               # memory operand address
		$R[0] = ($R[0] + 2);	               # increment A0 (PC)
		if($r2 != 0)			               # indexing ?
		{
			$a = ($a + $R[$r2]);               # yes
		}
		$a = $a & 0xFFFE;                      # effective address
		# $r1 is n
		$t1 = 1;
		while($t1 <= $r1)
		{
			$R[$A[$t1]] = $M[$a];              # load content of effective address
			$a = $a + 2;
			$t1 = $t1 + 1;
		}			
	}
	elsif($md == 3)				               # indirect memory/indexed memory operand
	{
		# T6/T7 ML* n,m or ML* n,m,r2
		$a = $M[$R[0]];			               # memory operand address
		$R[0] = ($R[0] + 2);	               # increment A0 (PC)
		if($r2 != 0)			               # indexing ?
		{
			$a = ($a + $R[$r2]);               # yes
		}
		$a = $M[$a & 0xFFFE] & 0xFFFE;         # memory content = effective address
		# $r1 is n
		$t1 = 1;
		while($t1 <= $r1)
		{
			$R[$A[$t1]] = $M[$a];              # load content of effective address
			$a = $a + 2;
			$t1 = $t1 + 1;
		}
	}
	else
	{
		# $md == 0
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "ML10");
	}
	$R[2] = setCRvalue($R[2]);	# condition register set by A1	
}
elsif($opc == 8)
{
	# Multiply
	if($r1 == 0)
	{
		my $s = 0;		                               # collect sign result
		$t2 = $R[4];                                   # A2				
		if($md == 0)
		{
			# T1 MUR r2
			$t1 = $R[$r2];                             # operand
		}
		elsif($md == 1)
		{
			if($r2 != 0)
			{
				# T3 MUR* r2
				$t1 = $M[($R[$r2] & 0xFFFE)];          # operand
			}
			else
			{
				# T2 MUK lk
				$t1 = $M[$R[0]];                       # lk, operand
				$R[0] = ($R[0] + 2);	               # increment A0 (PC)
			}
		}
		elsif($md == 2)
		{
			# T4/T5 MU m or MU m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $a & 0xFFFE;                   # effective address
			$t1 = $M[$a];                       # operand
		}
		elsif($md == 3)
		{
			# T6/T7 MU* m or MU* m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $M[$a & 0xFFFE] & 0xFFFE;      # memory content = effective address
			$t1 = $M[$a];                       # operand	
		}
		if($t1 & SIGN)
		{
			$t1 = ((~ $t1) + 1) & DMASK;        # sign - => 2-compl
			$s = $s ^ 1;
		}
		if($t2 & SIGN)
		{
			$t2 = ((~ $t2) + 1) & DMASK;        # sign - => 2-compl
			$s = $s ^ 1;
		}
		$t3 = ($t1 * $t2) & DPDMASK;            # A2 * operand
		if($t3 != 0)
		{
			if($s)
			{
				$t3 = ((~ $t3) & DPDMASK) + 1;  # sign - => 2-compl
				if( ! ($t3 & DPSIGN))
				{
					$overflow = 1;
				}
			}
			else
			{
				if($t3 & DPSIGN)
				{
					$overflow = 1;
				}
			}
		}
		# print "MU" . prhex(32,$t1) . prhex(32,$t2) . prhex(32,$t3);
		setCR2s($t3);
		putdbls($t3);
	}
	else
	{	
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc,"MUL10");
	}
}
elsif($opc == 9)
{
	if($r1 == 0)
	{
		# Divide
		my $s  = 0;		                               # collect sign result (quotient)
		my $sr = 0;                                    # sign remainder (gets sign of dividend)
		$t1 = getdbls();					           # A1,A2 dividend				
		if($md == 0)
		{
			# T1 DVR r2
			$t2 = $R[$r2];                             # operand, divisor
		}
		elsif($md == 1)
		{
			if($r2 != 0)
			{
				# T3 DVR* r2
				$t2 = $M[($R[$r2] & 0xFFFE)];          # operand, divisor
			}
			else
			{
				# T2 DVK lk
				$t2 = $M[$R[0]];                       # lk, operand, divisor
				$R[0] = ($R[0] + 2);	               # increment A0 (PC)
			}
		}
		elsif($md == 2)
		{
			# T4/T5 DV m or DV m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $a & 0xFFFE;                   # effective address
			$t2 = $M[$a];                       # operand, divisor
		}
		elsif($md == 3)
		{
			# T6/T7 DV* m or DV* m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $M[$a & 0xFFFE] & 0xFFFE;      # memory content = effective address
			$t2 = $M[$a];                       # operand, divisor
		}
		if($t1 & DPSIGN)
		{
			$t1 = (((~ $t1) & DPDMASK) + 1) & DPDMASK;	# make positive (2-compl)
			$s = $sr = $s ^ 1;							# and save sign / sign remainder
		}
		if($t2 & SIGN)
		{
			$t2 = (((~ $t2) & DMASK) + 1) & DMASK;
			$s = $s ^ 1;
		}
		if($t1 == 0)
		{
			$t3 = 0;	                        # dividend = 0 => result = 0
		}
		elsif($t2 == 0)
		{
			$overflow = 1;                      # divisor = 0
			$t3 = 0x7FFF;
			$t1 = 0;
		}
		else
		{
			$t3 = int($t1 / $t2);		        # quotient
			if($t3 & 0xFFFF8000)
			{
				$overflow = 1; 			        # quotient bigger than 15 bit
			}
			if($s)
			{
				$t3 = (((~ $t3)+ 1) & DMASK);
			}	
			$t1 = $t1 % $t2;			        # remainder
			if($sr)                             # remainder negative ?
			{
				$t1 = (((~ $t1) + 1) & DMASK);
			}			
		}
		$R[4] = setCRvalue($t3);                # A2 = quotient
		$R[2] = $t1;                            # A1 = remainder
	}
	else
	{	
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc,"DIV10");
	}
}
elsif($opc == 10)
{
	# double add
	if($r1 == 0)
	{
		$t1 = getdbls();								# A1,A2 double accumulator
		if($md == 0)
		{
			# T1 DAR r2
			$t2 = $R[$r2] << 15;
			if($r2 == 0xE)
			{
				$r2 = 1;
			}
			else
			{
				$r2 = $r2 + 2;
			}
			if($r2 > 0xF)
			{
				# illegal instruction
				ilcerror($R[0]-2, $IR, $opc,"DADD10a");
			}
			$t2 = $t2 | ($R[$r2] & MMASK);				# operand
		}
		elsif($md == 1)
		{
			if($r2 != 0)
			{
				# T3 DAR* r2
				$t2 = $M[($R[$r2] & 0xFFFE)] << 15;
				$t2 = $t2 | ($M[(($R[$r2]+2) & 0xFFFE)] & MMASK);
			}
			else
			{
				# T2 DAK lk1,lk2
				$t2 = $M[$R[0]];         
				$R[0] = ($R[0] + 2);	               	# increment A0 (PC)	
				$t3 = $M[$R[0]];		
				$R[0] = ($R[0] + 2);	               	# increment A0 (PC)	
				$t2 = ($t2 << 15) | ($t3 & MMASK);		# operand
			}
		}
		elsif($md == 2)
		{
			# T4/T5 DA m or DA m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $a & 0xFFFE;                   # effective address
			$t2 = $M[$a] << 15; 
			$t2 = $t2 | $M[($a + 2) & 0xFFFE] & MMASK; # operand
		}
		elsif($md == 3)
		{
			# T6/T7 DA* m or DA* m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $M[$a & 0xFFFE] & 0xFFFE;      # memory content = effective address
			$t2 = $M[$a] << 15; 
			$t2 = $t2 | $M[($a + 2) & 0xFFFE] & MMASK; # operand		
		}
		$t3 = $t1 + $t2;
		# overflow ?
		if((((~$t1) ^ $t2) & ($t1 ^ $t3)) & DPSIGN)
		{
			$overflow = 1;
		}
		setCR2s($t3);
		putdbls($t3);		
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc,"DADD10");
	}
}
elsif($opc == 11)
{
	# double sub
	if($r1 == 0)
	{
		$t1 = getdbls();								# A1,A2 double accumulator
		if($md == 0)
		{
			# T1 DSR r2
			$t2 = $R[$r2] << 15;
			if($r2 == 0xE)
			{
				$r2 = 1;
			}
			else
			{
				$r2 = $r2 + 2;
			}
			if($r2 > 0xF)
			{
				# illegal instruction
				ilcerror($R[0]-2, $IR, $opc,"DSUB10a");
			}
			$t2 = $t2 | ($R[$r2] & MMASK);				# operand
		}
		elsif($md == 1)
		{
			if($r2 != 0)
			{
				# T3 DSR* r2
				$t2 = $M[($R[$r2] & 0xFFFE)] << 15;
				$t2 = $t2 | ($M[(($R[$r2]+2) & 0xFFFE)] & MMASK);
			}
			else
			{
				# T2 DSK lk1,lk2
				$t2 = $M[$R[0]];         
				$R[0] = ($R[0] + 2);	               	# increment A0 (PC)	
				$t3 = $M[$R[0]];		
				$R[0] = ($R[0] + 2);	               	# increment A0 (PC)	
				$t2 = ($t2 << 15) | ($t3 & MMASK);		# operand
			}
		}
		elsif($md == 2)
		{
			# T4/T5 DS m or DS m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $a & 0xFFFE;                   # effective address
			$t2 = $M[$a] << 15; 
			$t2 = $t2 | $M[($a + 2) & 0xFFFE] & MMASK; # operand
		}
		elsif($md == 3)
		{
			# T6/T7 DS* m or DS* m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);        	    # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $M[$a & 0xFFFE] & 0xFFFE;      # memory content = effective address
			$t2 = $M[$a] << 15; 
			$t2 = $t2 | $M[($a + 2) & 0xFFFE] & MMASK; # operand		
		}
		$t3 = $t1 - $t2;
		# underflow ? (-)
		if((($t1 ^ $t2) & ((~$t2) ^ $t3)) & DPSIGN)
		{
			$overflow = 1;
		}	
		setCR2s($t3);
		putdbls($t3);
	}
	else
	{	
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc,"DSUB10");
	}
}
elsif($opc == 12)
{
	# character transfer
	if($r1 != 0)
	{
		# load
		if($md == 0)				            # register operand
		{
			# T1 ECR r1,r2
			$t1 = ($R[$r2] >> 8) & 0xFF;
			$R[$r1] = (($R[$r2] << 8) | $t1) & 0xFFFF;
		}
		elsif($md == 1)				            # indirect register operand
		{
			# T3 LCR r1,r2, T2 LCK r1,lk
			$a = $R[$r2];
			$t1 = $M[$a & 0xFFFE];
			if($a & 1)
			{
				$t1 = $t1 & 0xFF;               # 1 = right hand byte
			}
			else
			{
				$t1 = ($t1 >> 8) & 0xFF;        # 0 = left hand byte
			}
	
			if($r2 == 0)
			{
				# T2 indirect via A0: LCK r1,lk
				$R[0] = ($R[0] + 2);	        # increment A0 (PC)
			}
		}
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 LC r1,m or LC r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			# $a is byte address !
			$t1 = $M[$a & 0xFFFE];              # get word with addressed byte
			if($a & 1)
			{
				$t1 = $t1 & 0xFF;               # 1 = right hand byte
			}
			else
			{
				$t1 = ($t1 >> 8) & 0xFF;        # 0 = left hand byte
			}
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 LC* r1,m or LC* r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $M[$a & 0xFFFE];               # memory content = effective byte address
			# $a is byte address !
			$t1 = $M[$a & 0xFFFE];              # get word with addressed byte
			if($a & 1)
			{
				$t1 = $t1 & 0xFF;               # 1 = right hand byte
			}
			else
			{
				$t1 = ($t1 >> 8) & 0xFF;        # 0 = left hand byte
			}		
		}
		$R[$r1] = ($R[$r1] & 0xFF00) | $t1;	
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "LC10");	
	}		
}
elsif($opc == 13)
{
	# compare word
	if($md == 0)				                # register operand
	{
		# T1 CWR r1,r2
		$t1 = $R[$r1]; $t2 = $R[$r2];
	}
	elsif($md == 1)				                # indirect register operand
	{
		# T3 CWR* r1,r2, T2 CWK r1,lk
        $t2 = $M[($R[$r2] & 0xFFFE)];
		if($r2 == 0)
		{
			# T2 indirect via A0: CWK r1,lk
			$R[0] = ($R[0] + 2);	# increment A0 (PC)
		}
		$t1 = $R[$r1];		
	}
	elsif($md == 2)				                # memory/indexed memory operand
	{
		# T4/T5 CW r1,m or CW r1,m,r2
		$a = $M[$R[0]];			                # memory operand address
		$R[0] = ($R[0] + 2);	                # increment A0 (PC)
		if($r2 != 0)			                # indexing ?
		{
			$a = ($a + $R[$r2]);                # yes
		}
		$a = $a & 0xFFFE;                       # effective address
		$t1 = $R[$r1]; $t2 = $M[$a]; 
	}
	elsif($md == 3)				                # indirect memory/indexed memory operand
	{
		# T6/T7 CW* r1,m or CW* r1,m,r2
		$a = $M[$R[0]];			                # memory operand address
		$R[0] = ($R[0] + 2);	                # increment A0 (PC)
		if($r2 != 0)			                # indexing ?
		{
			$a = ($a + $R[$r2]);                # yes
		}
		$a = $M[$a & 0xFFFE] & 0xFFFE;          # memory content = effective address
		$t1 = $R[$r1]; $t2 = $M[$a];        
	}
	# direct subtraction of t1 and t2 is not possible; prevent over/underflow
	# over/undeflow with subtraction is only possible with t1 and t2 having 
	# different signs (and when comparing 0x8000)
	if(($t1 ^ $t2) & SIGN)
	{
		# sign t1 and t2 are different, so underflow is possible
		if($t1 & SIGN)
		{
			# sign t1 is - (and so sign t2 is +)
			$t3 = -1;		                    # t1 < t2
		}
		else
		{
			# sign t1 is + (and so sign t2 is -)
			$t3 = 1;                            # t1 > t2
		}
	}
	elsif($t1 == $t2)                           # t1 == t2
	{
		# t1 and t2 can be 0x8000, the smallest neg integer. subtraction
		# (making 2complement) gives overflow
		$t3 = 0;
	}
	else
	{
		$t3 = $t1 - $t2;                       
	}
	setCRvalue($t3);                            # set condition register
}
elsif($opc == 14)
{
	# return
	if($r2 != 0)
	{
		# return from function, RTN r2
		$R[$r2] = ($R[$r2] + 2) & 0xFFFE;
		$t1 = $M[$R[$r2]];                            # PSW
		$CR = ($t1 >> 8) & 0x3;                       # restore saved CR
		$R[$r2] = ($R[$r2] + 2) & 0xFFFE;	
		$R[0] = $M[$R[$r2]];                          # restore saved P
		if($r2 == 0xF)                                 # A15 ?
		{
			$PL =  ($t1 >> 10) & 0x3F;                # restore saved PL
			$ENB = 1;
			$UM  = $t1 & 1;                           # restore saved user/system mode	
		}
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "RTN10");			
	}
}
elsif($opc == 15)
{
	# 1-complement
	if($r1 != 0)
	{
		if($md == 0)				            # register operand
		{
			# T1 C1R r1,r2
			$t3 = ~ $R[$r2];
		}
		elsif($md == 1)				            # indirect register operand
		{
			# T3 C1R* r1,r2
			$t3 = ~ $M[($R[$r2] & 0xFFFE)];
		}
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 C1 r1,m or C1 r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $a & 0xFFFE;                   # effective address
			$t3 = ~ $M[$a];                     # complement content of effective address         
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 C1* r1,m or C1* r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $M[$a & 0xFFFE] & 0xFFFE;      # memory content = effective address
			$t3 = ~ $M[$a];                     # complement content of effective address               
		}
		$R[$r1] = setCRvalue($t3);
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "C110");
	}	
}
else
{
	# illegal instruction
	ilcerror($R[0]-2, $IR, $opc,"OPC10");
}
} # end format10


sub format11 {
#
# emulate the format 1 instructions	with the ls bit == 1:

$opc = ($IR >> 11) & 0xF;   # opcode
$r1  = ($IR >> 7)  & 0xF;   # r1 or cnd shifted 1 left	
$r2  = ($IR >> 1)  & 0xF;   # r2
$md  = ($IR >> 5)  & 0x3;   # adressing mode

# print "F11 $opc $r1 $r2 $md\n";

if($opc == 0)
{
	# store
	if($md == 1)				             # indirect register operand
	{
		# T3A/B STR r1,r2
		$M[($R[$r2] & 0xFFFE)] = $R[$r1];
		if($r2 == 0xF)
		{
			# A15, the stack pointer is addressed
			$R[0xF] = ($R[0xF] - 2) & 0xFFFE;
			if($R[0xF] < 0x100) 
			{
				trap(LKM);		            # A15 stack overflow
			}
		}
	}
	elsif($md == 2)				            # memory/indexed memory operand
	{
		# T4/T5 ST r1,m or ST r1,m,r2
		$a = $M[$R[0]];			            # memory operand address
		$R[0] = ($R[0] + 2);	            # increment A0 (PC)
		if($r2 != 0)			            # indexing ?
		{
			$a = ($a + $R[$r2]);            # yes
		}
		$a = $a & 0xFFFE;                   # effective address
		$M[$a] = $R[$r1];                   # store content r1 in effective address
	}
	elsif($md == 3)				            # indirect memory/indexed memory operand
	{
		# T6/T7 ST* r1,m or ST* r1,m,r2
		$a = $M[$R[0]];			            # memory operand address
		$R[0] = ($R[0] + 2);	            # increment A0 (PC)
		if($r2 != 0)			            # indexing ?
		{
			$a = ($a + $R[$r2]);            # yes
		}
		$a =  $M[$a & 0xFFFE] & 0xFFFE;     # memory content = effective address
		$M[$a] = $R[$r1];                   # store content r1 in effective address
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "ST11");
	}
}
elsif($opc == 2)
{
	# add, increment memory
	if($r1 != 0)
	{
		if($md == 1)				            # indirect register operand
		{
			# T3 ADRS r1,r2
			$a  = $R[$r2];
		}
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 ADS r1,m or ADS r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 ADS* r1,m or ADS* r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
		    $a = $M[$a & 0xFFFE];               # memory content = effective address
		}
		else
		{
			# illegal instruction
			ilcerror($R[0]-2, $IR, $opc, "ADD11");
		}
		$a = $a & 0xFFFE;                       # effective address
		$t1 = $R[$r1]; $t2 = $M[$a];
		$t3 = $t1 + $t2;                        # add content of effective address		
		# overflow? (+)
		if((((~$t1) ^ $t2) & ($t1 ^ $t3)) & SIGN)
		{
			$overflow = 1;
		}		
		$M[$a] = setCRvalue($t3);               # store result in effective address
	}
	else
	{
		if($md == 1)				            # indirect register operand
		{
			# T3 IMR r2
			$a = $R[$r2];
		}
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 IM m or IM m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 IM* m or IM* m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a =  $M[$a & 0xFFFE];              # memory content = effective address
		}
		else
		{
			# illegal instruction
			ilcerror($R[0]-2, $IR, $opc, "IM11");
		}
		$a = $a & 0xFFFE;                       # effective address		
		$t1 = 1; $t2 = $M[$a];             
		$t3 = $t1 + $t2;                        # add 1 to content of effective addres		
		# overflow? (+)
		if((((~$t1) ^ $t2) & ($t1 ^ $t3)) & SIGN)
		{
			$overflow = 1;
		}		
		$M[$a] = setCRvalue($t3);               # store result in effective address		
	}
} 
elsif($opc == 3)
{
	# subtract, two complement
	if($r1 != 0)
	{	
		if($md == 1)				            # indirect register operand
		{
			# T3 SURS r1,r2
			$a  = $R[$r2];                      # effective address
		}
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 SUS r1,m or SUS r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 SUS* r1,m or SUS* r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a =  $M[$a & 0xFFFE];              # memory content = effective address
		}
		else
		{
			# illegal instruction
			ilcerror($R[0]-2, $IR, $opc,"SU11");
		}
		$a = $a & 0xFFFE;		                # effective address
		$t1 = $R[$r1]; $t2 = $M[$a];
		$t3 = $t1 - $t2;                        # subtract content of effective address
		# underflow? (-)
		if((($t1 ^ $t2) & ((~$t2) ^ $t3)) & SIGN)
		{
			$overflow = 1;
		}		
		$M[$a] = setCRvalue($t3);               # store result in effective address
	}
	else
	{
		if($md == 1)				            # indirect register operand
		{
			# T3 C2R r2
			$a  = $R[$r2];                      # effective address
		}
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 C2 m or C2 m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}	
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 C2* m or C2* m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a =  $M[$a & 0xFFFE];              # memory content = effective address			
		}
		else
		{
			# illegal instruction
			ilcerror($R[0]-2, $IR, $opc,"C211");
		}
		$a = $a & 0xFFFE;
		$t1 = 0; $t2 = $M[$a];
		$t3 = $t1 - $t2;                        # subtract content of effective address			
		# underflow? (-)
		if((($t1 ^ $t2) & ((~$t2) ^ $t3)) & SIGN)
		{
			$overflow = 1;
		}			
		$M[$a] = setCRvalue($t3);               # store result in effective address
	}
}
elsif($opc == 4)
{
	# AND/TM
	if($r1 != 0)
	{
		if($md == 0)				            # register operand
		{
			# T1 TM r1,r2
			setCRvalue($R[$r1] & $R[$r2]);                         
		}
		else
		{
			if($md == 1)				            # indirect register operand
			{
				# T3 ANRS r1,r2
				$a  = $R[$r2] & 0xFFFE;             # effective address
			}
			elsif($md == 2)				            # memory/indexed memory operand
			{
				# T4/T5 ANS r1,m or ANS r1,m,r2
				$a = $M[$R[0]];			            # memory operand address
				$R[0] = ($R[0] + 2);	            # increment A0 (PC)
				if($r2 != 0)			            # indexing ?
				{
					$a = ($a + $R[$r2]);            # yes
				}
				$a = $a & 0xFFFE;                   # effective address                		
			}
			elsif($md == 3)				            # indirect memory/indexed memory operand
			{
				# T6/T7 ANS* r1,m or ANS* r1,m,r2
				$a = $M[$R[0]];			            # memory operand address
				$R[0] = ($R[0] + 2);	            # increment A0 (PC)
				if($r2 != 0)			            # indexing ?
				{
					$a = ($a + $R[$r2]);            # yes
				}
				$a =  $M[$a & 0xFFFE] & 0xFFFE;     # memory content = effective address           			
			}
			$M[$a] = setCRvalue($R[$r1] & $M[$a]);  # store result in effective address
		}
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "AND11");
	}	
}
elsif($opc == 5)
{
	# OR
	if($r1 != 0)
	{
		if($md == 1)				            # indirect register operand
		{
			# T3 ORRS r1,r2
			$a  = $R[$r2] & 0xFFFE;             # effective address
        }
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 ORS r1,m or ORS r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a = $a & 0xFFFE;                   # effective address                		
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 ORS* r1,m or ORS* r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a =  $M[$a & 0xFFFE] & 0xFFFE;     # memory content = effective address           			
		}
		else
		{
			# illegal instruction
			ilcerror($R[0]-2, $IR, $opc, "OR11b");
		}	
		$M[$a] = setCRvalue($R[$r1] | $M[$a]);  # store result in effective address
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "OR11a");
	}	
}	
elsif($opc == 6)
{
	# XOR/TNM
	if($r1 != 0)
	{
		if($md == 0)				            # register operand
		{
			# T1 TNM r1,r2
			setCRvalue($R[$r1] ^ $R[$r2]);                         
		}
		else
		{
			if($md == 1)				            # indirect register operand
			{
				# T3 XRRS r1,r2
				$a  = $R[$r2] & 0xFFFE;             # effective address
			}
			elsif($md == 2)				            # memory/indexed memory operand
			{
				# T4/T5 XRS r1,m or XRS r1,m,r2
				$a = $M[$R[0]];			            # memory operand address
				$R[0] = ($R[0] + 2);	            # increment A0 (PC)
				if($r2 != 0)			            # indexing ?
				{
					$a = ($a + $R[$r2]);            # yes
				}
				$a = $a & 0xFFFE;                   # effective address                		
			}
			elsif($md == 3)				            # indirect memory/indexed memory operand
			{
				# T6/T7 XRS* r1,m or XRS* r1,m,r2
				$a = $M[$R[0]];			            # memory operand address
				$R[0] = ($R[0] + 2);	            # increment A0 (PC)
				if($r2 != 0)			            # indexing ?
				{
					$a = ($a + $R[$r2]);            # yes
				}
				$a =  $M[$a & 0xFFFE] & 0xFFFE;     # memory content = effective address           			
			}
			$M[$a] = setCRvalue($R[$r1] ^ $M[$a]);  # store result in effective address
		}
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "XOR11");
	}	
}
elsif($opc == 7)
{
	# print "multiple store $md\n";
	$t1 = 1;
	# multiple store
	if($md == 1)				            # indirect register operand
	{
		# T3A/B MSR n,r2
		# $r1 is n
		if($r2 == 0xF)
		{
			# A15, the stack pointer is addressed (adapt maybe for R[0xF] is not even **************
		    $a = $R[$r2] & 0xFFFE; # effective address (save start value of $R[0xF] because a15 can be put in the stack)
			while($t1 <= $r1)
			{
				$M[$a] = $R[$A[$t1]];
				$a=$a-2;
				if($a < 0x100)
				{
					$R[0xF]=$a;					
					trap(LKM);		        # A15 stack overflow
				}					
				$t1 = $t1 + 1;
			}
			$R[0xF]=$a;			
		}
		else
		{
			$a = $R[$r2] & 0xFFFE;              # effective address						
			while($t1 <= $r1)
			{
				$M[$a] = $R[$A[$t1]];
				$a = $a + 2;
				$t1 = $t1 + 1;
			}
		}
	}
	elsif($md == 2)				            # memory/indexed memory operand
	{
		# print "MS\n";
		# T4/T5 MS n,m or MS n,m,r2
		$a = $M[$R[0]];			            # memory operand address
		$R[0] = ($R[0] + 2);	            # increment A0 (PC)
		if($r2 != 0)			            # indexing ?
		{
			$a = ($a + $R[$r2]);            # yes
		}
		$a = $a & 0xFFFE;                   # effective address
		
		while($t1 <= $r1)
		{
			$M[$a] = $R[$A[$t1]];           # store content if effective address
			# print "MS $t1 $r1 " . prhex(16,$a) . prhex(16,$M[$a]) . "\n";
			$a = $a + 2;
			$t1 = $t1 + 1;
		}			
	}
	elsif($md == 3)				            # indirect memory/indexed memory operand
	{
		# T6/T7 MS* n,m or MS* n,m,r2
		$a = $M[$R[0]];			            # memory operand address
		$R[0] = ($R[0] + 2);	            # increment A0 (PC)
		if($r2 != 0)			            # indexing ?
		{
			$a = ($a + $R[$r2]);            # yes
		}
		$a =  $M[$a & 0xFFFE] & 0xFFFE;     # memory content = effective address
		while($t1 <= $r1)
		{
			$M[$a] = $R[$A[$t1]];           # store content if effective address
			$a = $a + 2;
			$t1 = $t1 + 1;
		}				
	}
	else
	{
		# md == 0
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "MS11");
	}
}	
elsif($opc == 12)
{
	# character transfer
	if($r1 !=0)
	{
		# store
		if($md == 1)				            # indirect register operand
		{
			# T3 SCR r1,r2
			$a = $R[$r2];
			# $a is byte address
			$t1 = $M[$a & 0xFFFE];              # get word with addressed byte
			if($a & 1)
			{
				$t1 = ($t1 & 0xFF00) | ($R[$r1] & 0xFF); # right hand byte
			}
			else
			{
				$t1 = ($t1 & 0x00FF) | ($R[$r1] << 8);   # left hand byte
			}
		}
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 SC r1,m or SC r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			# $a is effective byte address
			$t1 = $M[$a & 0xFFFE];              # get word with addressed byte
			if($a & 1)
			{
				$t1 = ($t1 & 0xFF00) | ($R[$r1] & 0xFF); # right hand byte
			}
			else
			{
				$t1 = ($t1 & 0x00FF) | ($R[$r1] << 8);   # left hand byte
			}
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 SC* r1,m or SC* r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a =  $M[$a & 0xFFFE];              # memory content = effective byte address
			# $a is effective byte address
			$t1 = $M[$a & 0xFFFE];              # get word with addressed byte
			if($a & 1)
			{
				$t1 = ($t1 & 0xFF00) | ($R[$r1] & 0xFF); # right hand byte
			}
			else
			{
				$t1 = ($t1 & 0x00FF) | ($R[$r1] << 8);   # left hand byte
			}
		}
		else
		{
			# illegal instruction
			ilcerror($R[0]-2, $IR, $opc, "SC11b");
		}
		$M[$a & 0xFFFE] = $t1 & 0xFFFF;              # replace word
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "SC11a");	
	}
}
elsif($opc == 13)
{
	# compare characters
	if($r1 !=0)
	{	
		if($md == 1)				                # indirect register operand
		{
			# T3 CCR r1,r2, T2 CCK r1,lk
			$a  = $R[$r2];
			$t1 = $R[$r1] & 0xFF; $t2 = $M[($a & 0xFFFE)];
			if($a & 1)
			{
				$t2 = $t2 & 0xFF;                   # 1 = right hand byte
			}
			else
			{
				$t2 = ($t2 >> 8) & 0xFF;            # 0 = left hand byte
			}		
			if($r2 == 0)
			{
				# T2 indirect via A0: CWK r1,lk
				$R[0] = ($R[0] + 2);	# increment A0 (PC)
			}
		}
		elsif($md == 2)				                # memory/indexed memory operand
		{
			# T4/T5 CC r1,m or CC r1,m,r2
			$a = $M[$R[0]];			                # memory operand address
			$R[0] = ($R[0] + 2);	                # increment A0 (PC)
			if($r2 != 0)			                # indexing ?
			{
				$a = ($a + $R[$r2]);                # yes
			}
			$t1 = $R[$r1] & 0xFF; $t2 = $M[($a & 0xFFFE)]; 
			if($a & 1)
			{
				$t2 = $t2 & 0xFF;                   # 1 = right hand byte
			}
			else
			{
				$t2 = ($t2 >> 8) & 0xFF;            # 0 = left hand byte
			}				
		}
		elsif($md == 3)				                # indirect memory/indexed memory operand
		{
			# T6/T7 CC* r1,m or CC* r1,m,r2
			$a = $M[$R[0]];			                # memory operand address
			$R[0] = ($R[0] + 2);	                # increment A0 (PC)
			if($r2 != 0)			                # indexing ?
			{
				$a = ($a + $R[$r2]);                # yes
			}
			$a = $M[$a & 0xFFFE];                   # memory content = effective address
			$t1 = $R[$r1] & 0xFF; $t2 = $M[($a & 0xFFFE)];  
			if($a & 1)
			{
				$t2 = $t2 & 0xFF;                   # 1 = right hand byte
			}
			else
			{
				$t2 = ($t2 >> 8) & 0xFF;            # 0 = left hand byte
			}				
		}
		else
		{
			# illegal instruction
			ilcerror($R[0]-2, $IR, $opc, "CC11b");			
		}		
		# t1 and t2 contain the characters to compare
		$t3 = $t1 - $t2; 
		setCRvalue($t3);                        # set condition register
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "CC11a");			
	}		
}
elsif($opc == 14)
{
	# call function
	if($r1 != 0)
	{
		# call function
		if($md == 0)				                    # register operand
		{
			# T1 CFR r1,r2
			if($r1 != 0xF)                              # not A15
			{
				$M[$R[$r1]] = $R[0];                    # save P
				$R[$r1] = $R[$r1] - 2;                  # decrease link stack
				$M[$R[$r1]] = Fpsw();                   # save PSW
				$R[$r1] = $R[$r1] - 2;                  # decrease link stack
				$R[0] = $R[$r2] & 0xFFFE;
			}
			else                                        # via A15
			{
				$M[$R[0xF]] = $R[0];                    # save P
				$R[0xF] = $R[0xF] - 2;                  # decrease link stack
				if($R[0xF] < 0x100)
				{
					trap(LKM);		                    # A15 stack overflow
				}				
				$M[$R[0xF]] = Fpsw();                   # save PSW
				$R[0xF] = $R[0xF] - 2;                  # decrease link stack
				if($R[0xF] < 0x100)
				{
					trap(LKM);		                    # A15 stack overflow
				}					
				$R[0] = $R[$r2] & 0xFFFE;				
			}
		}
		elsif($md == 1)				                    # indirect register operand
		{
			# T2/T3 CF r1,lk   CFR* r1,r2
			$t1 =  $M[$R[$r2]& 0xFFFE] & 0xFFFE;        # save new P value
			if($r2 == 0)
			{
				# T2 CF r1,lk
				$R[0] = ($R[0] + 2);                    # increment A0 (PC)
			}
			if($r1 != 0xF)    			
			{
				$M[$R[$r1]] = $R[0];                    # save P on stack
				$R[$r1] = $R[$r1] - 2;                  # decrease link stack
				$M[$R[$r1]] = Fpsw();                   # save PSW on stack
				$R[$r1] = $R[$r1] - 2;                  # decrease link stack
				$R[0] = $t1;                            # jump
			}				
			else                                        # via A15
			{
				$M[$R[0xF]] = $R[0];                    # save P on stack
				$R[0xF] = $R[0xF] - 2;                  # decrease link stack
				if($R[0xF] < 0x100)
				{
					trap(LKM);		                    # A15 stack overflow
				}				
				$M[$R[0xF]] = Fpsw();                   # save PSW on stack
				$R[0xF] = $R[0xF] - 2;                  # decrease link stack
				if($R[0xF] < 0x100)
				{
					trap(LKM);		                    # A15 stack overflow
				}					
				$R[0] = $t1;		                    # jump		
			}
		}
		elsif($md == 2)				                    # memory/indexed memory operand
	    {	
			# T4/T5 CFI r1,m or CFI r1,m,r2
			$a = $M[$R[0]];		                        # memory operand address
			$R[0] = ($R[0] + 2);                        # increment A0 (PC)
			if($r2 != 0)			                    # indexing ?
			{
				$a = ($a + $R[$r2]);                    # yes
			}
			$a = $a & 0xFFFE;                           # effective address
			if($r1 != 0xF)    			
			{
				$M[$R[$r1]] = $R[0];                    # save P
				$R[$r1] = $R[$r1] - 2;                  # decrease link stack
				$M[$R[$r1]] = Fpsw();                   # save PSW
				$R[$r1] = $R[$r1] - 2;                  # decrease link stack
				$R[0] = $M[$a];
			}				
			else                                        # via A15
			{
				$M[$R[0xF]] = $R[0];                    # save P
				$R[0xF] = $R[0xF] - 2;                  # decrease link stack
				if($R[0xF] < 0x100)
				{
					trap(LKM);		                    # A15 stack overflow
				}				
				$M[$R[0xF]] = Fpsw();                   # save PSW
				$R[0xF] = $R[0xF] - 2;                  # decrease link stack
				if($R[0xF] < 0x100)
				{
					trap(LKM);		                    # A15 stack overflow
				}					
				$R[0] = $M[$a];				
			}		
		}
		elsif($md == 3)				                    # indirect memory/indexed memory operand
		{
			# T6/T7 CFI* r1,m or CFI* r1,m,r2
			$a = $M[$R[0]];			                    # memory operand address
			$R[0] = ($R[0] + 2);	                    # increment A0 (PC)
			if($r2 != 0)			                    # indexing ?
			{
				$a = ($a + $R[$r2]);                    # yes
			}
			$a = $M[$a & 0xFFFE] & 0xFFFE;              # memory content = effective address
			if($r1 != 0xF)    			
			{
				$M[$R[$r1]] = $R[0];                    # save P
				$R[$r1] = $R[$r1] - 2;                  # decrease link stack
				$M[$R[$r1]] = Fpsw();                   # save PSW
				$R[$r1] = $R[$r1] - 2;                  # decrease link stack
				$R[0] = $M[$a];
			}				
			else                                        # via A15
			{
				$M[$R[0xF]] = $R[0];                    # save P
				$R[0xF] = $R[0xF] - 2;                  # decrease link stack
				if($R[0xF] < 0x100)
				{
					trap(LKM);		                    # A15 stack overflow
				}				
				$M[$R[0xF]] = Fpsw();                   # save PSW
				$R[0xF] = $R[0xF] - 2;                  # decrease link stack
				if($R[0xF] < 0x100)
				{
					trap(LKM);		                    # A15 stack overflow
				}					
				$R[0] = $M[$a];				
			}		
		}
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "CF1!");	
	}		
}
elsif($opc == 15)
{
	# 1-complement
	if($r1 == 0)
	{
		if($md == 1)				            # indirect register operand
		{
			# T3 C1RS r1,r2
			$a  = $R[$r2];                      # effective address
        }
		elsif($md == 2)				            # memory/indexed memory operand
		{
			# T4/T5 C1S r1,m or C1S r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);         	    # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}               		
		}
		elsif($md == 3)				            # indirect memory/indexed memory operand
		{
			# T6/T7 C1S* r1,m or C1S* r1,m,r2
			$a = $M[$R[0]];			            # memory operand address
			$R[0] = ($R[0] + 2);	            # increment A0 (PC)
			if($r2 != 0)			            # indexing ?
			{
				$a = ($a + $R[$r2]);            # yes
			}
			$a =  $M[$a & 0xFFFE];              # memory content = effective address           			
		}
		else
		{
			# illegal instruction
			ilcerror($R[0]-2, $IR, $opc, "C111b");
		}
		$a = $a & 0xFFFE;                       # effective address 
		$M[$a] = setCRvalue(~ $M[$a]);          # store result in effective address
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "C111a");
	}	
}	
else
{
	# illegal instruction
	ilcerror($R[0]-2, $IR, $opc,"OPC11");
}

} # end format11

sub format0 {
#
# emulate the format 0 instructions:
# 	- short constant instructions
# 	- shift instructions
# 	- i/o instructions
# 	- miscellanious instructions
$opc = ($IR >> 11) & 0xF;	# opcode
$r3  = ($IR >> 7)  & 0xE;   # r3 or cnd (shifted 1 left) 
$k   = $IR & 0xFF;          # shortconstant
				
if($opc == 0)
{
	# load constant LDK
	$R[$r3] = $k;			# CR unchanged !!!!
}
elsif($opc == 1)
{
	# absolute branche AB
	# $r3 is the condition code in operation
	$r3 = $r3 >> 1; # position the condition code	
	if(evalcond($r3))
	{
		$R[0] = $k;
	}
}
elsif($opc == 2)
{
	# add constant ADK
	$t1 = $R[$r3];
	$t2 = $t1 + $k;
	# overflow? (+)
	if((((~$t1) ^ $k) & ($t1 ^ $t2)) & SIGN)
	{
		$overflow = 1;
	}
	$R[$r3] = setCRvalue($t2);
}
elsif($opc == 3)
{
	# subtract constant SUK
	$t1 = $R[$r3];
	$t2 = $t1 - $k;
	# underflow? (-)
	if((($t1 ^ $k) & ((~$k) ^ $t2)) & SIGN)
	{
		$overflow = 1;
	}
	$R[$r3] = setCRvalue($t2);
}
elsif($opc == 4)
{
	# logical and constant ANK - MIS: HLT, INH, RIT
	if($r3 != 0)
	{
		# ANK
		$R[$r3] = $R[$r3] & $k;
		setCR1($r3);
	}
	else
	{
		if($k == 0x7F)
		{
			# HLT
			$RUN = 0;
			print "\nExecution stopped with a HLT\n";
		}
		elsif($k == 0xBF)
		{
			# INH
			$ENB = 0;
		}
		else
		{
			my $t1 = $k & 0xC1;
			if($t1 == 0xC1)
			{
				# RIT
				# resetting more internal interrupts in 0ne instruction is allowed				
				$t1 = ($k & 0x3F) >> 1; # 5 bits resetting more internal interrupts in 0ne instruction is allowed
				$t1 = ~ $t1;
				
				if($t1 & 4) # 0x1B
				{
					$intreg = $intreg & (~ RTC);     # reset clock interrupt
				}
				elsif($t1 & 2) # 0x1D
				{
					$intreg = $intreg & (~ LKM);     # reset LKM interrupt
				}
				elsif($t1 & 16) # 0x0F
				{
					$intreg = $intreg & (~ CP);      # reset control panel interrupt
				}
				elsif($t1 & 8) # 0x17
				{
					$intreg = $intreg & (~ PWR);     # reset power failure interrupt
				}
				elsif($t1 & 1) # 0x1E
				{
					$intreg = $intreg & (~ MP);      # reset memory protect interrupt
				}
				# ignore other bits				
			}
			else
			{
				# illegal instruction
				ilcerror($R[0]-2, $IR, $opc, "F04");
			}
		}
	}
}
elsif($opc == 5)
{
	# logical or constant ORK - MIS: LKM ENB SMD
	if($r3 != 0)
	{
		# ORK
		$R[$r3] = $R[$r3] | $k;
		setCR1($r3);
	}
	else
	{
		if($k == 0x04)
		{
			# LKM
			trap(LKM);
		}
		elsif($k == 0x40)
		{
			# ENB
			$ENB = 1;
		}
		elsif($k == 0x01)
		{
			# SMD
			# $UM  = 1;  # No memory protection implemented
		}
        # ignore setting of other bits
	}
}
elsif($opc == 6)
{
	# excl or constant XRK
	if($r3 != 0)
	{
		# XRK
		$R[$r3] = $R[$r3] ^ $k;
		setCR1($r3);
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "F06");
	}
}
elsif($opc == 7)
{
	# shift: single/double word shift, normalize
	my $tlng;
	my $i;
	$sc   = $k & 0x1F;                      # shift count (5 bit)
	$sfun = ($k & 0xE0) >> 5;               # shift function (3 bit)
	$t1   = $sc;
	if($r3 != 0)
	{
		# single register shifts
				
		if($sfun == 0)
		{
			# SLA: single left arithmetic shift
			if($t1 > 16) { $t1 = 16 };      # limit to 16 
			$t2 = $R[$r3] & SIGN;           # save sign
			$i = 0;
			while($i < $t1)
			{
				$R[$r3] = $R[$r3] << 1;
				if(($R[$r3] & SIGN) != $t2)
				{
					$overflow = 1;
				}
				$i = $i + 1;
			}
			setCR1($r3);
		}
		elsif($sfun == 1)
		{
			# SRA: single right arithmetic shift
			if($t1 > 16) { $t1 = 16 };      # limit to 16 
			$t2 = $R[$r3] & SIGN;           # save sign
			$i=0;
			while($i < $t1)
			{
				$R[$r3] = $R[$r3] >> 1;
				$R[$r3] = $R[$r3] | $t2;
				$i = $i + 1;
			}
			setCR1($r3);
		}
		elsif($sfun == 2)
		{
			# SLL: single left logical shift
		    if($t1 > 16) 
			{ 
				$R[$r3] = 0;               # > 16 ? all 0 
			}
			else
			{
				$R[$r3] = ($R[$r3] << $t1) & DMASK;
			}
			setCR1($r3);
		}
		elsif($sfun == 3)
		{
			# SRL: single right logical shift
			if($t1 > 16) 
			{ 
				$R[$r3] = 0;               # > 16 ? all 0 
			}
			else
			{
				$R[$r3] = ($R[$r3] >> $t1) & DMASK;
			}
			setCR1($r3);
		}
		elsif($sfun == 6)
		{
			# SLC: single left circular shift
			$t2 = $t1 % 16;               # mod 16, t2 is remainder
			$R[$r3] = (($R[$r3] << $t2) | ($R[$r3] >> (16 -$t2)) & DMASK);
			setCR1($r3);
		}
		elsif($sfun == 7)
		{
			# SRC: single right circular shift
			$t2 = $t1 % 16;               # mod 16, t2 is remainder
			$R[$r3] = (($R[$r3] >> $t2) | ($R[$r3] << (16 -$t2)) & DMASK);
			setCR1($r3);
		}
		elsif($sfun == 4)
		{
			# SLN: single left normalize shift
			$r2 = ($k & 0x1E) >> 1;
			$t2 = $R[$r3] & SIGN;          # save sign (bit 0)
			$t3 = $t2 >> 1;                # sign shifted 1 right for comparison			
			$i = 0;
			while($i < 16)
			{
				$t1 = $R[$r3] & 0x4000;    # extract bit 1
				if($t3 == $t1)             # bit 0 equal to bit 1?
				{
					$R[$r3] = $R[$r3] << 1;
					$i = $i + 1;
				}
				else
				{
					last;
				}
			}
			$R[$r3] = ($R[$r3] & MMASK) | $t2;
			$R[$r2] = $i;
		}
		elsif($sfun == 5)
		{
			# SRN: single right normalize shift
			$r2 = ($k & 0x1E) >> 1;
			$t2 = $R[$r3] & SIGN;          # save sign
			$i = 0;
			while($i < 16)
			{
				$t1 = $R[$r3] & 1;
				if($t1 == 0)
				{
					$R[$r3] = ($R[$r3] >> 1) | $t2;
					$i = $i + 1;
				}
				else
				{
					last;
				}
			}
			$R[$r3] = ($R[$r3] & MMASK) | $t2;
			$R[$r2] = $i;
		}
	}
	else
	{
		# double register shifts (A1-A2)
				
		if($sfun == 0)
		{
			# DLA: double left arithmetic shift
			$tlng = getdbls();				# get A1, A2 as long
			$t2 = $tlng & DPSIGN;           # save sign
			$i = 0;
			while($i < $t1)
			{
				$tlng = $tlng << 1;
				if(($tlng & DPSIGN) != $t2)
				{
					$overflow = 1;
				}
				$i = $i + 1;
			}
			setCR2s($tlng);	                 # set condition code
			putdbls($tlng);                  # restore A1, A2
		}
		elsif($sfun == 1)
		{
			# DRA: double right arithmetic shift
			$tlng = getdbls();				# get A1, A2 as long
			$t2 = $tlng & DPSIGN;           # save sign
			$i=0;
#			if($t1 != 0) # special case for 0 shifts: keep A1/A2 as is (from testprogram CPB8K)
#           the P852 in case of 0 shifts did not change bit-0 of A2 (if it is 1, it remains 1)
#           the P854 (and later models) makes bit-0 of A2 always 0 also in case of zero shifts.
#           This implementation follows approach (at this moment): bit-0 A2 is always made 0
#			{
				while($i < $t1)
				{
					$tlng = $tlng >> 1;
					$tlng = $tlng | $t2;
					$i = $i + 1;
				}
#			}
			setCR2s($tlng);
			putdbls($tlng);                  # restore A1, A2
		}
		elsif($sfun == 2)
		{
			# DLL: double left logical shift
			$tlng = getdblu();				 # get A1, A2 as long
			$tlng = $tlng << $t1;
			setCR2u($tlng);
			putdblu($tlng);                  # restore A1, A2
		}
		elsif($sfun == 3)
		{
			# DRL: double right logical shift
			$tlng = getdblu();				 # get A1, A2 as long
			$tlng = $tlng >> $t1;
			setCR2u($tlng);
			putdblu($tlng);                  # restore A1, A2
		}
		elsif($sfun == 6)
		{
			# DLC: double left circular shift
			$tlng = getdblu();				 # get A1, A2 as long
			$tlng = ($tlng << $t1) | ($tlng >> (32 - $t1));
			setCR2u($tlng);
			putdblu($tlng);                  # restore A1, A2
		}
		elsif($sfun == 7)
		{
			# DRC: double right circular shift
			$tlng = getdblu();				 # get A1, A2 as long
			$tlng = ($tlng >> $t1) | ($tlng << (32 - $t1));
			setCR2u($tlng);
			putdblu($tlng);                  # restore A1, A2
		}
		elsif($sfun == 4)
		{
			# DLN: double left normalize shift
			$r2 = ($k & 0x1E) >> 1;
			$tlng = getdbls();				 # get A1, A2 as long
			$t2 = $tlng & DPSIGN;            # save sign
			$t3 = $t2 >> 1;                  # sign shifted 1 right for comparison			
			$i = 0;
			while($i < 31)
			{
				$t1 = $tlng & 0x20000000;
				if($t3 == $t1)
				{
					$tlng = $tlng << 1;
					$i = $i + 1;
				}
				else
				{
					last;
				}
			}
			$tlng = ($tlng & DPMMASK) | $t2;
			putdbls($tlng);                  # restore A1, A2
			$R[$r2] = $i;
		}
		elsif($sfun == 5)
		{
			# DRN: double right normalize shift
			$r2 = ($k & 0x1E) >> 1;
			$tlng = getdbls();				 # get A1, A2 as long
			$t2 = $tlng & DPSIGN;            # save sign
			$i = 0;
			while($i < 31)
			{
				$t1 = $tlng & 1;
				if($t1 == 0)
				{
					# $R[$r3] = ($R[$r3] >> 1) | $t2; # fout
					$tlng = ($tlng >> 1) | $t2;					
					$i = $i + 1;
				}
				else
				{
					last;
				}
			}
			$tlng = ($tlng & DPMMASK) | $t2;
			putdbls($tlng);                  # restore A1, A2
			$R[$r2] = $i;
		}
	}		
}
elsif($opc == 8)
{
	# IO: CIO OTR - MIS: WIM WMP WM2
	# no Memory Protection implemented (WPM and WM2)
	if($r3 !=0)
	{
		if($k == 0)
		{
			# WIM , write interrupt mask register (a 1 for a disabled interrupt)
			$intmsk = $R[$r3];
		}
		else
		{
			my $bit8 = $k & 0x80;
			my $bit9 = $k & 0x40;                 # function bit
			my $devaddr = $k & 0x3F;
			my $d = $DEV{$devaddr};               # pointer to device block
			if($d eq undef)
			{
				# unknown device address
				$CR = 3;
			}
			elsif($bit8 == 0x80)
			{
				# CIO
				$n = $$d[CIO];                    # CIO routine of device
				&$n($bit9);                       # call
			}
			else
			{
				# OTR
				$n = $$d[OTR];                    # OTR routine of device
				&$n($bit9);                       # call
			}
		}
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "IO1");
	}
}
elsif($opc == 9)
{
	# IO: INR SST TST - MIS: RIL
	if($r3 !=0)
	{
		if($k == 0)
		{
			# RIL , read interrupt line status to A3
			# in the mask register is a 1 for an disabled interrupt
			$R[$r3] = ($intreg & (~ $intmsk)) & 0xFFFF; # state of common line
		}
		else
		{
			my $bit8 = $k & 0x80;
			my $bit9 = $k & 0x40;
			my $devaddr = $k & 0x3F;
			my $d = $DEV{$devaddr};               # pointer to device block
			if($d eq undef)
			{
				# unknown device address
				$CR = 3;
			}
			elsif($bit8 == 0)
			{
				# INR
				$n = $$d[INR];                    # INR routine of device
				&$n($bit9);                       # call			
			}
			elsif($bit9 == 0)
			{
				# TST
				$n = $$d[TST];                    # TST routine of device
				&$n();                            # call			
			}
			else
			{
				# SST
				$n = $$d[SST];                    # SST routine of device
				&$n();                            # call	
			}
		}
	}
	else
	{
		# illegal instruction
		ilcerror($R[0]-2, $IR, $opc, "IO2");
	}
}
elsif($opc == 0xA)
{
	# RF
	# $r3 (4 bit) is the condition code in operation >> 1
	$r3 = $r3 >> 1;
	if(evalcond($r3))
	{
		$R[0] = ($R[0] + $k);
	}
}
elsif($opc == 0xB)
{
	# RB
	# $r3 (4 bit) is the condition code in operation
	$r3 = $r3 >> 1;
	if(evalcond($r3))
	{
		$R[0] = ($R[0] - $k);
	}
}
else
{
	# illegal instruction
	ilcerror($R[0]-2, $IR, $opc, "OPC00");	
}
} # end format0

########################### emulator subroutines #########################
# Evaluate the Condition code in an operation against the current CR value
# If true, the function returns a 1, else 0.
sub evalcond {
	my $cd = shift;  	# condition code (3 bit) from operation
	my $not;
	# print "cond $cd CR $CR\n"; ##############################
	if($cd == 7)
	{
		return 1;	# unconditional
	}
	$not  = $cd & 4;
	$cd   = $cd & 3;
	if($not)        # not bit
	{
		if($cd == $CR)
		{
			return 0;
		}
		else
		{
			return 1;
		}
	}
	else
	{
		if($cd == $CR)
		{
			return 1;
		}
		else
		{
			return 0;
		}
	}
}

# set Condition Register according value of r1 or r3 (or temporary register in @R)
sub setCR1 {
	my $r = shift;
	my $t = $R[$r];
	if($overflow)
	{
		$CR = 3;                 # overflow
	}
	elsif($t & 0x8000)
	{
		$CR = 2;                 # result LT 0
	}
	elsif($t & 0x7FFF)
	{
		$CR = 1;                 # result GT 0
	}
		else
	{
		$CR = 0;                 # result = 0
	}
	# print "setCR1 " . prhex(32,$R[$r]) . " CR $CR\n";  ##############
	$R[$r] = $t & 0xFFFF;
	$overflow = 0;
}

# set Condition Register according  a given value
# the masked 16-bit value is reurned
sub setCRvalue {
	my $t = shift;
	if($overflow)
	{
		$CR = 3;                 # overflow
	}
	elsif($t & 0x8000)
	{
		$CR = 2;                 # result LT 0
	}
	elsif($t & 0x7FFF)
	{
		$CR = 1;                 # result GT 0
	}
		else
	{
		$CR = 0;                 # result = 0
	}
	# print "setCR1 " . prhex(32,$R[$r]) . " CR $CR\n";  ##############
	$overflow = 0;
    return $t & 0xFFFF;
}

# set Condition Register according value of A1, A2 as 31 bit signed long integer
# bit 1 (A1 bit 0) is sign bit (bit 0 A2 excluded)
# to be used in combination of getdbls/putdbls
sub setCR2s {
	my $l = shift;
	if($overflow)               # DLA + arith ops like double add, etc.
	{
		$CR = 3;                 # overflow
	}	
	elsif($l & 0x40000000)
	{
		$CR = 2;                 # result LT 0
	}
	elsif($l & 0x3FFFFFFF)
	{
		$CR = 1;                 # result GT 0
	}
	else
	{
		$CR = 0;                 # result = 0
	}
	# print "setCRs " . prhex(32,$l) . " CR $CR\n";       ################
	$overflow = 0;
}

# set Condition Register according value of A1, A2 as 32 bit signed long integer
# bit 0 (A1 bit 0) is sign bit
# to be used in combination of getdblu/putdblu
sub setCR2u {
	my $l = shift;
	if($l & 0x80000000)
	{
		$CR = 2;                 # result LT 0
	}
	elsif($l & 0x7FFFFFFF)
	{
		$CR = 1;                 # result GT 0
	}
	else
	{
		$CR = 0;                 # result = 0
	}
	# print "setCRu " . prhex(32,$l) . " CR $CR\n";       ################
}

# In the following 4 routines A1 an A2 are accessed as long accumulator.
# !!!!!! A1 = $R[2] A2=$R[4] !!!!!!!!
# get double signed
# returns A1, A2 as signed 31 value
sub getdbls {
	return (($R[2] << 15) | ($R[4] & MMASK));
}

# get double unsigned
# returns A1, A2 as unsigned 32 value
sub getdblu {
	return (($R[2] << 16) | $R[4] );
}

# put double signed
# stores signed 31 bit value in A1, A2
sub putdbls {
	my $x = shift;
	$x = $x & 0x7FFFFFFF;
	$R[2] = ($x >> 15) & DMASK;
	$R[4] = ($x & MMASK);
}

# put double unsigned
# stores unsigned 32 bit value in A1, A2
sub putdblu {
	my $x = shift;
	$R[2] = ($x >> 16) & DMASK;
	$R[4] = $x & DMASK;
}

# internal interrupt
sub trap {
	my $int = shift;
	my $pc = $R[0] - 2;
	$intreg = $intreg | $int;
	if($DEBUG)
	{
		$int = $int >> 16;         # print priority flipflop state
		print "\nTrap" . prhex(8,$int) . " at" . prhex(16,$pc) . "\n";
		print "Opcode =" . prhex(4,$opc) . " Instruction =" . prhex(16,$IR) . "\n";
		trace($pc);
	}
}

# Illegal instruction 
sub ilcerror {
	my $lc    = shift;
	my $instr = shift;
	my $opc   = shift;
	my $ident = shift;
	if($DEBUG)
	{
		print "\nERROR: Illegal Instruction at" . prhex(16,$lc) . " IR:" . prhex(16,$instr) . " OPCODE:" . prhex(4,$opc) . " $ident\n";
    }
	if($StopOnIllInstr)
	{
		print "\nERROR: Illegal Instruction at" . prhex(16,$lc) . " IR:" . prhex(16,$instr) . " OPCODE:" . prhex(4,$opc) . " $ident\n";		
		$RUN = 0;
	}
	else
	{
		trap(LKM);      # trap illegal instruction, PL=1
	}
}

# Illegal IOP channel condition
sub mpxerror {
	if($StopOnMpxError)
	{
		trace($pcsaved);
		$RUN=0;
	}
}

# format PSW
sub Fpsw {
#	my $PL;             # priority level 			psw bit 0:5
#	my $CR;             # condition register		psw bit 6:7
#	my $overflow=0;     # (aid to set CR, must be 0)
#	my $RUN;            # run bit               	psw bit 8
#	my $ENB;            # interrupt enable bit		psw bit 9
#	my $CP;             # control panel interrupt	psw bit 10      X
#	my $PWR;            # power fail interrupt      psw bit 11      X
#	my $RTC;            # real time clock interrupt psw bit 12      X
#	my $LKM;            # link monitor interrupt    psw bit 13      X
#						# + stack overflow and illegal instruction
#	my $MP;             # memory protect interrupt  psw bit 14      X
#	my $UM;             # user mode indicator       psw bit 15
#                       ########### X bits are part of $intreg

my $i = ($intreg >> 16) & 0x1F;      # isolate the 5 internal interrupts represented in the PSW
$PSW = ((((((((($PL << 2) + $CR) << 1) + $RUN) << 1) + $ENB) << 5) + $i) << 1) | $UM;
return $PSW;
}

sub reset_cpu {
	$PL  = 0x3F;     #??? or is the reset value 0 ???
	$CR  = 0;
	$overflow = 0;
	$RUN = 0;
	$ENB = 0;        #??? probably 1 after reset (all bootstraps have INH as 1st instruction
	$UM  = 0;        # system mode
	$PSW = 0;
	$intreg = 0;  # interrupts
	$intmsk = 0;  # interrupt mask
	$brkreg = 0;  # MPX breaks
	# all registers, except A0 to 0
	$R[1]  = 0;
	$R[2]  = 0;	
	$R[3]  = 0;	
	$R[4]  = 0;
	$R[5]  = 0;	
	$R[6]  = 0;		
	$R[7]  = 0;
	$R[8]  = 0;	
	$R[9]  = 0;	
	$R[10] = 0;
	$R[11] = 0;	
	$R[12] = 0;	
	$R[13] = 0;
	$R[14] = 0;	
	$R[15] = 0;	
#########################################################
# as a debugging aid, A15 is set to a predefined value: #
	$R[15] = 0x1FE;                                     #
# for the same reason ENB is set to 0 and PL to /3F     #
#########################################################
}

# the memory array M should only be used for even addresses
# in case an oneven address contains a value, there is an error
# in the simulators memory access
# on even address, also 1-bits outside the lower 16 part of each
# word should not be there
sub mem_check {
	my $fl = shift;
	my $b = 0;
	my $i = 1;
	while($i < 0x10000)
	{
		if($i & 1)
		{
			# uneven address
			if($M[$i] ne undef)
			{
				print "ERROR! value" . prhex(16,$M[$i]) . " is found at the uneven address" . prhex(16,$i) . "\n";
				$b=1;
			}
		}
		else
		{
			# even address
			if($M[$i] & 0xFFFF0000)
			{
				print "ERROR! wrong bits" . prhex(32,$M[$i]) . " found at the even address" . prhex(16,$i) . "\n";
				$b=1;
			}
		}
		$i = $i + 1;
	}
	$i=0;
	while($i < 0x10)
	{
		if($R[$i] & 0xFFFF0000)
		{
			print "ERROR! wrong bits" . prhex(32,$R[$i]) . " found in Register $Reg[$i] \n";
			$b=1;
		}
		$i = $i + 1;
	}
	if($fl)
	{
		print "Memory and Registers Checked\n";
	}
	return $b;
}

###################### Controller subroutines #################################

sub reset_all {
	$program_started = 0;
	reset_cpu();
	reset_tty();
	reset_clk();
	reset_ptr();
	reset_ptp();
	reset_lpt();
	reset_crd();
	reset_md0();
}

# trace can be used in two ways:
# 1) with pc=A0/P it displays the state of the cpu
# 2  with pc=location of an executed instruction it provides the state of
#    the cpu after the execution of that instruction
my $pctrace = 0;
my $trrep   = 0;
sub trace {
	my $pc = shift;
	my $psw = Fpsw();
	
	if($tracefile) # dump trace in file ?
	{
		if($pc != $pctrace)
		{
			$pctrace=$pc;
			$trrep=0;		
			print "  P" . prhex(16,$pc) . "    M" . prhex(16,$M[$pc]) . "  PSW" . prhex(16,$psw);
			print " CR" . prhex(4,$CR); 
			print " IR" . prhex(24,$intreg) . " IMSK" . prhex(16,$intmsk) . "   BR" . prhex(16,$brkreg) . "\n";
			print " A1";
			print prhex(16,$R[2]);
			print prhex(16,$R[4]);
			print prhex(16,$R[6]);
			print prhex(16,$R[8]);
			print prhex(16,$R[10]);
			print prhex(16,$R[12]);
			print prhex(16,$R[14]);
	
			print prhex(16,$R[1]);
			print prhex(16,$R[3]);
			print prhex(16,$R[5]);
			print prhex(16,$R[7]);
			print prhex(16,$R[9]);
			print prhex(16,$R[11]);
			print prhex(16,$R[13]);
			print prhex(16,$R[15]);
			print "\n";
		}
		else
		{
			if($trrep < 5)
			{
				print TRACE "=\n";
			}
			$trrep=$trrep+1;
		}
	}		
	else
	{
			print "  P" . prhex(16,$pc) . "    M" . prhex(16,$M[$pc]) . "  PSW" . prhex(16,$psw);
			print " CR" . prhex(4,$CR); 
			print " IR" . prhex(24,$intreg) . " IMSK" . prhex(16,$intmsk) . "   BR" . prhex(16,$brkreg) . "\n";
			print " A1";
			print prhex(16,$R[2]);
			print prhex(16,$R[4]);
			print prhex(16,$R[6]);
			print prhex(16,$R[8]);
			print prhex(16,$R[10]);
			print prhex(16,$R[12]);
			print prhex(16,$R[14]);
	
			print prhex(16,$R[1]);
			print prhex(16,$R[3]);
			print prhex(16,$R[5]);
			print prhex(16,$R[7]);
			print prhex(16,$R[9]);
			print prhex(16,$R[11]);
			print prhex(16,$R[13]);
			print prhex(16,$R[15]);
			print "\n";
	}
}

######################### Simulator event handler
sub activate_event {
	my $d = shift;		# device for which event is started
	my $t = shift;      # event time out value
	my $cur;
	my $prev = 0;
	my $acc = 0;        # acccumulated time of active events
	
    if($t < 0) { return -1 };                # error
	if(is_active($d)) { return 0 };          # already active
	
	$time = $time + ($currentinterval - $interval);
	# scan event queue (= linked activated devices)
	for($cur = $event_queue; $cur != 0; $cur = $$cur[NEXT])
	{
		if($t < ($acc + $$cur[TIMEOUT]))
		{
			last;
		}
		$acc = $acc + $$cur[TIMEOUT];
		$prev = $cur;
	}
	if($prev == 0)       # insert new event at head of the queue
	{
		$cur = $$d[NEXT] = $event_queue;   # = 0
		$event_queue = $d;                 # = head
	}
	else                 # insert at previous in the queue
	{
		$cur = $$d[NEXT] = $$prev[NEXT];   # = 0
        $$prev[NEXT] = $d;
	}
	$$d[TIMEOUT] = $t - $acc;
	if($cur != 0)
	{
		# make time of next element relative to new one
		$$cur[TIMEOUT] = $$cur[TIMEOUT] - $$d[TIMEOUT];	
	}
	$interval = $$event_queue[TIMEOUT];
	$currentinterval = $interval;
	return 0;
}

sub cancel_event {
    my $d = shift;                # event t o cancel
	my $cur;
	my $next = 0;
	
    if($event_queue == 0) { return 0 };	
	$time = $time + ($currentinterval - $interval);
	if($event_queue == $d)
	{
		$next = $event_queue = $$d[NEXT];
	}
	else
	{
		for($cur = $event_queue; $cur != 0; $cur = $$cur[NEXT])
		{
			if($$cur[NEXT] = $d)
			{
				$next = $$cur[NEXT] = $$d[NEXT];
				last;
			}
		}
	}
	if($next != 0)
	{
		$$next[TIMEOUT] = $$next[TIMEOUT] + $$d[TIMEOUT];
	}
	$$d[0] = 0;
	if($event_queue != 0)
	{
		$interval = $$event_queue[TIMEOUT];
	}
	else
	{
		$interval = $queue_empty_time;
	}
	$currentinterval = $interval;
	return 0;
}

sub is_active {
	my $d = shift;
	my $acc = 0;
	my $cur;
	for($cur = $event_queue; $cur != 0; $cur = $$cur[NEXT])
	{
		if($cur == $event_queue)
		{
			if($interval > 0)
			{
				$acc = $acc + $interval;
			}
		}
		else
		{
			$acc = $acc + $$cur[TIMEOUT];
		}
		if($cur == $d)
		{
			return $acc + 1;
		}
	}
	return 0;
}

sub process_events {
	my $d;
	my $err;
	
	$time = $time + $currentinterval;
	if($event_queue == 0)
	{
		$interval = $queue_empty_time;
		$currentinterval = $interval;
		$err = 0;
	}
	else
	{
		do
		{
			$d = $event_queue;          # remove 1st element from queue
			$event_queue = $$d[NEXT];   # set next
			$$d[NEXT] = 0;
			if($event_queue != 0)
			{
				$interval = $$event_queue[TIMEOUT]; # set new interval
			}
			else
			{
				$interval = $queue_empty_time;
			}
			$currentinterval = $interval;
			if($$d[SERVICE] != 0)
			{
				my $sub = $$d[SERVICE];
				$err = &$sub;           # execute event service routine
			}
			else
			{
				$err = 0;
			}
			# print "p";
		}
		while(($err == 0) && ($interval == 0));
	}
	return $err;
}

sub show_event_queue {
	my $cur;
	my $d;
	my $t;
	print "Time: $time Current interval: $currentinterval Actual interval: $interval\n";
	for($cur = $event_queue; $cur != 0; $cur = $$cur[NEXT])
	{
		$d = $$cur[NAME];
		$t = $$cur[TIMEOUT];
		print "Device $d Current event time $t\n";
	}
}

######################### CLK device routines
# Real time clocks were delivered with several frequencies (1 ms, 10 ms, etc.),
# depending on the applied cristal. The default was a line frequency clock(20 ms or
# 16.6 ms).

sub cio_clk {
	my $f = shift;                  # function bit
	my $stat = $CLK[STATE] >> 16;   # get status/cu state
	if($f == 0)
	{
		# CIO STOP, always accepted
		if($stat == EXE)
		{
			$CLK[STATE] = $CLK[STATE] | (WREQ << 16);     # state = wait state requested
		}
		else
		{
			 $CLK[STATE] = INAC;
		}
		$CR = 0;
	}
	elsif($stat == INAC)
	{
		# CIO START, accepted if state == inactive
		# F=1 => CIO start clock
		$CLK[STATE] = (EXE << 16);                    # state = execute	
		$CLK[BUFFER] = 0;                             # reset sync period
		$CLK[CTIME] =  $CLK[TIME];                    # start with default interval
		activate_event(\@CLK,$CLK[CTIME]);            # schedule event
		$CR = 0;		
	}
	else
	{
		$CR = 1;                    # not accepted                    
	}	
}

# Clock service routine 
# processes the 'execute state' of the simulated clock device.

sub svc_clk {
	my $stat = ($CLK[STATE] >> 16);	
	if($stat == EXE)
	{
		$intreg = $intreg | RTC;                       # set the ready flipflop
		###### clock sync code
        if($CLK[BUFFER] == 0)                          # new sync period ?
		{
			@clk0 = gettimeofday ();                   # fix start point
		}
		elsif($CLK[BUFFER] == $CLK[SYNC])
		{
		#	# sync period ended
			$laps = tv_interval (\@clk0);              # elapsed sync time in secs
			$pulses = $CLK[CTIME] * $CLK[SYNC];
			# print "$laps secs $pulses pulses\n";
		#	# new interval time
			$CLK[CTIME] = int(($pulses / $laps) / $CLK[FREQ]);
		#	# start a new sync period
			@clk0 = gettimeofday ();                   # fix new start point
			$CLK[BUFFER] = 0;
	    }
		$CLK[BUFFER] = $CLK[BUFFER] + 1;               # sync  period
		activate_event(\@CLK,$CLK[CTIME]);             # schedule a new event
		
	}
	elsif($stat & WREQ)
	{
		# switch directly to inactive state 
		# (so no SST is required to make the clock inactive)
		$CLK[STATE] = INAC;                            # cu = inactive
	}
}

sub inr_clk {
	return 0; # nop
}

sub otr_clk {
	return 0; # nop
}

sub sst_clk {
	return 0; # nop
}

sub tst_clk {
	if( ($CLK[STATE] >> 16) )
	{
		$R[$r3] = 1;                                   # busy, not in inactive state
	}
	else
	{
		$R[$r3] = 0;
	}		
	$CR = 0;
}

sub state_clk {
		print "CU" . prhex(20,$CLK[STATE]) . " INTERVAL $CLK[CTIME]\n";
}

sub reset_clk {
	$intreg = $intreg & (~ RTC);                       # reset the ready flipflop
	$CLK[STATE]  = INAC;                               # state = inactive + 0
	$CLK[BUFFER] = 0;                                  # reset buffer
	$CLK[CTIME]  = 0;
}
		
######################### TTY device routines

sub cio_tty {
	my $f = shift;                  # function bit
	my $stat = $TTY[STATE] >> 16;   # get status/cu state
	if($f == 0)
	{
		# CIO STOP, always accepted
		if(($stat == IEXE) || ($stat == OEXE) || ($stat == IEXC) || ($stat == OEXC))
		{
			$TTY[STATE] = $TTY[STATE] | (WREQ << 16);     # state = wait state requested
		}
		$CR = 0;
	}
	elsif($stat == INAC)
	{
		# CIO START, accepted if state == inactive
		if($R[$r3] == 0)
		{
			# OUTPUT
			$TTY[STATE] = (OEXC << 16);        # state = reset + output + exchange
			$intreg = $intreg | IASR1;         # set the ready flipflop
		}
		else
		{
			# INPUT
			# tty event is already scheduled
		    $TTY[STATE] = (IEXE << 16);        # state = reset + input + execute
			$TTY[BUFFER]  = 0;                 # reset tty buffer
		}
		$CR = 0;
	}
	else
	{
		$CR = 1;                               # not accepted                    
	}
}
	
# tty service routine 
# processes the 'execute state' of the simulated tty cu+device, and
# - simulated panel interrupt (key CntrlP)
# - stopping of the simulator by pressing the stop key CntrE
#
sub svc_tty {
	my $key;
	my $stat = ($TTY[STATE] >> 16);
	my $k;
	
	if(defined ($key = ReadKey(-1)))
	{
		$k = ord ($key);
		if($k == 0xA || $k == 0xD)     # ENTER provides 0A under Unix and 0D under Windows
		{
			print "\r\n";
		}
		else
		{
			if($TTY[CSET] == 64)
			{
				print UP($key);              # Echo, suppress LF(ENTER), which is converted to CR
			}
			else
			{
				print $key;
			}
		}
				
		if($k == 16)	# CntrP
		{
			print "\nPanel interrupt\n";
			trap(CP);                                 # set interrupt
		}
		elsif($k == 5)  # CntrE
		{
			print "\nKeyboard interrupt\n";
			$RUN = 0;
		}
		elsif( $stat == IEXE )                        # input + execute
		{
			if($TTY[BUFFER] == 0)                     # buffer reset?
			{
				if($TTY[CSET] == 64)
				{
					$TTY[BUFFER] = ord(UP($key));     # save character in buffer
				}
				else
				{
					$TTY[BUFFER] = $k;                # save character in buffer	
				}			
				$intreg = $intreg | IASR1;            # set the ready flipflop
				$TTY[STATE] = $TTY[STATE] & 0xFFFF;           # reset cu state bits
				$TTY[STATE] = $TTY[STATE] | (IEXC << 16);     # state = input + exchange
				# wait for INR
			}
			else
			{
				print "Error: TTY-INR did not reset the buffer\n";
			}
		}
		elsif( $stat == IEXC )                        # input + exchange
		{
			# throughput error (INR did not process)
			$intreg = $intreg | IASR1;                # set the ready flipflop
		    $TTY[STATE] = $TTY[STATE] & 0xFFFF;       # reset cu state bits
		    $TTY[STATE] = $TTY[STATE] | (WAIT << 16); # cu state = wait
			$TTY[STATE] = $TTY[STATE] | 2;            # set throughput error bit in state	
		}
		else
		{
			; # ignore key
		}
	}
	
	if( $stat == OEXE )                               # output + execute
	{
		$k=$TTY[BUFFER];                             # print buffered character
		if($k == 0xD)
		{
			print "\r\n";
		}
		elsif($k == 0xA)
		{
			;
		}
		else
		{
			print chr($k);
		}		
		$intreg = $intreg | IASR1;                    # set the ready flipflop
		$TTY[STATE] = $TTY[STATE] & 0xFFFF;           # reset cu state bits
		$TTY[STATE] = $TTY[STATE] | (OEXC << 16);     # state = output + exchange
		# wait for OTR
	}
	
	if( $stat & WREQ)
	{
		# stop io requested (CIO stop)
		if($stat == (OEXE + WREQ))
		{
			if(!($intreg & IASR1))
			{
				# not ready, so still a buffered output character pending
				$k = $TTY[BUFFER];                    # print buffered character
 				if($k == 0xD)
				{
					print "\r\n";
				}
				elsif($k == 0xA)
				{
					;
				}
				else
				{
					print chr($k);
				}	                      
			}
		}
		$intreg = $intreg | IASR1;                    # set the ready flipflop
		$TTY[STATE] = $TTY[STATE] & 0xFFFF;           # reset cu state bits
		$TTY[STATE] = $TTY[STATE] | (WAIT << 16);     # cu state = wait
	}
	# schedule next tty event
	activate_event(\@TTY, $TTY[TIME]);
}

sub inr_tty {
	my $k;
	
	if( ($TTY[STATE] >> 16) == IEXC )
	{
		# cu in exchange state
		$k = $TTY[BUFFER] & 0x7F;                     # buffered character
		if($k == 0xA)                                 # buffered character to r3
		{
			$R[$r3] = 0xD;
		}
		else
		{
			$R[$r3] = $k;
		}               
		$intreg = $intreg & (~ IASR1);                # reset the ready flipflop
		$TTY[STATE] = $TTY[STATE] & 0xFFFF;           # reset cu state bits
		$TTY[STATE] = $TTY[STATE] | (IEXE << 16);     # state = input + execute
		$TTY[BUFFER] = 0;                             # reset the buffer
		$CR = 0;
	}
	else
	{
		$CR = 1;
	}	
}

sub otr_tty {
	if( ($TTY[STATE] >> 16) == OEXC )
	{
		# cu in exchange state
		$TTY[BUFFER] = $R[$r3] & 0x7F;                # buffer character
		$intreg = $intreg & (~ IASR1);                # reset the ready flipflop
		$TTY[STATE] = $TTY[STATE] & 0xFFFF;           # reset cu state bits
		$TTY[STATE] = $TTY[STATE] | (OEXE << 16);     # state = output + execute
		$CR = 0;	
	}
	else
	{
		$CR = 1;
	}
}

sub sst_tty {
	if( ($TTY[STATE] >> 16) == WAIT)
	{
		# cu in wait state
		$TTY[STATE] = $TTY[STATE] & 0xFFFF;           # cu = inactive
		$R[$r3] = $TTY[STATE];                        # r3 = status
		$intreg = $intreg & (~ IASR1);                # reset the ready flipflop
		$CR = 0;
	}
	else
	{
		$CR = 1;	                                  # sst no accepted
	}	
}

sub tst_tty {
	if( ($TTY[STATE] >> 16) )
	{
		$R[$r3] = 1;                                  # busy, not in inactive state
	}
	else
	{
		$R[$r3] = 0;
	}		
	$CR = 0;
}

sub state_tty {
	print "CU" . prhex(20,$TTY[STATE]) . " CR" . prhex(4,$CR) . " BUFFER" . prhex(8,$TTY[BUFFER]) . " CSET $TTY[CSET]" . "\n";
}

sub reset_tty {
	$intreg = $intreg & (~ IASR1);                    # reset the ready flipflop
	$TTY[STATE]  = INAC;                              # state = inactive + 0
	$TTY[BUFFER] = 0;                                 # reset tty buffer
}

######################### PTR device routines

sub cio_ptr {
	my $f = shift;                  # function bit
	my $stat = $PTR[STATE] >> 16;   # get status/cu state
	if($f == 0)
	{
		# CIO STOP, always accepted
		if(($stat == IEXE) || ($stat == IEXC))
		{
			$PTR[STATE] = $PTR[STATE] | (WREQ << 16);     # state = wait state requested
		}
		$CR = 0;
	}
	elsif($stat == INAC)
	{
		# CIO START INPUT, accepted if state == inactive
	    $PTR[STATE] = (IEXE << 16);                   # state = reset + input + execute
		$PTR[BUFFER]  = 0;                            # reset buffer
		$CR = 0;
		# schedule a PTR event
		activate_event(\@PTR, $PTR[TIME]);
	}
	else
	{
		$CR = 1;                                      # not accepted                    
	}
}
	
# ptr service routine 
# processes the 'execute state' of the simulated ptr cu+device.
#
sub svc_ptr {
	my $stat = ($PTR[STATE] >> 16);  
	my $c;
	$PTR[STATE] = $PTR[STATE] & 0xFFFF;                       # reset cu state bits

	if( $stat & WREQ)
	{
		# stop io requested (CIO stop)
		$PTR[STATE] = $PTR[STATE] | (WAIT << 16);             # cu state = wait
	}	
	elsif( $stat == IEXE )                                    # input + execute
	{
		if($PTR[FILE] eq "")
		{
			# not attached, so not operable
	        $PTR[STATE] = $PTR[STATE] | (WAIT << 16);         # cu state = wait
			$PTR[STATE] = $PTR[STATE] | 1;                    # set not operable error bit in state
		}
		elsif($PTR[BUFFER] == 0)                              # buffer reset?
		{
			$c = read_ptr();
			if($c)            
			{
				$PTR[BUFFER] = $c;                            # save character in buffer
				$PTR[STATE]  = ($PTR[STATE] | (IEXC << 16));  # state = input + exchange
				# wait for INR
				# schedule next PTR event
				activate_event(\@PTR, $PTR[TIME]);
			}
			else
			{
				# end of tape
				$PTR[STATE] = $PTR[STATE] | (WAIT << 16);     # cu state = wait
				$PTR[STATE] = $PTR[STATE] | 0x20;             # set eot bit in state	
			}
		}
		else
		{
			print "Error: PTR-INR did not reset the buffer\n";
		}
	}
	elsif( $stat == IEXC )                                    # input + exchange
	{
		# throughput error (INR did not process)
	    $PTR[STATE] = $PTR[STATE] | (WAIT << 16);             # cu state = wait
		$PTR[STATE] = $PTR[STATE] | 2;                        # set throughput error bit in state
	}
	else
	{
		print "Error: PTR service request with CU in wrong state: $stat\n";
	}
	$intreg = $intreg | IPTR1;                                # set the ready flipflop
}

sub inr_ptr {
	
	if( ($PTR[STATE] >> 16) == IEXC )
	{
		# cu in exchange state
		$R[$r3] = $PTR[BUFFER] & 0xFF;                # buffered character
		$intreg = $intreg & (~ IPTR1);                # reset the ready flipflop
		$PTR[STATE] = $PTR[STATE] & 0xFFFF;           # reset cu state bits
		$PTR[STATE] = $PTR[STATE] | (IEXE << 16);     # state = input + execute
		$PTR[BUFFER] = 0;                             # reset the buffer
		$CR = 0;
	}
	else
	{
		$CR = 1;
	}	
}

sub sst_ptr {
	if( ($PTR[STATE] >> 16) == WAIT)
	{
		# cu in wait state
		$PTR[STATE] = $PTR[STATE] & 0xFFFF;           # cu = inactive
		$R[$r3] = $PTR[STATE];                        # r3 = status
		$intreg = $intreg & (~ IPTR1);                # reset the ready flipflop
		$CR = 0;
	}
	else
	{
		$CR = 1;	                                  # sst no accepted
	}	
}

sub tst_ptr {
	if( ($PTR[STATE] >> 16) )
	{
		$R[$r3] = 1;                                  # busy, if not in inactive state
	}
	else
	{
		$R[$r3] = 0;
	}		
	$CR = 0;
}

sub state_ptr {
	print "CU" . prhex(20,$PTR[STATE]) . " CR" . prhex(4,$CR) . " BUFFER" . prhex(8,$PTR[BUFFER]) . "\n";
    if($PTR[FILE] ne "")
	{
		print "Device PTR is attached to file: $PTR[FILE]\n";
	}
	else
	{
		print "Device PTR is not attached to a file\n";
	}
}

sub reset_ptr {
	$intreg = $intreg & (~ IPTR1);                    # reset the ready flipflop
	$PTR[STATE]  = INAC;                              # state = inactive + 0
	$PTR[BUFFER] = 0;                                 # reset buffer
	# if attached, reattach (set file at load point)
	if($PTR[FILE] ne "")
	{
		$file = $PTR[FILE];
		close ($PTR[FC]) || die "can't close $file: $!";
		open (PTR, $file) || die "Cannot open papertape inputfile $file: $!";
		$PTR[FC] = \*PTR;
		binmode $PTR[FC], ":raw";	
		print "File $file reattached to device PTR\n";
	}
}

######################### PTP device routines

sub cio_ptp {
	my $f = shift;                  # function bit
	my $stat = $PTP[STATE] >> 16;   # get status/cu state
	if($f == 0)
	{
		# CIO STOP, always accepted
		if(($stat == OEXE) || ($stat == OEXC))
		{
			$PTP[STATE] = $PTP[STATE] | (WREQ << 16);     # state = wait state requested
		}
		$CR = 0;
	}
	elsif($stat == INAC)
	{
		# CIO START OUTPUT, accepted if state == inactive
	    $PTP[STATE] = (OEXC << 16);                       # state = reset + output + exchange
		$intreg = $intreg | IPTP1;                        # set the ready flipflop
		activate_event(\@PTP,$PTP[TIME]);                 # schedule a new event
		$CR = 0;
	}
	else
	{
		$CR = 1;                                          # not accepted                    
	}
}

# ptp service routine 
# processes the 'execute state' of the simulated ptp cu+device.
#
sub svc_ptp {
	my $stat = ($PTP[STATE] >> 16);  
    #print "1 service ptp called $stat\n";

	if( $stat == OEXE )                                       # output + execute
	{
		if($PTP[FILE] eq "")
		{
			# not attached, so not operable
		    $PTP[STATE] = $PTP[STATE] & 0xFFFF;               # reset cu state bits			
	        $PTP[STATE] = $PTP[STATE] | (WAIT << 16);         # cu state = wait
			$PTP[STATE] = $PTP[STATE] | 1;                    # set not operable error bit in state
			$intreg = $intreg | IPTP1;                        # set the ready flipflop
		}
		else
		{
			punch_ptp($PTP[BUFFER]);
			$PTP[STATE] = $PTP[STATE] & 0xFFFF;               # reset cu state bits	
			$PTP[STATE]  = ($PTP[STATE] | (OEXC << 16));      # state = output + exchange
			$intreg = $intreg | IPTP1;                        # set the ready flipflop
			# wait for next OTR or CIO HALT
			# schedule next PTP event
			activate_event(\@PTP, $PTP[TIME]);
		}
	}
	elsif( $stat & WREQ)
	{
		# stop io requested (CIO stop)
		if($stat == (OEXE + WREQ))
		{
			if(!($intreg & IPTP1))
			{
				# not ready, so still a buffered output character pending
				punch_ptp($PTP[BUFFER]);
			}
		}
		$PTP[STATE] = $PTP[STATE] & 0xFFFF;                   # reset cu state bits
		$PTP[STATE] = $PTP[STATE] | (WAIT << 16);             # cu state = wait
		$intreg = $intreg | IPTP1;                            # set the ready flipflop
	}		
	else
	{
		# do nothing: wait for OTR or CIO HALT
		activate_event(\@PTP, $PTP[TIME]);
	}
}

sub otr_ptp {
	
	if( ($PTP[STATE] >> 16) == OEXC )
	{
		# cu in exchange state
		$PTP[BUFFER] = $R[$r3];                       # buffered character
		$intreg = $intreg & (~ IPTP1);                # reset the ready flipflop
		$PTP[STATE] = $PTP[STATE] & 0xFFFF;           # reset cu state bits
		$PTP[STATE] = $PTP[STATE] | (OEXE << 16);     # state = output + execute
		$CR = 0;
	}
	else
	{
		$CR = 1;
	}	
}

sub sst_ptp {
	if( ($PTP[STATE] >> 16) == WAIT)
	{
		# cu in wait state
		$PTP[STATE] = $PTP[STATE] & 0xFFFF;           # cu = inactive
		$R[$r3] = $PTP[STATE];                        # r3 = status
		$intreg = $intreg & (~ IPTP1);                # reset the ready flipflop
		$CR = 0;
	}
	else
	{
		$CR = 1;	                                  # sst no accepted
	}	
}

sub tst_ptp {
	if( ($PTP[STATE] >> 16) )
	{
		$R[$r3] = 1;                                  # busy, if not in inactive state
	}
	else
	{
		$R[$r3] = 0;
	}		
	$CR = 0;
}

sub state_ptp {
	print "CU" . prhex(20,$PTP[STATE]) . " CR" . prhex(4,$CR) . " BUFFER" . prhex(8,$PTP[BUFFER]) . "\n";
    if($PTP[FILE] ne "")
	{
		print "Device PTP is attached to file: $PTP[FILE]\n";
	}
	else
	{
		print "Device PTP is not attached to a file\n";
	}
}

sub reset_ptp {
	$intreg = $intreg & (~ IPTP1);                    # reset the ready flipflop
	$PTP[STATE]  = INAC;                              # state = inactive + 0
	$PTP[BUFFER] = 0;                                 # reset buffer
	# if attached, reattach
	if($PTP[FILE] ne "")
	{
		$file = $PTP[FILE];
		close ($PTP[FC]) || die "can't close $file: $!";
		open (PTP, ">:raw", $file) || die "Cannot open paper punch file $file: $!";	
		$PTP[FC] = \*PTP;
		print "File $file reattached to device PTP\n";
	}
}

####################### Moving Head Disk routines ############################
# The control unit is wired to a multiplex channel; so only an interrupt when the
# cu switches to wait state. Data is transferred to/from memory from/to the internal 
# disk control unit buffer by the multiplexer/controlunit by means of breaks
# The controlunit/multiplexer channel supports 2 disk drives, the 1st one has
# device number 0 and the 2nd one device number 1.
#

sub cio_md0 {
	my $f  = shift;	
	my $dn = 0;     # device number
	cio_md($f,$dn);
}

sub cio_md1 {
	my $f  = shift;	
	my $dn = 1;     # device number	
	cio_md($f,$dn);	
}

# common CIO processor
sub cio_md {
	my $f  = shift;
	my $dn = shift;
	my $stat = $MD0[STATE] >> 16;   # get status: cu state
	my $r;
	my $d;
	my $c;
	my $p;
	if($f == 0)
	{	
	    # CIO STOP, alsways accepted
		if(($stat == IEXE) || ($stat == IEXC) || ($stat == OEXC)|| ($stat == OEXC))
		{
			$MD0[STATE] = $MD0[STATE] | (WREQ << 16);     # state = wait state requested
		}
	}
	elsif($stat == INAC)
	{
		# CIO START, accepted if state == inactive
		$MD0[STATE] = 0;                                  # reset state
		$md0buf     = "";                                 # reset buffer
		$MD0[XLEN]  = 0;                                  # mpx transferred length in bytes				
		$r = $R[$r3] & 0x3;
		if($r == 0)
		{
			# read a physical sector, CIO START INPUT
			$MD0[SEC]  = ($R[$r3] >> 2) & 0x1F;           # physical sector to read: 0..31
			$MD0[STATE] = $dn << 6;                       # set disc/unit number
			$MD0[STATE] = $MD0[STATE] | (IEXE << 16);     # state = input + execute			
		    activate_event(\@MD0,$MD0[RWTIME]);           # schedule event for moving to sector in cylinder 		
		}
		elsif($r == 1)
		{
			#  write a physical sector, CIO START OUTPUT
			$MD0[SEC]  = ($R[$r3] >> 2) & 0x1F;           # physical sector to write: 0..31	
			$MD0[STATE] = $dn << 6;                       # set disc/unit number			
			$MD0[STATE] = $MD0[STATE] | (OEXC << 16);     # state = output + exchange			
		    activate_event(\@MD0,$MD0[TIME]);             # schedule event to start mpx transfer 		
		}
		elsif($r == 2)
		{
			# move head to another cylinder (SEEK)
			$r=($R[$r3] >> 3) & 0xFF;                     # number of cylinders to move
			$d=($R[$r3] >> 2) & 1;                        # direction, 1 => forward, 0 => back
			$MD0[STATE] = ($dn & 1) << 6;                 # set unit number
			$MD0[STATE] = $MD0[STATE] | (WREQ << 16); # cu state = wait request
			
			if($dn == 0 && $MD0[FILE] ne "")
			{
				if($d)
				{
					$c=$MD0[CYL] + $r;                    # cylinder to move to
				}
				else
				{
					$c=$MD0[CYL] - $r;                    # cylinder to move to
				}
				if($c >=0 && $c < CYLS)
				{
					$p = $c * SPT * NH * LSEC * 2;        # new position in chars
					if(sysseek(MD0, $p, 0) ne undef)      # position heads to requested cylinder
					{
						$MD0[CYL] = $c;
						$MD0[STATE] = $MD0[STATE] | DONCYL;   						
						# schedule event for moving $r cylinders
						activate_event(\@MD0,(($r * $MD0[TRTIME]) + $MD0[RWTIME])); 
					}
					else
					{
						# seek error
						$MD0[STATE] = $MD0[STATE] | DSEEKERR | DONCYL;    					
						activate_event(\@MD0,$MD0[RWTIME]);       # schedule event
					}
				}
				else
				{
					# seek error
					$MD0[STATE] = $MD0[STATE] | DSEEKERR | DONCYL;      					
					activate_event(\@MD0,$MD0[RWTIME]);       # schedule event
				}
			}
			elsif($dn == 1 && $MD1[FILE] ne "")
			{
				if($d)
				{
					$c=$MD1[CYL] + $r;                    # cylinder to move to
				}
				else
				{
					$c=$MD1[CYL] - $r;                    # cylinder to move to
				}
				
				if($c >=0 && $c < CYLS)
				{
					$p = $c * SPT * NH * LSEC * 2;        # new position in chars
					if(sysseek(MD1, $p, 0) ne undef)      # position heads to requested cylinder
					{
						$MD1[CYL] = $c;
						$MD0[STATE] = $MD0[STATE] | DONCYL; 						
						# schedule event for moving $r cylinders
						activate_event(\@MD0,(($r * $MD0[TRTIME]) + $MD0[RWTIME])); 
					}
					else
					{
						# seek error
						$MD0[STATE] = $MD0[STATE] | DSEEKERR | DONCYL;      					
						activate_event(\@MD0,$MD0[RWTIME]);       # schedule event
					}
				}
				else
				{
					# program error (cylinder number outoff range)
					$MD0[STATE] = $MD0[STATE] | DSEEKERR | DONCYL;;      					
					activate_event(\@MD0,$MD0[RWTIME]);       # schedule event
				}
			}			
			else
			{
				# not attached, so not operable
				$MD0[STATE] = $MD0[STATE] | DONCYL | DNOTOP;       # set not operable error bit in state					
			    activate_event(\@MD0,$MD0[RWTIME]);       # schedule event, 				
			}		
		}
		else
		{
			# seek to cylinder 0
			$MD0[STATE] = ($dn & 1) << 6;                  # set unit number			
			$MD0[STATE] = $MD0[STATE] | (WREQ << 16); # cu state = wait request			
			if($dn == 0 && $MD0[FILE] ne "")
			{			    
				if(sysseek(MD0, 0, 0) ne undef)                # position heads to cylinder 0
				{
					$MD0[CYL] = 0;
					$MD0[STATE] = $MD0[STATE] | DONCYL; 					
					# schedule event for moving to cylinder 0
					activate_event(\@MD0,(($MD0[CYL] * $MD0[TRTIME]) + $MD0[RWTIME])); 
				}
				else
				{
					# seek error
					$MD0[STATE] = $MD0[STATE] | DSEEKERR | DONCYL;      					
					activate_event(\@MD0,$MD0[RWTIME]);       # schedule event
				}
			}
			elsif($dn == 1 && $MD1[FILE] ne "")
			{
				if(sysseek(MD1, 0, 0) ne undef)                # position heads to cylinder 0
				{
					$MD1[CYL] = 0;
					$MD0[STATE] = $MD0[STATE] | DONCYL; 					
					# schedule event for moving to cylinder 0
					activate_event(\@MD0,(($MD1[CYL] * $MD0[TRTIME]) + $MD0[RWTIME])); 
				}
				else
				{
					# seek error
					$MD0[STATE] = $MD0[STATE] | DSEEKERR | DONCYL;      					
					activate_event(\@MD0,$MD0[RWTIME]);       # schedule event
				}
			}
			else
			{
				# not attached, so not operable
				$MD0[STATE] = $MD0[STATE] | DNOTOP;       # set not operable error bit in state					
			    activate_event(\@MD0,$MD0[RWTIME]);       # schedule event, 				
			}
		}
		$MD0[DN] = $dn; # save active drive number
		$CR = 0;
	}
	else
	{
		$CR = 1;                                          # CIO not accepted 
	}
}

sub svc_md0 {
	my $stat = $MD0[STATE] >> 16;
	my $dn   = $MD0[DN];                                  # number of active device
	my $pos;                                              # sector position in cylinder
    my $cyl;                                              # current cylider
	if($dn) {$cyl = $MD1[CYL]} else{$cyl=$MD0[CYL]};	  # set current cylinder
	
	if( $stat & WREQ)
	{
		# stop io requested (CIO stop)
    	$MD0[STATE] = $MD0[STATE] & 0xFFFF;                   # reset cu state bits			
		$MD0[STATE] = $MD0[STATE] | (WAIT << 16);             # cu state = wait
		$intreg = $intreg | IMD0;                             # set the ready flipflop			
	}	
	elsif( $stat == IEXE )                                    # input + execute
	{
		$MD0[STATE]  = $MD0[STATE] & 0xFFFF;                  # reset cu state bits			
		if(($dn == 0 && $MD0[FILE] eq "") || ($dn == 1 && $MD1[FILE] eq ""))
		{
			# not attached, so not operable
			$MD0[STATE] = DNOTOP;                             # set not operable error bit in state			
			$MD0[STATE] = $MD0[STATE] | (WAIT << 16);         # cu state = wait
		    $intreg = $intreg | IMD0;                         # set the ready flipflop	
		}
		else
		{
			# read requested physical sector into the device buffer
			$pos = (($cyl * SPT * NH) + $MD0[SEC]) * LSEC * 2;  # sector position in cylinder				
			if($dn==0)
			{
				sysseek(MD0,$pos,0);                          # position and
				$stat = sysread(MD0, $md0buf, LSEC*2);        # read sector
			}
			else
			{
				sysseek(MD1,$pos,0);                          # position
				$stat = sysread(MD1, $md0buf, LSEC*2);        # read sector
			}
			
			#print "b md read result: $stat\n";
			# check sector identifier read sector as an integrity check (real 
			# disk had only cylinder number as identifier)
			# format sector identifier
			# xxxxx sector (5 bit) 0..31 (!!this is the physical sector number!!)
			#      00        (2 bit)
			#        xxxxxxxxx cylinder (9 bit) 0..202
			my $s = unpack("C",substr($md0buf,0,1)) >> 3;     # physical sector number on disk
			my $c = unpack("C",substr($md0buf,1,1));          # cylinder (X1210/1215 cyl number fits in 8 bit)
            #print "s  ". prhex(32,$s) . " c " . prhex(32,$c) . "\n";
			#print "ps ". prhex(32,$MD0[SEC]) . " cyl " . prhex(32,$cyl) . "\n";
            #print "read physical sector $MD0[SEC]\n";
			#print "sec s $s $MD0[SEC] cyl c $c $cyl\n";		

			#if(($s == $MD0[SEC]) && ($stat == LSEC*2) && ($c == $cyl))    # check physical sector number, length an cylinder        
			if(($stat == LSEC*2) && ($c == $cyl))    # check length an cylinder as on real X1201/X1215 
			{
				# internal buffer ok and ready for mpx transfer
				$MD0[BUFFER] = 0;                             # reset word buffer
				$MD0[STATE]  = ($MD0[STATE] | (IEXC << 16));  # state = input + exchange
				# wait for INR
				activate_event(\@MD0, $MD0[TIME]);            # schedule event for mpx transfer
			}
			else
			{
				#error
				$MD0[STATE] = $MD0[STATE] | (WAIT << 16);     # cu state = wait				
				if($stat != LSEC*2)
				{
					# incorrect length
					$MD0[STATE] = $MD0[STATE] | DLENGERR;     # set error bit in state
				}
				else
				{
					# sector identifier error
					$MD0[STATE] = $MD0[STATE] | DDATAERR;     # set error bit in state					
				}
				$intreg = $intreg | IMD0;                     # set the ready flipflop					
			}
		}			
	}
	elsif( $stat == IEXC )                                    # input + exchange
	{
		# move internal device buffer via mpx into memory buffer
		if($MD0[BUFFER] == 0)
		{
			$brkreg = $brkreg | BMD0;                         # set an mpx break request
			activate_event(\@MD0,$MD0[TIME]);                 # schedule next mpx event			
		}
		else
		{
			# throughput error (INR/break did not process)
    	    $MD0[STATE] = $MD0[STATE] & 0xFFFF;               # reset cu state bits				
			$MD0[STATE] = $MD0[STATE] | (WAIT << 16);         # cu state = wait
			$MD0[STATE] = $MD0[STATE] | DTHROERR;             # set throughput error bit in state
		    $intreg = $intreg | IMD0;                         # set the ready flipflop			
		}
	}
	elsif( $stat == OEXC )                                    # output + exchange
	{
		# move memory buffer into internal device buffer via mpx
		$brkreg = $brkreg | BMD0;                             # set an mpx break request
		activate_event(\@MD0,$MD0[TIME]);                     # schedule next mpx event
	}	
	elsif( $stat == OEXE )
	{
		# write device buffer into the requested sector
		$MD0[STATE] = $MD0[STATE] & 0xFFFF;                   # reset cu state bits			
		if(($dn == 0 && $MD0[FILE] eq "") || ($dn == 1 && $MD1[FILE] eq ""))
		{
			# not attached, so not operable
			$MD0[STATE] = DNOTOP;                             # set not operable error bit in state				
			$MD0[STATE] = $MD0[STATE] | (WAIT << 16);         # cu state = wait
		    $intreg = $intreg | IMD0;                         # set the ready flipflop	
		}
		else
		{				
			# check sector identifier to write as a disk integrety check
			# format sector identifier
			# xxxxx sector (5 bit) 0..31 (!!this is the physical sector number!!)
			#      00        (2 bit)
			#        xxxxxxxxx cylinder (9 bit) 0..202
			my $l = length($md0buf);                          # length of the requested buffer to write
			my $s = unpack("C", substr($md0buf,0,1)) >> 3;    # physical sector number IN the sector buffer to write
			my $c = unpack("C", substr($md0buf,1,1));         # cylinder (X1210/1215 cyl number fits in 8 bit)
            # $MD0[SEC] is the physical sector requested to write
			#if($l == 2*LSEC && $c == $cyl && $s == $MD0[SEC])
			if($l == 2*LSEC && $c == $cyl)			
			{
				# sector format ok, write the sector
				$pos = (($cyl * SPT * NH) + $MD0[SEC]) * LSEC * 2;  # sector position in cylinder
				if($dn==0)
				{
					sysseek(MD0,$pos,0);                      # position and				
					$stat = syswrite(MD0, $md0buf, LSEC*2);   # write sector
				}
				else
				{
					sysseek(MD1,$pos,0);                      # position and
					$stat = syswrite(MD1, $md0buf, LSEC*2);   # write sector
				}	
				# and check the result
				if($stat == LSEC*2)            
				{
					# internal buffer transferred to disk
					$MD0[STATE]  = ($MD0[STATE] | (WREQ << 16));  # state = wait request
					activate_event(\@MD0, $MD0[RWTIME]);          # schedule event write event
				}
				else
				{
					# incorrect length				
					$MD0[STATE] = $MD0[STATE] | (WAIT << 16);     # cu state = wait
					$MD0[STATE] = $MD0[STATE] | DLENGERR;         # set error bit in state
					$intreg = $intreg | IMD0;                     # set the ready flipflop				
				}				
			}
			else
			{
				# error in the format of the sector buffer to write; not written to disk			
				$MD0[STATE] = $MD0[STATE] | (WAIT << 16);     # cu state = wait				
				if($l != 2*LSEC)
				{
					# incorrect length
					$MD0[STATE] = $MD0[STATE] | DLENGERR;       # set error bit in state
				}
				#if($s != $MD0[SEC] || $c != $cyl)               # check sector identifier
				if($c != $cyl)                                  # check sector identifier				
				{
					# sector identifier in buffer is corrupted => program error
					# !!!!!!on the real P800 disk, the sector number was not part of the
					# sector identier!!!!!!!!!! (so that test must be removed later)
					$MD0[STATE] = $MD0[STATE] | DPROGERR;         # set error bit in state
				}
				$intreg = $intreg | IMD0;                     # set the ready flipflop				
			}
		}
	}
	else	
	{
		print "Error: MD0 service request with CU in wrong state: $stat\n";
	}	
}

sub inr_md0 {
	# byte or word transfer (internal buffer => memory)
	my $f = shift;    # function bit
	my $t = shift;    # transfer mode: 1=byte, 0=word 
	my $c;
	my $pos = $MD0[XLEN];
	if( $pos < LSEC*2)
	{
		if($t)
		{
			# byte transfer
			$c = substr($md0buf,$pos,1);
			$MD0[BUFFER] = unpack("C", $c);
			$MD0[XLEN]   = $MD0[XLEN] + 1;
		}
		else
		{
			# word transfer
			$c = substr($md0buf,$pos,1);
			$MD0[BUFFER] = unpack("C", $c) << 8;
			$c = substr($md0buf,($pos+1),1);
			$c = unpack("C", $c);
			$MD0[BUFFER] = $MD0[BUFFER] | $c;
			$MD0[XLEN]   = $MD0[XLEN] + 2;
		}
	}
	else
	{
		$MD0[BUFFER] = 0;
	}
}

sub otr_md0 {
	# byte or word transfer (memory => internal buffer)		
	my $f = shift;    # function bit
	my $t = shift;    # transfer mode: 1=byte, 0=word
	my $c;
	if($t)
	{
		# byte transfer
		$c = pack("C", ($MD0[BUFFER] & 0xFF));
		$md0buf=$md0buf . $c;
		$MD0[XLEN] = $MD0[XLEN] + 1;
	}
	else
	{
		# word transfer
        $c = pack("C", (($MD0[BUFFER] >> 8) & 0xFF));		
		$md0buf=$md0buf . $c;
   		$c = pack("C", ($MD0[BUFFER] & 0xFF));  
		$md0buf=$md0buf . $c;
		$MD0[XLEN] = $MD0[XLEN] + 2;		
	}
}

sub sst_md0 {
	if( ($MD0[STATE] >> 16) == WAIT)
	{
		# cu in wait state
		$MD0[STATE] = $MD0[STATE] & 0xFFFF;           # cu = inactive
		$R[$r3] = $MD0[STATE];                        # r3 = status
		$intreg = $intreg & (~ IMD0);                 # reset the ready flipflop
		$CR = 0;
	}
	else
	{
		$CR = 1;	                                  # sst no accepted
	}	
}

sub tst_md0 {
	if( ($MD0[STATE] >> 16) )
	{
		$R[$r3] = 1;                                  # busy, if not in inactive state
	}
	else
	{
		$R[$r3] = 0;
	}		
	$CR = 0;
}

sub state_md0 {
    $t1 = length($md0buf);
	print "CU" . prhex(20,$MD0[STATE]) . " CR" . prhex(4,$CR) . " BUFFER-LENGTH" . prhex(8,$t1) . " TRL" . prhex(16,$MD0[XLEN]) . "\n";
    print "Selected drive:$MD0[DN] PSector:$MD0[SEC] Drive-0 at Cyl:$MD0[CYL] Drive-1 at Cyl:$MD1[CYL]\n";
    if($MD0[FILE] ne "")
	{
		print "Device MD0 is attached to file: $MD0[FILE]\n";
	}
	else
	{
		print "Device MD0 is not attached to a file\n";
	}
    if($MD1[FILE] ne "")
	{
		print "Device MD1 is attached to file: $MD1[FILE]\n";
	}
	else
	{
		print "Device MD1 is not attached to a file\n";
	}	
}

sub reset_md0 {
	$intreg = $intreg & (~ IMD0);                     # reset the ready flipflop
	$brkreg = $brkreg & (~ BMD0);                     # reset break flipflop	
	$MD0[STATE]  = INAC;                              # state = inactive + 0
	$MD0[BUFFER] = 0;                                 # reset buffer
	$MD0[XLEN]   = 0;                                 # reset transferred length
	$MD0[DN]     = 0;                                 # reset to primary device
	$MD0[SEC]    = 0;                                 # reset physical sector
	$MD0[CYL]    = 0;								  # disk 0 at load point
	$MD1[CYL]    = 0;                                 # disk 1 at load point
	$md0buf      = "";
	# if attached, reattach (set file at load point)
	if($MD0[FILE] ne "")
	{
		$file = $MD0[FILE];
		close (MD0) || die "can't close $file: $!";
		open (MD0, "+<:raw", $file) || die "Cannot open  disk file $file: $!";
		print "File $file reattached to device MD0\n";
	}
	if($MD1[FILE] ne "")
	{
		$file = $MD1[FILE];
		close (MD1) || die "can't close $file: $!";
		open (MD1, "+<:raw", $file) || die "Cannot open  disk file $file: $!";
		print "File $file reattached to device MD1\n";
	}	
}

######################### CRD device routines
# The control unit is wired to a multiplex channel; so only an interrupt when the
# cu switches to wait state. Data is transferred to memory from the internal 
# cardreader buffer by the multiplexer/control unit by means of breaks
#
sub cio_crd {
	my $f = shift;                  # function bit
	my $stat = $CRD[STATE] >> 16;   # get status/cu state
	if($f == 0)
	{
		# CIO STOP, always accepted
		if(($stat == IEXE) || ($stat == IEXC))
		{
			$CRD[STATE] = $CRD[STATE] | (WREQ << 16);     # state = wait state requested
		}
		$CR = 0;
	}
	elsif($stat == INAC)
	{
		# CIO START INPUT, accepted if state == inactive
	    $CRD[STATE] = (IEXE << 16);                       # state = reset + input + execute
		# $intreg = $intreg | ILP;                        # set the ready flipflop
		# $brkreg = $brkreg | BLP;                        # set an mpx break request
        $crdbuf = "";                                     # reset buffer
		activate_event(\@CRD,$CRD[RTIME]);                # schedule a event, move next card
		$CR = 0;
	}
	else
	{
		$CR = 1;                                          # not accepted                    
	}
}		
		
sub svc_crd {
	my $stat = ($CRD[STATE] >> 16);  
	
	if( $stat & WREQ)
	{
		# stop io requested (CIO stop)
    	$CRD[STATE] = $CRD[STATE] & 0xFFFF;                   # reset cu state bits			
		$CRD[STATE] = $CRD[STATE] | (WAIT << 16);             # cu state = wait
		$intreg = $intreg | ICR;                              # set the ready flipflop			
	}	
	elsif( $stat == IEXE )                                    # input + execute
	{
		if($CRD[FILE] eq "")
		{
			# not attached, so not operable
    	    $CRD[STATE] = $CRD[STATE] & 0xFFFF;               # reset cu state bits				
	        $CRD[STATE] = $CRD[STATE] | (WAIT << 16);         # cu state = wait
			$CRD[STATE] = $CRD[STATE] | 1;                    # set not operable error bit in state
		    $intreg = $intreg | ICR;                          # set the ready flipflop	
		}
		else
		{
			# read next card into the device buffer
	        my $fc = $CRD[FC];	
			$CRD[RCLEN] = 0;                                 # read length in bytes			
			$stat = read($fc, $crdbuf, 160);
			if($stat == 160)            
			{
				$CRD[BUFFER] = 0;                            # reset word buffer
    	        $CRD[STATE] = $CRD[STATE] & 0xFFFF;          # reset cu state bits					
				$CRD[STATE]  = ($CRD[STATE] | (IEXC << 16));  # state = input + exchange
				# wait for INR
				# schedule next CRD event
				activate_event(\@CRD, $CRD[TIME]);
			}
			else
			{
				# incorrect length (or no card)
    	        $CRD[STATE] = $CRD[STATE] & 0xFFFF;           # reset cu state bits						
				$CRD[STATE] = $CRD[STATE] | (WAIT << 16);     # cu state = wait
				$CRD[STATE] = $CRD[STATE] | 0x20;             # set eof bit in state
		        $intreg = $intreg | ICR;                      # set the ready flipflop				
			}
		}
	}
	elsif( $stat == IEXC )                                    # input + exchange
	{
		if($CRD[BUFFER] == 0)
		{
			$brkreg = $brkreg | BCR;                          # set an mpx break request
			activate_event(\@CRD,$CRD[TIME]);                 # schedule next event			
		}
		else
		{
			# throughput error (INR/break did not process)
    	    $CRD[STATE] = $CRD[STATE] & 0xFFFF;               # reset cu state bits				
			$CRD[STATE] = $CRD[STATE] | (WAIT << 16);         # cu state = wait
			$CRD[STATE] = $CRD[STATE] | 2;                    # set throughput error bit in state
		    $intreg = $intreg | ICR;                          # set the ready flipflop			
		}
	}
	else
	{
		print "Error: CRD service request with CU in wrong state: $stat\n";
	}
}

sub inr_crd {
	# word tarnsfer is assumed!
    my $pos = $CRD[RCLEN]; 
	my $c;
	if($pos < 159)
	{
		$c = substr($crdbuf,$pos,1);
		$CRD[BUFFER] = unpack("C", $c) << 8;
		$c = substr($crdbuf,($pos+1),1);
		$c = unpack("C", $c);
		$CRD[BUFFER] = $CRD[BUFFER] | $c;
		$CRD[RCLEN] = $CRD[RCLEN] + 2;
	}
	else
	{
		$CRD[BUFFER] = 0;
	}
}

sub sst_crd {
	if( ($CRD[STATE] >> 16) == WAIT)
	{
		# cu in wait state
		$CRD[STATE] = $CRD[STATE] & 0xFFFF;           # cu = inactive
		$R[$r3] = $CRD[STATE];                        # r3 = status
		$intreg = $intreg & (~ ICR);                  # reset the ready flipflop
		$CR = 0;
	}
	else
	{
		$CR = 1;	                                  # sst no accepted
	}	
}

sub tst_crd {
	if( ($CRD[STATE] >> 16) )
	{
		$R[$r3] = 1;                                  # busy, if not in inactive state
	}
	else
	{
		$R[$r3] = 0;
	}		
	$CR = 0;
}

sub state_crd {
    $t1 = length($crdbuf);
	print "CU" . prhex(20,$CRD[STATE]) . " CR" . prhex(4,$CR) . " BUFFER-LENGTH" . prhex(8,$t1) . " TRL" . prhex(8,$CRD[RCLEN]) . "\n";
    if($CRD[FILE] ne "")
	{
		print "Device CRD is attached to file: $CRD[FILE]\n";
	}
	else
	{
		print "Device CRD is not attached to a file\n";
	}				
}

sub reset_crd {
	$intreg = $intreg & (~ ICR);                      # reset the ready flipflop
	$brkreg = $brkreg & (~ BCR);                      # reset break flipflop	
	$CRD[STATE]  = INAC;                              # state = inactive + 0
	$CRD[BUFFER] = 0;                                 # reset buffer
	$CRD[RCLEN]  = 0;                                 # reset transferred length
	$crdbuf      = "";
	# if attached, reattach (set file at load point)
	if($CRD[FILE] ne "")
	{
		$file = $CRD[FILE];
		close ($CRD[FC]) || die "can't close $file: $!";
		open (CRD, $file) || die "Cannot open card reader inputfile $file: $!";
		$CRD[FC] = \*CRD;
		binmode $CRD[FC], ":raw";	
		print "File $file reattached to device CRD\n";
	}

}

######################### LPT device routines
# The control unit is wired to a multiplex channel; so only an interrupt when the
# cu switches to wait state. Data is transferred to the internal lineprinter buffer
# by the multiplexer/control unit by means of breaks
# The 1st 2 bytes in the buffer are interpreted as carriage control characters,
# supported as:
# - 0D0A => skip 1 line, after printing the line
# - 0A0A => skip 2 lines, after printing the line
# - 0D0C => print line and skip to top of next new page
# - 0A0C => print line, skip 1 line and skip to top of next new page
#
sub cio_lpt {
	my $f = shift;                  # function bit
	my $stat = $LPT[STATE] >> 16;   # get status/cu state
	if($f == 0)
	{
		# CIO STOP, always accepted
		if(($stat == OEXE) || ($stat == OEXC))
		{
			$LPT[STATE] = $LPT[STATE] | (WREQ << 16);     # state = wait state requested
		}
		$CR = 0;
	}
	elsif($stat == INAC)
	{
		# CIO START OUTPUT, accepted if state == inactive
	    $LPT[STATE] = (OEXC << 16);                       # state = reset + output + exchange
		# $intreg = $intreg | ILP;                        # set the ready flipflop
		# $brkreg = $brkreg | BLP;                        # set an mpx break request
        $lptbuf = "";                                     # reset buffer
		activate_event(\@LPT,$LPT[TIME]);                 # schedule a event
		$CR = 0;
	}
	else
	{
		$CR = 1;                                          # not accepted                    
	}
}

sub svc_lpt {
	my $stat = ($LPT[STATE] >> 16); 

	if( $stat & WREQ)
    {
		$LPT[STATE] = $LPT[STATE] & 0xFFFF;                   # reset cu state bits
		$LPT[STATE] = $LPT[STATE] | (WAIT << 16);             # cu state = wait
		$intreg = $intreg | ILP;                              # set the ready flipflop	
	}
	elsif( $stat == OEXC )                                    # output + exchange
	{
		$brkreg = $brkreg | BLP;                              # set an mpx break request
		activate_event(\@LPT,$LPT[TIME]);                     # schedule next event
	}
	elsif( $stat == OEXE )                                    # output + execute	
	{
		if($LPT[FILE] eq "")
		{
			# not attached, so not operable
		    $LPT[STATE] = $LPT[STATE] & 0xFFFF;               # reset cu state bits			
	        $LPT[STATE] = $LPT[STATE] | (WAIT << 16);         # cu state = wait
			$LPT[STATE] = $LPT[STATE] | 1;                    # set not operable error bit in state
			$intreg = $intreg | ILP;                          # set the ready flipflop
		}
		else
		{
	        my $fc = $LPT[FC];
			my $c1 = ord(substr($lptbuf,0,1));
			my $c2 = ord(substr($lptbuf,1,1));
			my $l  = length($lptbuf);
			# print $l . prhex(8,$c1) . prhex(8,$c2) . "\n";
			
			# skip before printing the buffer
			if($LPT[SKIP] == SKIP_BP)
			{
				if($c1 == 0xA)
				{
					# new line
					print $fc "\r\n";
					$LPT[LCUR] = $LPT[LCUR] + 1;
				}
				if($c2 == 0xA)
				{
					# new line
					print $fc "\r\n";
					$LPT[LCUR] = $LPT[LCUR] + 1;
				}
				if($c2 == 0xC && $LPT[LCUR] != 0)
				{
					# new page if not on new page position
					while( $LPT[LCUR] !=  $LPT[LMAX] + $LPT[NPAG])
					{
						print $fc "\r\n";
						$LPT[LCUR] = $LPT[LCUR] + 1;
					}
					$LPT[LCUR] = 0;
				}
				if($LPT[LCUR] == $LPT[LMAX] )			# end of page ?
				{
					# new page
					while( $LPT[LCUR] != ( $LPT[LMAX] + $LPT[NPAG]))
					{
						print $fc "\r\n";
						$LPT[LCUR] = $LPT[LCUR] + 1;
					}
					$LPT[LCUR] = 0;
				}
			}			
			
			if($LPT[SKIP] == SKIP_NO)
			{
				# no printer control => just print the buffer
				print $fc $lptbuf;
            }
			else
			{
				# printer control: either skip before or after print (set by SKIP_BP or SKIP_AP)
				if($l > 2)
				{
					# strip control characters and print the line
					$lptbuf = substr($lptbuf,2,$l-2);
					print $fc $lptbuf;
				}
			}
			
			# skip page after printing the buffer
			if($LPT[SKIP] == SKIP_AP)
			{
				if($c1 == 0xA)
				{
					# new line
					print $fc "\r\n";
					$LPT[LCUR] = $LPT[LCUR] + 1;
				}
				if($c2 == 0xA)
				{
					# new line
					print $fc "\r\n";
					$LPT[LCUR] = $LPT[LCUR] + 1;
				}
				if($c2 == 0xC && $LPT[LCUR] != 0)
				{
					# new page if not on new page position
					while( $LPT[LCUR] !=  $LPT[LMAX] + $LPT[NPAG])
					{
						print $fc "\r\n";
						$LPT[LCUR] = $LPT[LCUR] + 1;
					}
					$LPT[LCUR] = 0;
				}
				if($LPT[LCUR] == $LPT[LMAX] )			# end of page ?
				{
					# new page
					while( $LPT[LCUR] != ( $LPT[LMAX] + $LPT[NPAG]))
					{
						print $fc "\r\n";
						$LPT[LCUR] = $LPT[LCUR] + 1;
					}
					$LPT[LCUR] = 0;
				}
			}
			
			$LPT[STATE] = $LPT[STATE] | (WREQ << 16);             # state = wait state requested
			activate_event(\@LPT,$LPT[PTIME]);                    # schedule next event (print time)
		}
	}
	else
	{
		print "LP-SVC: Lineprinter CU in unexpected state\n";
	}
}

sub otr_lpt {

	# cu in exchange state (byte transfer is assumed!)
	$lptbuf = $lptbuf . chr($LPT[BUFFER]);
}

sub sst_lpt {
	if( ($LPT[STATE] >> 16) == WAIT)
	{
		# cu in wait state
		$LPT[STATE] = $LPT[STATE] & 0xFFFF;           # cu = inactive
		$R[$r3] = $LPT[STATE];                        # r3 = status
		$intreg = $intreg & (~ ILP);                  # reset the ready flipflop
		$CR = 0;
	}
	else
	{
		$CR = 1;	                                  # sst no accepted
	}	
}

sub tst_lpt {
	if( ($LPT[STATE] >> 16) )
	{
		$R[$r3] = 1;                                  # busy, if not in inactive state
	}
	else
	{
		$R[$r3] = 0;
	}		
	$CR = 0;
}

sub state_lpt {
    $t1 = length($lptbuf);
	print "CU" . prhex(20,$LPT[STATE]) . " CR" . prhex(4,$CR) . " BUFFER-LENGTH" . prhex(8,$t1) . "\n";
	print "BUFFER:.$lptbuf.\n";
    if($LPT[FILE] ne "")
	{
		print "Device LPT is attached to file: $LPT[FILE]\n";
	}
	else
	{
		print "Device LPT is not attached to a file\n";
	}
}

sub reset_lpt {
	$intreg = $intreg & (~ ILP);                      # reset the ready flipflop
	$brkreg = $brkreg & (~ BLP);                      # reset break flipflop
	$LPT[STATE]  = INAC;                              # state = inactive + 0
    $lptbuf	= "";                                     # reset buffer
	$LPT[LCUR]  = 0;                                  # set on line 0 current page 	
	# if attached, reattach
	if($LPT[FILE] ne "")
	{
		$file = $LPT[FILE];
		close ($LPT[FC]) || die "can't close $file: $!";
		open (LPT, ">$file") || die "Cannot open line printer output file $file: $!";	
		$LPT[FC] = \*LPT;
		print "File $file reattached to device LPT\n";
	}
}
############################################################################

sub prx { 
my $v = shift;
my $s = sprintf ("%x", $v);
print $s;
}

sub prhex {
	# print hex
	# $n is number of bits
	# $v is value
	my $n = shift;
	my $v = shift;
	my $s;
	if($n == 0)
	{
		return "";
	}
	elsif($n<5)
	{
		$s = sprintf(" %01X", $v);	
	}
	elsif($n<9)
	{
		$s = sprintf(" %02X", $v);
	}
	elsif($n<13)
	{		
		$s = sprintf(" %03X", $v);
	}
	elsif($n<17)
	{		
		$s = sprintf(" %04X", $v);
	}
	elsif($n<21)
	{
		$s = sprintf(" %05X", $v);
	}
	elsif($n<25)
	{						
		$s = sprintf(" %06X", $v);
	}
	elsif($n<29)
	{		
		$s = sprintf(" %07X", $v);
	}
	elsif($n<33)
	{
		$s = sprintf(" %08X", $v);
	}
	# print "$s ";
	return $s;
}

sub rtrim {
	my $s = shift;
	my $l = length($s);
	my $i = $l;
	if ($l > 0)
	{
		while($i > 0)
		{
			if((substr($s,$i-1,1)) ne " ")
			{ 
				last;
			}
			else
			{
				$i=$i-1;
			}
		}
	}
	else
	{
		return "";
	}
	if($i == 0)
	{
		return "";
	}
	else
	{
		return substr($s,0,$i);
	}
}

sub prdec
{
	# print number
	# $n is number of digits (string)
	# $v is value
	my $n = shift;
	my $v = shift;
	my $s;
	my $format="\%0" . $n . "d";
	$s = sprintf($format, $v);
	# print "$s ";
	return "  $s  ";
}

# convert hex string in expression to binary
sub cvhex
{
	my $s = shift;
	my $i = 0;
	my $r=0;
	my $d=0;
	my $c;
	if(($c=substr($s,$i,1)) =~ /[0-9A-F]/ )	# at least 1 hex digit required
	{
		while (($c = substr($s,$i,1)) =~ /[0-9A-F]/)
		{
			$d = $cvdig{$c};
			$r = ($r << 4) + $d;
			$i = $i + 1;
		}
		
		if ($r > 0xFFFF)
		{
			return -1;
		}
		return $r;
	}
	else
	{
		return -1;
	}
}

# convert decimal string in expression to binary
sub cvdec
{
	my $s = shift;
	my $i = 0;
	my $r=0;
	my $d=0;
	my $c;
	if(($c=substr($s,$i,1)) =~ /[0-9]/ )	# at least 1 hex digit required
	{
		while (($c = substr($s,$i,1)) =~ /[0-9]/)
		{
			$d = $cvdig{$c};
			$r = ($r * 10) + $d;
			$i = $i + 1;
		}
		
		if ($r > 0xFFFF)
		{
			return -1;
		}
		return $r;
	}
	else
	{
		return -1;
	}
}

# translate string to upper case
sub UP {
	my $s = shift;
    $_ = $s;
	tr/a-z/A-Z/;
	return $s = $_;
}

# read octad from papertape, attached to filecode IN (used by SCP load commands)
sub nextchar {
	my $c;
	read(IN,$c,1);
	$c = unpack("C", $c); 	#get new current char
	return ($c & 0xFF);
}

# read word (2 octads) from papertape
sub nextword {
	my $w = nextchar();
	$w = $w << 8;
	$w = $w + nextchar();;
	return $w;
}

# read octad from papertape, attached to filecode PTR
sub read_ptr {
	my $c;
	if(	read($PTR[FC],$c,1) )
	{
		$c = unpack("C", $c); 	         # get next char
		return (($c & 0xFF) | 0x8000);   # nonzero if not eot
	}
	else
	{
		return 0;
	}
}

# punch a byte on papertape connected to filecode PTP
sub punch_ptp {
    my($a) = @_;
    my $byte;
	my $fc = $PTP[FC];
    $byte = pack("C", ($a & 0xFF) );
    print $fc $byte;
}
